Package: ukbnmr
Type: Package
Title: Removal of Unwanted Technical Variation from UK Biobank NMR
        Metabolomics Biomarker Data
Version: 3.3.2
Authors@R: 
    person(given=c("Scott", "C"), family="Ritchie", email="sritchie73@gmail.com", 
           role=c("aut", "cre"), comment=c(ORCID="0000-0002-8454-9548"))
BugReports: https://github.com/sritchie73/ukbnmr/issues
Description: A suite of utilities for working with the UK Biobank 
    <https://www.ukbiobank.ac.uk/> Nuclear Magnetic Resonance spectroscopy (NMR) 
    metabolomics data <https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220>. 
    Includes functions for extracting biomarkers from decoded UK Biobank field 
    data, removing unwanted technical variation from biomarker concentrations,
    computing an extended set of lipid, fatty acid, and cholesterol fractions, 
    and for re-deriving composite biomarkers and ratios after adjusting data 
    for unwanted biological variation. For further details on methods see
    Ritchie SC et al. Sci Data (2023) <doi:10.1038/s41597-023-01949-y>.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 2.10)
Imports: data.table, bit64, lubridate, MASS
Suggests: knitr, rmarkdown
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Config/Needs/development: roxygen2
NeedsCompilation: no
Packaged: 2026-04-12 10:11:01 UTC; sr827
Author: Scott C Ritchie [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8454-9548>)
Maintainer: Scott C Ritchie <sritchie73@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-12 10:20:02 UTC
Built: R 4.5.2; ; 2026-04-12 13:11:28 UTC; unix

Package: ukbnmr
Type: Package
Title: Removal of Unwanted Technical Variation from UK Biobank NMR
        Metabolomics Biomarker Data
Version: 3.4
Authors@R: 
    person(given=c("Scott", "C"), family="Ritchie", email="sritchie73@gmail.com", 
           role=c("aut", "cre"), comment=c(ORCID="0000-0002-8454-9548"))
BugReports: https://github.com/sritchie73/ukbnmr/issues
Description: A suite of utilities for working with the UK Biobank 
    <https://www.ukbiobank.ac.uk/> Nuclear Magnetic Resonance spectroscopy (NMR) 
    metabolomics data <https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220>. 
    Includes functions for extracting biomarkers from decoded UK Biobank field 
    data, removing unwanted technical variation from biomarker concentrations,
    computing an extended set of lipid, fatty acid, and cholesterol fractions, 
    and for re-deriving composite biomarkers and ratios after adjusting data 
    for unwanted biological variation. For further details on methods see
    Ritchie SC et al. Sci Data (2023) <doi:10.1038/s41597-023-01949-y>.
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Depends: R (>= 2.10)
Imports: data.table, bit64, lubridate, MASS
Suggests: knitr, rmarkdown
VignetteBuilder: knitr
Config/Needs/development: roxygen2
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-07 10:55:03 UTC; sr827
Author: Scott C Ritchie [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8454-9548>)
Maintainer: Scott C Ritchie <sritchie73@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-07 11:10:02 UTC
Built: R 4.5.2; ; 2026-07-07 13:45:01 UTC; unix

# Generated by roxygen2: do not edit by hand

export(compute_extended_ratio_qc_flags)
export(compute_extended_ratios)
export(compute_lipoprotein_fatty_acids)
export(compute_lipoprotein_fatty_acids_flags)
export(extract_biomarker_qc_flags)
export(extract_biomarkers)
export(extract_sample_qc_flags)
export(recompute_derived_biomarker_qc_flags)
export(recompute_derived_biomarkers)
export(remove_technical_variation)
import(bit64)
import(data.table)
importFrom(MASS,rlm)
importFrom(lubridate,ymd)
importFrom(lubridate,ymd_hms)
importFrom(stats,coef)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,ppoints)
importFrom(stats,qnorm)
importFrom(stats,sd)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# library(ukbnmr)
# library(data.table) # for fast reading and writing of csv files using fread() and fwrite()
# 
# # Load exported field data saved by the Table Exporter tool on the RAP
# exported <- fread("path/to/exported_ukbiobank_phenotype_data.csv")
# 
# nmr <- extract_biomarkers(exported)
# biomarker_qc_flags <- extract_biomarker_qc_flags(exported)
# sample_qc_flags <- extract_sample_qc_flags(exported)
# 
# fwrite(nmr, file="path/to/nmr_biomarker_data.csv")
# fwrite(biomarker_qc_flags, file="path/to/nmr_biomarker_qc_flags.csv")
# fwrite(sample_qc_flags, file="path/to/nmr_sample_qc_flags.csv")

## -----------------------------------------------------------------------------
library(ukbnmr)

exported <- ukbnmr::test_data # see help("test_data") for more details

nmr <- extract_biomarkers(exported)
biomarker_qc_flags <- extract_biomarker_qc_flags(exported)
sample_qc_flags <- extract_sample_qc_flags(exported)

## ----eval=FALSE---------------------------------------------------------------
# library(ukbnmr)
# library(data.table) # for fast reading and writing of csv files using fread() and fwrite()
# 
# # Load exported field data saved by the Table Exporter tool on the RAP
# exported <- fread("path/to/exported_ukbiobank_phenotype_data.csv")
# 
# processed <- remove_technical_variation(exported)
# 
# fwrite(processed$biomarkers, file="path/to/nmr_biomarker_data.csv")
# fwrite(processed$biomarker_qc_flags, file="path/to/nmr_biomarker_qc_flags.csv")
# fwrite(processed$sample_processing, file="path/to/nmr_sample_qc_flags.csv")
# fwrite(processed$log_offset, file="path/to/nmr_biomarker_log_offset.csv")
# fwrite(processed$outlier_plate_detection, file="path/to/outlier_plate_info.csv")

## -----------------------------------------------------------------------------
library(ukbnmr)

exported <- ukbnmr::test_data # see help("test_data") for more details

processed <- remove_technical_variation(exported)

## ----eval=FALSE---------------------------------------------------------------
# library(ukbnmr)
# library(data.table) # for fast reading and writing of csv files using fread() and fwrite()
# library(MASS) # Robust linear regression
# 
# # Load exported field data saved by the Table Exporter tool on the RAP
# exported <- fread("path/to/exported_ukbiobank_phenotype_data.csv")
# 
# # First we need to remove the effects of technical variation before removing the
# # biological covariates
# processed <- remove_technical_variation(exported)
# tech_qc <- processed$biomarkers
# 
# # Write out all the component results as above
# fwrite(tech_qc, file="path/to/nmr_biomarker_data.csv")
# fwrite(processed$biomarker_qc_flags, file="path/to/nmr_biomarker_qc_flags.csv")
# fwrite(processed$sample_processing, file="path/to/nmr_sample_qc_flags.csv")
# fwrite(processed$log_offset, file="path/to/nmr_biomarker_log_offset.csv")
# fwrite(processed$outlier_plate_detection, file="path/to/outlier_plate_info.csv")
# 
# # Second, we can now create an additional dataset that has been adjusted for
# # age, sex, and BMI.
# 
# # First, we need to extract the relevant fields from UK biobank and format as
# # above, assuming the relevant fields for age (field #21003), sex (field #31),
# # and BMI (field #21001) have also been saved by the Table Exporter tool in
# # path/to/exported_ukbiobank_phenotype_data.csv
# baseline_covar <- exported[, .(eid, age=p21003_i0, sex=p31_i0, bmi=p21001_i0)]
# repeat_covar <- exported[, .(eid, age=p21003_i1, sex=p31_i1, bmi=p21001_i1)]
# covar <- rbind(idcol="visit_index", "0"=baseline_covar, "1"=repeat_covar)
# covar[, visit_index := as.integer(visit_index)]
# 
# # Combine covariates with processed NMR biomarker data. Filter the processed
# # biomarkers only to the non-derived set, as we will plan to rederive the sums
# # and ratios after adjusting for age sex and bmi
# non_derived <- ukbnmr::nmr_info[Type == "Non-derived", Biomarker]
# non_derived <- intersect(non_derived, names(nmr)) # In case some biomarkers are not in the data
# bio_qc <- tech_qc[, .SD, .SDcols=c("eid", "visit_index", non_derived)]
# bio_qc <- covar[bio_qc, on = .(eid, visit_index), nomatch=0]
# 
# # Transform to long format so we can operate on all biomarkers in bulk
# bio_qc <- melt(bio_qc, id.vars=c("eid", "visit_index", "age", "sex", "bmi"),
#   variable.name="biomarker")
# 
# # Log transform biomarkers prior to adjustment. To do so, we need to add a small
# # offset to any measurements equal to 0. This code is copied directly from the
# # internals of the remove_technical_variation() function
# log_offset <- bioqc[!is.na(value),
#   .(Minimum=min(value), Minimum.Non.Zero=min(value[value != 0])),
#   by=biomarker]
# log_offset[, Log.Offset := ifelse(Minimum == 0, Minimum.Non.Zero / 2, 0)]
# bio_qc[log_offset, on = .(biomarker), log_value := log(value + Log.Offset)]
# 
# # Now adjust for age, sex, and (log) BMI using robust linear regression from the MASS package
# bio_qc[, adjusted := rlm(biomarker ~ age + factor(sex) + log(bmi))$residuals, by=biomarker]
# 
# # The adjusted residuals are normally distributed with mean 0. We can rescale to
# # absolute units by subtracting the mean of the original log concentrations, then
# # undoing the log transformation. This code is copied directly from the internals
# # of the remove_technical_variation() function.
# bio_qc[, adjusted := adjusted + as.vector(coef(rlm(value ~ 1)))[1], by=biomarker]
# bio_qc[, adjusted := exp(adjusted)]
# bio_qc[log_offset, on = .(biomarker), adjusted := adjusted - Log.Offset] # remove log offset
# 
# # Some values that were 0 are now < 0, apply small right shift for these biomarkers
# # (the shift should be very small, essentially numeric error. This is worth double
# # checking!). Again, this code is copied directly from the internals of the
# # remove_technical_variation() function.
# shift <- bio_qc[, .(Right.Shift=-pmin(0, min(adjusted))), by=biomarker]
# log_offset <- log_offset[shift, on = .(biomarker)]
# bio_qc[log_offset, on = .(biomarker), adjusted := adjusted + Right.Shift]
# 
# # Cast back to wide format so that each biomarker has its own column
# bioqc <- dcast(bio_qc, eid + visit_index ~ biomarker, value.var="adjusted")
# 
# # Now we recompute the composite biomarkers and derived ratios and save the result.
# bio_qc <- recompute_derived_biomarkers(bio_qc)
# fwrite(bio_qc, file="path/to/nmr_biomarkers_adjusted_for_covariates.csv")
# 
# # You may also want to aggregate and save the quality control flags for each sample from
# # the biomarkers underlying each derived biomarker or ratio, adding them as additional
# # columns to the input data (see help("recompute_derived_biomarker_qc_flags")).
# biomarker_qc_flags <- recompute_derived_biomarker_qc_flags(nmr)
# fwrite(biomarker_qc_flags, file="path/to/biomarker_qc_flags.csv")

#' Extract NMR metabolomic biomarkers from a data.frame of UK Biobank fields
#'
#' Given an input \code{data.frame} loaded from a dasaset of
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{NMR metabolomics fields}
#' extracted by the Table Exporter tool on the
#' \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform},
#' this function extracts the NMR metabolomics biomarkers giving them short variable
#' names as listed in the \code{\link{nmr_info}} information data sheet
#' available in this package.
#'
#' @details
#' Data sets extracted on the \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}
#' have one row per UK Biobank participant, whose project specific sample
#' identifier is given in the first column named "eid". Columns following this
#' follow a \href{https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#database-columns}{naming scheme}
#' based on the unique identifier of each field, assessment visit, and (optionally if relevant)
#' repeated measurement of "p<field_id>_i<visit_index>_a<repeat_index>". For example,
#' the measurement of 3-Hydroxybutyrate at baseline assessment has the column name
#' "p23474_i0". For the UKB NMR data, measurements are available at baseline assessment
#' and the first repeat assessment (e.g. "p23474_i1"). For the UKB NMR data, the
#' <repeat_index> is reserved for cases where biomarker measurements have more than
#' one QC Flag (see \code{\link{extract_biomarker_qc_flags}()}).
#'
#' The \code{data.frame} returned by this function gives each field a unique
#' recognizable name, with measurements from baseline and repeat assessment
#' given in separate rows. The "visit_index" column immediately after the "eid"
#' column indicates whether the biomarker measurement was quantified from the
#' blood samples taken at baseline assessment (visit_index == 0) or first repeat
#' assessment (visit_index == 1). Rows are uniquely identifiable
#' by the combination of entries in columns "eid" and "visit_index".
#'
#' This function will also work with data predating the Research Analysis Platform,
#' including data sets extracted by the \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#' tool and/or the \code{ukbtools} R package.
#'
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' A \code{data.table} will be returned instead of a \code{data.frame} if the
#' the user has loaded the package into their R session.
#'
#' @param x \code{data.frame} with column names "eid" followed by extracted
#'           fields e.g. "p23474_i0", "p23474_i1", \dots, "p23467_i1".
#' @return a \code{data.frame} or \code{data.table} with column names "eid",
#'         and "visit_index", followed by columns for each biomarker
#'        e.g. "bOHbutyrate", \dots, "Valine".
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' nmr <- extract_biomarkers(ukb_data)
#'
#' @export
extract_biomarkers <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  if (detect_format(x, type="biomarkers") == "processed") {
    return(x) # return as is; process_data() would make a copy first
  } else {
    returnDT(process_data(x, type="biomarkers"))
  }
}

#' Extract NMR biomarker QC flags from a data.frame of UK Biobank fields
#'
#' Given an input \code{data.frame} loaded from a dasaset of
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{NMR metabolomics QC indicator fields}
#' extracted by the Table Exporter tool on the
#' \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform},
#' this function extracts the
#' \href{https://biobank.ndph.ox.ac.uk/showcase/showcase/docs/nmrm_app4.pdf}{quality control (QC) flags for the NMR metaolomics biomarkers}
#' giving them short variable names as listed in the \code{\link{nmr_info}} information data sheet
#' available in this package. QC Flags are separated by "; " in each column where
#' there are multiple QC Flags for a single measurement.
#'
#' @details
#' #' Data sets extracted on the \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}
#' have one row per UK Biobank participant, whose project specific sample
#' identifier is given in the first column named "eid". Columns following this
#' follow a \href{https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#database-columns}{naming scheme}
#' based on the unique identifier of each field, assessment visit, and (optionally if relevant)
#' repeated measurement of "p<field_id>_i<visit_index>_a<repeat_index>". For example,
#' the QC flags for the measurement of 3-Hydroxybutyrate at baseline assessment
#' has the column names "p23774_i0_a0", "p23774_i0_a1", and "p23774_i0_a2"; indicating
#' that the 3-Hydroxybutyrate measurement can have up to three QC flags per
#' sample at baseline assessment. Measurements for blood samples collected at the
#' first repeat assessment have 1 in the visit index, e.g. for 3-Hydroxybutyrate
#' at the first repeat assessment there are three columns "p23774_i1_a0",
#' "p23774_i1_a1", "p23774_i1_a2".
#'
#' The \code{data.frame} returned by this function gives each field a unique
#' recognizable name, with measurements from baseline and repeat assessment
#' given in separate rows. The "visit_index" column immediately after the "eid"
#' column indicates whether the biomarker measurement was quantified from the
#' blood samples taken at baseline assessment (visit_index == 0) or first repeat
#' assessment (visit_index == 1). Where multiple QC flags were present at the
#' same measurement, these are collated into a single entry with the multiple QC
#' flags separated by a "; ". Rows are uniquely identifiable by the combination
#' of entries in columns "eid" and "visit_index".
#'
#' This function will also work with data predating the Research Analysis Platform,
#' including data sets extracted by the \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#' tool and/or the \code{ukbtools} R package.
#'
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' A \code{data.table} will be returned instead of a \code{data.frame} if the
#' the user has loaded the package into their R session.
#'
#' @param x \code{data.frame} with column names "eid" followed by extracted
#'           fields e.g. "p23774_i0_a0", \dots "p23774_i1_a0", \dots, "p23767_i1".
#' @return a \code{data.frame} or \code{data.table} with column names "eid",
#'        and "visit_index" followed by columns for each biomarker
#'        e.g. "bOHbutyrate", \dots, "Valine".
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' biomarker_qc_flags <- extract_biomarker_qc_flags(ukb_data)
#'
#' @export
extract_biomarker_qc_flags <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  if (detect_format(x, type="biomarker_qc_flags") == "processed") {
    return(x) # return as is; process_data() would make a copy first
  } else {
    returnDT(process_data(x, type="biomarker_qc_flags"))
  }
}

#' Extract NMR sample QC flags from a data.frame of UK Biobank fields
#'
#' Given an input \code{data.frame} loaded from a dasaset of
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{NMR metabolomics processing fields}
#' extracted by the Table Exporter tool on the
#' \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}, this function
#' extracts the
#' \href{https://biobank.ndph.ox.ac.uk/showcase/showcase/docs/nmrm_app4.pdf}{sample quality control flags for the NMR metabolomics biomarker data} giving them short variable names as listed in the
#' \code{\link{sample_qc_info}} information data sheet available in this package.
#'
#' @details
#' Data sets extracted on the \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}
#' have one row per UK Biobank participant, whose project specific sample
#' identifier is given in the first column named "eid". Columns following this
#' follow a \href{https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#database-columns}{naming scheme}
#' based on the unique identifier of each field, assessment visit, and (optionally if relevant)
#' repeated measurement of "p<field_id>_i<visit_index>_a<repeat_index>". For example,
#' the Shipment Plate for each sample collected at baseline assessment has the
#' column name "p23649_i0". For the UKB NMR data, measurements are available at
#' baseline assessment and the first repeat assessment (e.g. "p23649_i1"). For the
#' UKB NMR data, the <repeat_index> is reserved for cases where biomarker
#' measurements have more than one QC Flag (see \code{\link{extract_biomarker_qc_flags}()}).
#'
#' The \code{data.frame} returned by this function gives each field a unique
#' recognizable name, with measurements from baseline and repeat assessment
#' given in separate rows. The "visit_index" column immediately after the "eid"
#' column indicates whether the biomarker measurement was quantified from the
#' blood samples taken at baseline assessment (visit_index == 0) or first repeat
#' assessment (visit_index == 1). Rows are uniquely identifiable
#' by the combination of entries in columns "eid" and "visit_index".
#'
#' This function will also work with data predating the Research Analysis Platform,
#' including data sets extracted by the \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#' tool and/or the \code{ukbtools} R package.
#'
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' A \code{data.table} will be returned instead of a \code{data.frame} if the
#' the user has loaded the package into their R session.
#'
#' @param x \code{data.frame} with column names "eid" followed by extracted
#'           fields e.g. "p23649_i0", "p23649_i1", \dots, "p23655_i1".
#' @return a \code{data.frame} or \code{data.table} with column names "eid"
#'        and "visit_index", followed by columns for each sample
#'        QC tag, e.g. "Shipment.Plate", \dots, "Low.Protein".
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' sample_qc_flags <- extract_sample_qc_flags(ukb_data)
#'
#' @import bit64
#' @export
extract_sample_qc_flags <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  if (detect_format(x, type="sample_qc_flags") == "processed") {
    return(x) # return as is; process_data() would make a copy first
  } else {
    returnDT(process_data(x, type="sample_qc_flags"))
  }
}

detect_format <- function(x, type) {
  # Silence CRAN NOTES about undefined global variables (columns in data.tables)
  UKB.Field.ID <- QC.Flag.Field.ID <- NULL

  if (type == "biomarkers") {
    field_ids <- ukbnmr::nmr_info[, na.omit(UKB.Field.ID)]
  } else if (type == "biomarker_qc_flags") {
    field_ids <- ukbnmr::nmr_info[, na.omit(QC.Flag.Field.ID)]
  } else if (type == "sample_qc_flags") {
    field_ids <- ukbnmr::sample_qc_info$UKB.Field.ID
  } else {
    stop("internal error: 'type' must be one of \"biomarkers\", \"biomarker_qc_flags\", or \"sample_qc_flags\"")
  }

  raw_ukb_cols <- c(paste0(field_ids, "-0.0"), paste0(field_ids, "-1.0"))
  ukbtools_cols <- c(paste0(field_ids, "_0_0"), paste0(field_ids, "_1_0"))
  ukbconv_r_cols <- c(paste0("f.", field_ids, ".0.0"), paste0("f.", field_ids, ".1.0"))
  ukb_rap_cols <- c(paste0("p", field_ids, "_i0"), paste0("p", field_ids, "_i1"))
  if (!is.data.frame(x)) {
    stop("Input data must be a data.frame or data.table")
  }
  if (type != "sample_qc_flags" & length(intersect(names(x), ukbnmr::nmr_info[["Biomarker"]])) > 0) {
    return("processed")
  } else if (type == "sample_qc_flags" & length(intersect(names(x), ukbnmr::sample_qc_info[["Name"]])) > 0) {
    return("processed")
  } else if (length(intersect(names(x), raw_ukb_cols)) > 0) {
    return("raw")
  } else if (length(intersect(gsub(".*_f", "", names(x)), ukbtools_cols)) > 0) {
    return("ukbtools")
  } else if (length(intersect(names(x), ukbconv_r_cols)) > 0) {
    return("ukbconv_r")
  } else if (length(intersect(names(x), ukb_rap_cols)) > 0) {
    return("ukb_rap")
  } else {
    if (type == "biomarkers") {
      stop("NMR biomarker fields (#23400-23648) not found in input data")
    } else if (type == "biomarker_qc_flags") {
      stop("NMR biomarker QC Flag fields (#23700-23948) not found in input data")
    } else if (type == "sample_qc_flags") {
      stop("NMR sample QC Flag fields (#23649-23660) not found in input data")
    } else {
      stop("internal error: 'type' must be one of \"biomarkers\", \"biomarker_qc_flags\", or \"sample_qc_flags\"")
    }
  }
}

#' Remove technical variation from NMR biomarker data in UK Biobank.
#'
#' @param x \code{data.frame} containing a tabular phenotype dataset extracted by the
#' \href{https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#create-a-tsv-or-csv-file-using-table-exporter}{Table Exporter} tool on the
#' \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}
#' containing the project specific sample id (eid) and all fields (and instances)
#' relating to the NMR metabolomics data (i.e. fields listed in showcase categories
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{220},
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{221}, and
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{222}.
#' @param remove.outlier.plates logical, when set to \code{FALSE} biomarker
#' concentrations on outlier shipment plates (see details) are not set to
#' missing but simply flagged in the \code{biomarker_qc_flags} \code{data.frame}
#' in the returned \code{list}.
#' @param skip.biomarker.qc.flags logical, when set to \code{TRUE} biomarker QC
#' flags are not processed or returned.
#' @param version version of the QC algorithm to apply. Defaults to 3, the latest
#' version of the algorithm (see details).
#'
#' @details
#' Three versions of the QC algorithm have been developed. Version 1 was designed
#' based on the first phase of data released to the public covering ~120,000
#' UK Biobank participants. Version 2 made several improvements to the algorithm
#' based on the subsequent second public release of data covering an additional
#' ~150,000 participants. Version 3, the default, makes some further minor tweaks
#' primarily so that the algorithm is compatible with the full public data
#' release covering all ~500,000 participants.
#'
#' Details on the impact of these algorithms on technical variation in the latest
#' UK Biobank data are provided in the package vignette and on the github readme.
#'
#' ## Version 1
#' Setting version to 1 applies the algorithm as exactly described in Ritchie S.
#' C. *et al.* _Sci Data_ 2023. In brief, this multi-step procedure applies the
#' following steps in sequence:
#'
#' 1. First biomarker data is filtered to the 107 biomarkers that cannot be
#'    derived from any combination of other biomarkers.
#' 2. Absolute concentrations are log transformed, with a small offset
#'    applied to biomarkers with concentrations of 0.
#' 3. Each biomarker is adjusted for the time between sample preparation
#'    and sample measurement (hours).
#' 4. Each biomarker is adjusted for systematic differences between rows
#'    (A-H) on the 96-well shipment plates.
#' 5. Each biomarker is adjusted for remaining systematic differences
#'    between columns (1-12) on the 96-well shipment plates.
#' 6. Each biomarker is adjusted for drift over time within each of the six
#'    spectrometers. To do so, samples are grouped into 10 bins, within each
#'    spectrometer, by the date the majority of samples on their respective
#'    96-well plates were measured.
#' 7. Regression residuals after the sequential adjustments are
#'    transformed back to absolute concentrations.
#' 8. Samples belonging to shipment plates that are outliers of
#'    non-biological origin are identified and set to missing.
#' 9. The 61 composite biomarkers and 81 biomarker ratios are recomputed
#'    from their adjusted parts.
#' 10. An additional 76 biomarker ratios of potential biological
#'     significance are computed.
#'
#' At each step, adjustment for technical covariates is performed using
#' \link[MASS:rlm]{robust linear regression}. Plate row, plate column, and
#' sample measurement date bin are treated as factors, using the group with the
#' largest sample size as reference in the regression.
#'
#' ## Version 2
#' Version 2 of the algorithm modifies the above procedure to (1)
#' adjust for well and column within each processing batch separately in steps 4
#' and 5, and (2) in step 6 instead of splitting samples into 10 bins per
#' spectrometer uses a fixed bin size of approximately 2,000 samples per bin,
#' ensuring samples measured on the same plate and plates measured on the same
#' day are grouped into the same bin.
#'
#' The first modification was made as applying version 1 of the algorithm to the
#' combined data from the first and second tranche of measurements revealed
#' introduced stratification by well position when examining the corrected
#' concentrations in each data release separately.
#'
#' The second modification was made to ensure consistent bin sizes across data
#' releases when correcting for drift over time. Otherwise, spectrometers used
#' in multiple data releases would have different bin sizes when adjusting
#' different releases. A bin split is also hard coded on spectrometer 5 between
#' plates 0490000006726 and 0490000006714 which correspond to a large change in
#' concentrations akin to a spectrometer recalibration event most strongly
#' observed for alanine concentrations.
#'
#' ## Version 3
#' Version 3 of the algorithm makes two further minor changes:
#'
#' 1. Imputation of missing sample measurement times has been improved. Previously,
#'    any samples missing time of measurement (N=3 in the phase 2 public release)
#'    had their time of measurement set to 00:00. In version 3, the time of measurement
#'    is set to the median time of measurement for that spectrometer on that day,
#'    which is between 12:00-13:00, instead of 00:00.
#' 2. Samples missing sample preparation dates and times (N=182 in the V20 UK
#'    Biobank data release) use their sample centrifugation date and time as a
#'    proxy to allow adjustment for time between sample prep and sample
#'    measurement. Sample centrifugation always takes place a short time after
#'    sample preparation. As sample centrifugation date and time is not made
#'    available via UK Biobank, this data and time is hard coded; all samples
#'    missing sample preparation date and time had a sample centrifugation date
#'    and time of 2022-12-20 06:39:03 in the extended advance access data, so
#'    we use this date and time for any samples missing sample preparation date
#'    and time.
#' 3. Underlying code for adjusting drift over time has been modified to accommodate
#'    the phase 3 public release, which includes one spectrometer with ~2,500 samples.
#'    Version 2 of the algorithm would split this into two bins, whereas version 3
#'    keeps this as a single bin to better match the bin sizes of the rest of the
#'    spectrometers.
#'
#' @return a \code{list} containing three \code{data.frames}: \describe{
#'   \item{biomarkers}{A \code{data.frame} with column names "eid",
#'        and "visit_index", containing project-specific sample identifier and
#'        UK Biobank visit index (0 for baseline assessment, 1 for first repeat
#'        assessment), followed by columns for each biomarker containing their
#'        absolute concentrations (or ratios thereof) adjusted for technical
#'        variation. See \code{\link{nmr_info}} for information on each biomarker.}
#'   \item{biomarker_qc_flags}{A \code{data.frame} with the same format as
#'         \code{biomarkers} with entries corresponding to the
#'         \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{quality
#'         control indicators} for each sample. "High plate outlier" and "Low
#'         plate outlier" indicate the value was set to missing due to systematic
#'         abnormalities in the biomarker's concentration on the sample's shipment
#'         plate compared to all other shipment plates (see Details). For
#'         composite and derived biomarkers, quality control flags are aggregates
#'         of any quality control flags for the underlying biomarkers from which
#'         the composite biomarker or ratio is derived.}
#'   \item{sample_processing}{A \code{data.frame} containing the
#'         \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{processing
#'         information and quality control indicators} for each sample, including
#'         those derived for removal of unwanted technical variation by this
#'         function. See \code{\link{sample_qc_info}} for details.}
#'   \item{log_offset}{A \code{data.frame} containing diagnostic information on
#'         the offset applied so that biomarkers with concentrations of 0 could
#'         be log transformed, and any right shift applied to prevent negative
#'         concentrations after rescaling adjusted residuals back to absolute
#'         concentrations. Should contain only biomarkers with minimum
#'         concentrations of 0 (in the "Minimum" column). "Minimum.Non.Zero"
#'         gives the smallest non-zero concentration for the biomarker.
#'         "Log.Offset" the small offset added to all samples prior to
#'         log transformation: half the mininum non-zero concentration.
#'         "Right.Shift" gives the small offset added to prevent negative
#'         concentrations that arise after rescaling residuals to log
#'         concentrations: this should be at least one order of magnitude
#'         smaller than the smallest non-zero value (i.e. the offset added
#'         should amount to noise in numeric precision for all samples). See
#'         publication for more details.}
#'   \item{outlier_plate_detection}{A \code{data.frame} containing diagnostic
#'         information and details of outlier plate detection. For each of the
#'         107 non-derived biomarkers, the median concentration on each of the
#'         1,352 plates was calculated, then plates were flagged as outliers if
#'         their median value deviated more than expected from the mean of plate
#'         medians. "Mean.Plate.Medians" gives the mean of the plate medians for
#'         each biomarker. "Lower.Limit" and "Upper.Limit" give the values below
#'         and above which plates are flagged as outliers based on their plate
#'         median. See publication for more details.}
#'   \item{algorithm_version}{Version of the algorithm run, either 1, 2, or 3
#'         (default).}
#' }
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' processed <- remove_technical_variation(ukb_data)
#'
#' @importFrom lubridate ymd_hms
#' @importFrom lubridate ymd
#' @importFrom stats coef
#' @importFrom stats sd
#' @importFrom stats qnorm
#' @importFrom stats ppoints
#' @importFrom stats median
#' @importFrom MASS rlm
#'
#' @export
remove_technical_variation <- function(
  x,
  remove.outlier.plates=TRUE,
  skip.biomarker.qc.flags=FALSE,
  version=3L
) {
  # Silence CRAN NOTES about undefined global variables (columns in data.tables)
  Well.Position.Within.Plate <- Well.Row <- Well.Column <- Sample.Measured.Date.and.Time <-
    Sample.Prepared.Date.and.Time <- Sample.Measured.Date <- Sample.Measured.Time <-
    Sample.Prepared.Date <- Sample.Prepared.Time <- Prep.to.Measure.Duration <-
    Spectrometer.Date.Bin <- Spectrometer <- Type <- Biomarker <- eid <- visit_index <-
    Shipment.Plate <- value <- Log.Offset <- Minimum <- Minimum.Non.Zero <-
    log_value <- adj <- Right.Shift <- outlier <- Lower.Limit <- Upper.Limit <-
    Name <- UKB.Field.ID <- N <- Plate.Measured.Date <- i.Sample.Measured.Date <-
    Processing.Batch <- Spectrometer.Group <- max_bin <- i.offset <- i.Processing.Batch <-
    NULL

  # Check for valid algorithm version
  if (version != 1L && version != 2L && version != 3L) stop("'version' must be 1, 2, or 3")

  message("Checking for revelant UKB fields...")

  # Check relevant fields exist
  f1 <- detect_format(x, type="biomarkers")
  f2 <- if (skip.biomarker.qc.flags) { "skipped" } else { detect_format(x, type="biomarker_qc_flags") }
  f3 <- detect_format(x, type="sample_qc_flags")

  # Make sure its not a "processed" dataset returned by extract_biomarkers() and
  # related functions.
  if (f1 == "processed" || f2 == "processed" || f3 == "processed") {
    stop("'x' should be a 'data.frame' containing UK Biobank field ID columns, not one extracted by extract_biomarkers() or related functions'")
  }

  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  message("Extracting and pre-processing data...")

  # Extract biomarkers, biomarker QC flags, and sample processing information
  bio <- process_data(x, type="biomarkers")
  if (!skip.biomarker.qc.flags) {
    bio_qc <- process_data(x, type="biomarker_qc_flags")
  }
  sinfo <- process_data(x, type="sample_qc_flags")

  message("Checking for required sample processing fields needed for QC procedure...")

  # Check required sample processing fields exist
  req <- c("Well.Position.Within.Plate", "Sample.Measured.Date.and.Time",
           "Sample.Prepared.Date.and.Time", "Spectrometer", "Shipment.Plate")
  miss_req <- setdiff(req, names(sinfo))
  if (length(miss_req) > 0) {
    err_txt <- ukbnmr::sample_qc_info[Name %in% miss_req]
    err_txt <- err_txt[, sprintf("%s (Field: %s)", Name, UKB.Field.ID)]
    err_txt <- paste(err_txt, collapse=", ")
    stop("Missing required sample processing fields: ", err_txt, ".")
  }

  if (version > 1L && !("Processing.Batch" %in% names(sinfo))) {
    warning("Processing.Batch missing (Field 20282), inferring from Shipment.Plate (Field 23649)")
    sinfo[plate_batch_map, on = list(Shipment.Plate), Processing.Batch := i.Processing.Batch]
    if (any(is.na(sinfo$Processing.Batch))) {
      warning("Inference of Processing.Batch from Shipment.Plate failed, reverting to algorithm version 1")
      version <- 1L
    }
  }

  message("Processing sample processing fields for QC procedure...")

  # Split out row and column information from 96-well plate information
  sinfo[, Well.Row := gsub("[0-9]", "", Well.Position.Within.Plate)]
  sinfo[, Well.Column := as.integer(gsub("[A-Z]", "", Well.Position.Within.Plate))]

  # Split out date and time for sample measurement and sample prep
  sinfo[, c("Sample.Measured.Date", "Sample.Measured.Time") := IDateTime(harmonize_datetime(Sample.Measured.Date.and.Time))]
  sinfo[, c("Sample.Prepared.Date", "Sample.Prepared.Time") := IDateTime(harmonize_datetime(Sample.Prepared.Date.and.Time))]

  # Handle the handful of samples missing a sample measurement time. In the above
  # these default to 00:00:00, instead, impute these to the median time of measurement
  # for that spectrometer on that day
  if (version == 3L) {
    missing_time <- sinfo[, which(is.na(suppressWarnings(ymd_hms(Sample.Measured.Date.and.Time))))]
    for (idx in missing_time) {
      this_date <- sinfo[idx, Sample.Measured.Date]
      this_spectrometer <- sinfo[idx, Spectrometer]
      median_measure_time <- sinfo[Sample.Measured.Date == this_date & Spectrometer == this_spectrometer, median(Sample.Measured.Time)]
      sinfo[idx, Sample.Measured.Time := median_measure_time]
    }
  }

  # In the V20 data release 182 samples are missing both a sample preparation
  # date and sample preparation time, which are needed to compute hours between
  # sample prep and sample measurement. In the advance access data, we also had
  # access to date+time of sample centrifugation, which always takes place a
  # short time after sample prep, so we substituted that date+time in as a proxy
  # for sample preparation date+time to QC the data. These samples all shared a
  # sample centrifugation date and time of 2022-12-20 06:39:03, which makes it
  # possible to hard-code this value in the package to substitute in for these
  # samples, as centrifugation date+time is not released by UK Biobank.
  if (version == 3L) {
    sinfo[is.na(Sample.Prepared.Date), Sample.Prepared.Date := as.IDate("2022-12-20")]
    sinfo[is.na(Sample.Prepared.Time), Sample.Prepared.Time := as.ITime("06:39:03")]
  }

  # Compute hours between sample prep and sample measurement
  sinfo[, Prep.to.Measure.Duration := duration_hours(
    Sample.Prepared.Date, Sample.Prepared.Time,
    Sample.Measured.Date, Sample.Measured.Time
  )]

  # Get majority date of sample measurement per shipment plate
  plate_measure_dates <- sinfo[,.N,by=list(Shipment.Plate, Sample.Measured.Date)]
  plate_measure_dates <- plate_measure_dates[,.SD[which.max(N)],by=Shipment.Plate]
  sinfo[plate_measure_dates, on = list(Shipment.Plate), Plate.Measured.Date := i.Sample.Measured.Date]

  # For version 2 of the algorithm, we need to hard code a bin split. The easiest way
  # to do this is by creating a dummy spectrometer column to split on, by giving the
  # spectrometer before and after the split a different name
  sinfo[, Spectrometer.Group := Spectrometer]
  if (version > 1L && "0490000006726" %in% sinfo[["Shipment.Plate"]]) {
    sinfo <- sinfo[order(Plate.Measured.Date)][order(Spectrometer)]
    sinfo[, Spectrometer.Group := as.character(Spectrometer.Group)]
    spec_to_split <- sinfo[Shipment.Plate == "0490000006726", Spectrometer.Group][1]
    sinfo[, row := .I]
    idx_to_split <- sinfo[Shipment.Plate == "0490000006726", max(row)]
    sinfo[Spectrometer.Group == spec_to_split & row > idx_to_split,
          Spectrometer.Group := paste(Spectrometer.Group, "Group 2")]
    sinfo[, row := NULL]
  }

  # Bin samples into either 10 equal size bins per spectrometer (algorithm
  # version 1) or into bins of roughly 2000 samples each (algorithm version 2)
  sinfo[, Spectrometer.Date.Bin := bin_dates(Plate.Measured.Date, version), by=Spectrometer.Group]

  # Offset date bin by spectrometer number to make bin allocation unique across
  # all spectrometers
  offset <- sinfo[, list(max_bin=max(Spectrometer.Date.Bin)), by=Spectrometer.Group]
  offset <- offset[order(Spectrometer.Group)]
  offset[-1, offset := offset[, cumsum(max_bin)[-.N]]]
  offset[1, offset := 0]
  sinfo[offset, on = list(Spectrometer.Group), Spectrometer.Date.Bin := Spectrometer.Date.Bin + i.offset]

  # Melt to long, filtering to non-derived biomarkers and dropping missing values
  bio <- melt(bio, id.vars=c("eid", "visit_index"), variable.name="Biomarker", na.rm=TRUE,
    measure.vars=intersect(names(bio), ukbnmr::nmr_info[Type == "Non-derived", Biomarker]))

  # add in relevant sample processing information
  if (version == 1L) {
    bio <- sinfo[, list(eid, visit_index, Prep.to.Measure.Duration, Well.Row,
                        Well.Column, Spectrometer.Date.Bin, Spectrometer,
                        Spectrometer.Group, Shipment.Plate)][
                          bio, on = list(eid, visit_index), nomatch=0]
  } else {
    bio <- sinfo[, list(eid, visit_index, Prep.to.Measure.Duration, Well.Row,
                        Well.Column, Spectrometer.Date.Bin, Spectrometer,
                        Spectrometer.Group, Processing.Batch, Shipment.Plate)][
                          bio, on = list(eid, visit_index), nomatch=0]
  }

  message("Determining log offsets for biomarker concentrations...")

  # Determine offset for log transformation (required for variables with measurements == 0):
  # Acetate, Acetoacetate, Albumin, bOHbutyrate, Clinical_LDL_C, Gly, Ile, L_LDL_CE, L_LDL_FC, M_LDL_CE
  log_offset <- bio[!is.na(value), list(Minimum=min(value), Minimum.Non.Zero=min(value[value != 0])),by=Biomarker]
  log_offset[, Log.Offset := ifelse(Minimum == 0, Minimum.Non.Zero / 2, 0)]

  # Get log transformed raw value
  bio[log_offset, on = list(Biomarker), log_value := log(value + Log.Offset)]

  message("Adjusting for time between sample prep and sample measurement...")

  # Adjust for time between sample prep and sample measurement
  bio[, adj := MASS::rlm(log_value ~ log(Prep.to.Measure.Duration))$residuals, by=Biomarker]

  message("Adjusting for within plate structure across 96-well plate rows A-H...")

  # Adjust for within plate structure across 96-well plate rows A-H
  if (version == 1L) {
    bio[, adj := MASS::rlm(adj ~ factor_by_size(Well.Row))$residuals, by=Biomarker]
  } else {
    bio[, adj := MASS::rlm(adj ~ factor_by_size(Well.Row))$residuals, by=list(Processing.Batch, Biomarker)]
  }

  message("Adjusting for within plate structure across 96-well plate columns 1-12...")

  # Adjust for within plate structure across 96-well plate columns 1-12
  if (version == 1L) {
    bio[, adj := MASS::rlm(adj ~ factor_by_size(Well.Column))$residuals, by=Biomarker]
  } else {
    bio[, adj := MASS::rlm(adj ~ factor_by_size(Well.Column))$residuals, by=list(Processing.Batch, Biomarker)]
  }

  message("Adjusting for drift over time within spectrometer...")

  if (version == 3L) {
    # Adjust for drift over time within spectrometer (only applicable to spectrometers with > 1 bin)
    to_adjust <- sinfo[, list(N=length(unique(Spectrometer.Date.Bin))), by=list(Spectrometer)]
    to_adjust <- to_adjust[N > 1, Spectrometer]
    bio[Spectrometer %in% to_adjust, adj := MASS::rlm(adj ~ factor_by_size(Spectrometer.Date.Bin))$residuals, by=list(Biomarker, Spectrometer.Group)]
  } else {
    bio[, adj := MASS::rlm(adj ~ factor_by_size(Spectrometer.Date.Bin))$residuals, by=list(Biomarker, Spectrometer.Group)]
  }

  message("Rescaling adjusted biomarkers to absolute concentrations...")

  # Rescale to absolute units. First, shift the residuals to have the same central
  # parameter estimate (e.g. mean, estimated via robust linear regression) as the
  # log concentrations
  bio[, adj := adj + as.vector(coef(MASS::rlm(log_value ~ 1)))[1], by=Biomarker]

  # Undo the log transform
  bio[, adj := exp(adj)]

  # Remove the log offset
  bio[log_offset, on = list(Biomarker), adj := adj - Log.Offset]

  # Some values that were 0 are now < 0, apply small right shift
  # for these biomarkers (shift is very small, i.e. the impact on
  # the distribution's median is essentially numeric error)
  shift <- bio[, list(Right.Shift=-pmin(0, min(adj))), by=Biomarker]
  log_offset <- log_offset[shift, on = list(Biomarker)]
  bio[log_offset, on = list(Biomarker), adj := adj + Right.Shift]

  message("Identifying outlier plates and setting their concentrations to NA...")

  # Identify and remove outlier plates.
  # Model plate medians as a normal distribution (across all 1,352 plates), then
  # we can flag outlier plates as those > 3.374446 standard deviations from the
  # mean where 3.374446 is derived from the range of the theoretical normal
  # distribution for 1,352 samples
  n_plates <- sinfo[, length(unique(Shipment.Plate))] # 1,352
  sdlim <- max(qnorm(ppoints(n_plates))) # 3.374446
  plate_medians <- bio[, list(value = median(adj)), by=list(Biomarker, Shipment.Plate)]

  outlier_lim <- plate_medians[, list(
    Lower.Limit = mean(value) - sd(value) * sdlim,
    Mean.Plate.Medians = mean(value),
    Upper.Limit = mean(value) + sd(value) * sdlim
  ), by=Biomarker]

  plate_medians[, outlier := "no"]
  plate_medians[outlier_lim, on = list(Biomarker, value < Lower.Limit), outlier := "low"]
  plate_medians[outlier_lim, on = list(Biomarker, value > Upper.Limit), outlier := "high"]

  # Add outlier plate tags to biomarker qc tags
  if (!skip.biomarker.qc.flags) {
    message("Adding outlier plates to measurement QC tags...")

    bio_qc <- melt(bio_qc, id.vars=c("eid", "visit_index"), variable.name="Biomarker", na.rm=TRUE,
                   measure.vars=intersect(names(bio_qc), ukbnmr::nmr_info[Type == "Non-derived", Biomarker]))

    outlier_flags <- bio[plate_medians[outlier != "no"], on = list(Shipment.Plate, Biomarker),
                         list(eid, visit_index, Biomarker, value=ifelse(
                           outlier == "high", "High outlier plate", "Low outlier plate"))]

    bio_qc <- rbind(bio_qc, outlier_flags)
    bio_qc <- bio_qc[, list(value = paste(value, collapse="; ")), by=list(eid, visit_index, Biomarker)]
  }

  # Remove outlier plates from biomarker data
  if (remove.outlier.plates) {
    bio <- bio[!plate_medians[outlier != "no"], on = list(Biomarker, Shipment.Plate)]
  }

  # Convert back to wide format
  bio <- dcast(bio, eid + visit_index ~ Biomarker, value.var="adj")
  if (!skip.biomarker.qc.flags) {
    bio_qc <- dcast(bio_qc, eid + visit_index ~ Biomarker, value.var="value")
  }

  message("Recalculating derived biomarkers...")

  # Compute derived biomarkers and ratios
  bio <- nightingale_composite_biomarker_compute(bio)
  bio <- nightingale_ratio_compute(bio)
  bio <- extended_ratios_compute(bio)

  if (!skip.biomarker.qc.flags) {
    message("Collating measurement QC tags for derived biomarkers...")

    bio_qc <- nightingale_composite_biomarker_flags(bio_qc)
    bio_qc <- nightingale_ratio_flags(bio_qc)
    bio_qc <- extended_ratios_flags(bio_qc)

    # Add back in samples with no QC flags for any biomarkers
    bio_qc <- bio_qc[bio[, list(eid, visit_index)], on = list(eid, visit_index)]
  }

  # Remove dummy Spectrometer.Group column from returned output
  sinfo[, Spectrometer.Group := NULL]

  message("Returning result...")

  # Return list
  if (!skip.biomarker.qc.flags) {
    return(list(
      biomarkers=returnDT(bio),
      biomarker_qc_flags=returnDT(bio_qc),
      sample_processing=returnDT(sinfo),
      log_offset=returnDT(log_offset[Log.Offset != 0 | Right.Shift != 0]),
      outlier_plate_detection=returnDT(outlier_lim),
      algorithm_version=version
    ))
  } else {
    return(list(
      biomarkers=returnDT(bio),
      sample_processing=returnDT(sinfo),
      log_offset=returnDT(log_offset[Log.Offset != 0 | Right.Shift != 0]),
      outlier_plate_detection=returnDT(outlier_lim),
      algorithm_version=version
    ))
  }
}

# Cast object to data.table, or if already a data.table, make a copy so we don't
# modify by reference
getDT <- function(x) {
  if (is.data.table(x)) {
    copy(x)
  } else {
    as.data.table(x)
  }
}

# If a user is not using the data.table package (i.e. its not loaded by the user
# in their R session) make sure to return a data.frame instead.
returnDT <- function(x) {
  if (is.data.table(x) && !("data.table" %in% .packages())) {
    setDF(x)
  } else {
    x[,] # Otherwise the first show() call will be empty
  }
  return(x)
}

# Is a vector an integer? It may have been loaded as character / numeric
is.integer <- function(x) {
  if (is.factor(x)) {
    return(FALSE)
  }
  tryCatch({
    as.integer(x)
    return(TRUE)
  }, warning=function(w) {
    return(FALSE)
  })
}

# Paste together QC flags for biomarkers when computing ratios
collate_flags <- function(...) {
  colnames <- sapply(substitute(list(...))[-1], deparse)
  cols <- list(...)

  # Throw appropriate error if column does not exist
  for (ii in seq_along(colnames)) {
    if (is.null(cols[[ii]])) {
      stop(sprintf("object '%s' not found", colnames[ii]))
    }
  }

  # If any of the input biomarkers have QC flags, aggregate, and add the respective
  # biomarker name. If no tags, return NA. If any column is itself an aggregated
  # set of flags (e.g. a ratio of Total_C / Total_L will be aggregated from the
  # sums of all cholesterols and lipids), don't add the biomarker name.
  tags_with_names <- sapply(seq_along(colnames), function(argIdx) {
    ifelse(is.na(cols[[argIdx]]), NA_character_,
      ifelse(grepl(":", cols[[argIdx]]), gsub(".$", "", cols[[argIdx]]),
        sprintf("%s: %s", colnames[argIdx], cols[[argIdx]])))
  })
  # For each row, get a single string of collated flags.
  collated_tags <- apply(tags_with_names, 1, function(row) {
    ifelse(all(is.na(row)), NA_character_, paste0(paste(stats::na.omit(row), collapse=". "), "."))
  })
  # In some cases, biomarkers may appear multiple times where they contribute to
  # e.g. both a numerator and denominator of a ratio. Split out the string and
  # make sure we only return each underlying biomarker's flags once.
  sapply(strsplit(collated_tags, split="\\. ?"), function(row) {
    ifelse(all(is.na(row)), NA_character_, paste0(paste(sort(unique(row)), collapse=". "), "."))
  })
}

# Convert an ITime object to numeric representation of hours since midnight
hours_decimal <- function(time) {
  hour(time) + minute(time)/60 + second(time)/3600
}

# Compute durations between events (hours)
duration_hours <- function(days1, time1, days2, time2) {
  (as.double(days2 - days1))*24 + (hours_decimal(time2) - hours_decimal(time1))
}

# Function for converting a set of numbers to a factor, ordering levels by group size
factor_by_size <- function(x) {
  # Suppress CRAN warnings for data.table columns
  N <- NULL

  group_sizes <- as.data.table(table(x))
  group_sizes <- group_sizes[order(-N)]
  factor(x, levels=group_sizes$x)
}

# Split a series of dates into equal size bins, ordered by date
bin_dates <- function(date, version) {
  # Suppress CRAN warnings for data.table columns
  bin <- NULL

  date_as_int <- as.integer(date)
  date_order <- as.integer(factor(date_as_int))
  if (version == 1L) {
    bin_size <- max(date_order)/10
    return(as.integer(ceiling(date_order/bin_size)))
  } else if (version == 2L) {
    n_bins <- floor(length(date)/2000)
    if (n_bins < 2) n_bins <- 2 # test_data not big enough to bin into groups of 2,000 samples, so just bin into 2
    bins <- cut(unique(date_order), n_bins, labels=FALSE)
    bin_map <- data.table(date=unique(date_order), bin=bins)
    return(bin_map[data.table(date=date_order), on=list(date), bin])
  } else if (version == 3L) {
    n_bins <- floor(length(date)/2000)
    if (n_bins < 2) {
      return(rep(1, length(date)))
    } else {
      bins <- cut(unique(date_order), n_bins, labels=FALSE)
      bin_map <- data.table(date=unique(date_order), bin=bins)
      return(bin_map[data.table(date=date_order), on=list(date), bin])
    }
  }
}

# Harmonize various date time formats so they can be passed to IDateTime
harmonize_datetime <- function(x) {
  y <- suppressWarnings(ymd_hms(x))
  y[is.na(y)] <- suppressWarnings(ymd(x[is.na(y)]))
  y
}

#' Tools for processing the UK Biobank NMR metabolomics biomarker data
#'
#'  @description
#' This package provides utilities for working with the
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{UK Biobank MR metabolomics data}.
#' Details are provided below, and in the package vignette (type \code{vignette("ukbnmr")} to view).
#'
#' @details
#' There are three groups of functions in this package: (1) data extraction,
#' (2) removal of technical variation, and (3) recomputing derived biomarkers
#' and biomarker ratios.
#'
#' All functions can be applied directly to raw data extracted from UK Biobank.
#'
#' This package also provides a \code{data.frame} of biomarker information, loaded
#' as \code{\link{nmr_info}}, and \code{data.frame} of sample processing information,
#' loaded as \code{\link{sample_qc_info}}.
#'
#' @section Data Extraction Functions:
#' The \code{\link{extract_biomarkers}()} function will take a phenotype dataset extracted on
#' the UK Biobank Research Analysis Platform by the Table Exporter tool, extract the
#' \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{NMR biomarker fields}
#' and give them short comprehensible column names as described in \code{\link{nmr_info}}.
#' Measurements are also split into multiple rows where a participant has measurements
#' at both baseline and first repeat assessment.
#'
#' The \code{\link{extract_biomarker_qc_flags}()} function will take a phenotype
#' dataset extracted on the UK Biobank Research Analysis Platform by the Table
#' Exporter tool, extract the \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{Nightingale quality control flags}
#' for each biomarker measurement, returning a single column per biomarker
#' (corresponding to respective columns output by \code{\link{extract_biomarkers}()}).
#'
#' The \code{\link{extract_sample_qc_flags}()} function will take a phenotype
#' dataset extracted on the UK Biobank Research Analysis Platform by the Table
#' Exporter tool and extract the \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{sample quality control tags}
#' for the Nightingale NMR metabolomics data.
#'
#' These functions will also work with older datasets predating the UK Biobank
#' Research Analysis Platform, e.g. those extracted by \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv},
#' and/or by the ukbtools R package.
#'
#' @section Removal of technical variation:
#' The \code{\link{remove_technical_variation}()} function  will take a phenotype
#' dataset extracted on the UK Biobank Research Analysis Platform by the Table
#' Exporter tool, extract all the biomarkers and QC flags, remove the effects of
#' technical variation on biomarker concentrations, and return a list
#' containing the adjusted NMR biomarker data, biomarker QC flags, and sample quality control
#' and processing information.
#'
#' This applies a multistep process as described in Ritchie \emph{et al.} 2023:
#'
#' \enumerate{
#'   \item{First biomarker data is filtered to the 107 biomarkers that
#'   cannot be derived from any combination of other biomarkers.}
#'   \item{Absolute concentrations are log transformed, with a small offset
#'   applied to biomarkers with concentrations of 0.}
#'   \item{Each biomarker is adjusted for the time between sample preparation
#'   and sample measurement (hours) on a log scale.}
#'   \item{Each biomarker is adjusted for systematic differences between rows
#'   (A-H) on the 96-well shipment plates.}
#'   \item{Each biomarker is adjusted for remaining systematic differences
#'   between columns (1-12) on the 96-well shipment plates.}
#'   \item{Each biomarker is adjusted for drift over time within each of the six
#'   spectrometers. To do so, samples are grouped into 10 bins, within each
#'   spectrometer, by the date the majority of samples on their respective
#'   96-well plates were measured.}
#'   \item{Regression residuals after the sequential adjustments are
#'   transformed back to absolute concentrations.}
#'   \item{Samples belonging to shipment plates that are outliers of
#'   non-biological origin are identified and set to missing.}
#'   \item{The 61 composite biomarkers and 81 biomarker ratios are recomputed
#'   from their adjusted parts.}
#'   \item{An additional 76 biomarker ratios of potential biological
#'   significance are computed.}
#' }
#'
#' Further details can be found in Ritchie S. C. \emph{et al.} Quality control
#' and removal of technical variation of NMR metabolic biomarker data in
#' ~120,000 UK Biobank participants, \emph{Sci Data} \strong{10}, 64 (2023). doi:
#' \href{https://www.nature.com/articles/s41597-023-01949-y}{10.1038/s41597-023-01949-y}
#'
#' @section  Methods for computing biomarker ratios:
#' The \code{\link{compute_extended_ratios}()} function will compute an extended
#' set of biomarker ratios expanding on the biomarkers available directly from
#' the Nightingale platform. A companion function, \code{\link{compute_extended_ratio_qc_flags}()},
#' will aggregate the QC flags for the biomarkers underlying each ratio.
#'
#' The \code{\link{recompute_derived_biomarkers}()} function will recompute all
#' composite biomarkers and ratios from 107 non-derived biomarkers, which is
#' useful for ensuring data consistency when adjusting for unwanted biological
#' variation. This includes the extended biomarker rations computed by the
#' \code{\link{compute_extended_ratios}()} function. A companion function,
#' \code{\link{recompute_derived_biomarker_qc_flags}()} will aggregate the QC
#' flags for the biomarkers underlying each composite biomarker and ratio.
#'
#' @name ukbnmr
#' @import data.table
#' @importFrom stats na.omit
#' @keywords package
"_PACKAGE"

# Internal functions for aggregating QC Flags when rederiving biomarkers

nightingale_ratio_flags <- function(x) {
  # Silence CRAN NOTES about undefined global variables (columns in the scope of x)
  ApoA1 <- ApoB <- ApoB_by_ApoA1 <- DHA <- DHA_pct <- IDL_C <- IDL_C_pct <-
    IDL_CE <- IDL_CE_pct <- IDL_FC <- IDL_FC_pct <- IDL_L <- IDL_PL <- IDL_PL_pct <-
    IDL_TG <- IDL_TG_pct <- L_HDL_C <- L_HDL_C_pct <- L_HDL_CE <- L_HDL_CE_pct <-
    L_HDL_FC <- L_HDL_FC_pct <- L_HDL_L <- L_HDL_PL <- L_HDL_PL_pct <- L_HDL_TG <-
    L_HDL_TG_pct <- L_LDL_C <- L_LDL_C_pct <- L_LDL_CE <- L_LDL_CE_pct <- L_LDL_FC <-
    L_LDL_FC_pct <- L_LDL_L <- L_LDL_PL <- L_LDL_PL_pct <- L_LDL_TG <- L_LDL_TG_pct <-
    L_VLDL_C <- L_VLDL_C_pct <- L_VLDL_CE <- L_VLDL_CE_pct <- L_VLDL_FC <- L_VLDL_FC_pct <-
    L_VLDL_L <- L_VLDL_PL <- L_VLDL_PL_pct <- L_VLDL_TG <- L_VLDL_TG_pct <- LA <-
    LA_pct <- M_HDL_C <- M_HDL_C_pct <- M_HDL_CE <- M_HDL_CE_pct <- M_HDL_FC <-
    M_HDL_FC_pct <- M_HDL_L <- M_HDL_PL <- M_HDL_PL_pct <- M_HDL_TG <- M_HDL_TG_pct <-
    M_LDL_C <- M_LDL_C_pct <- M_LDL_CE <- M_LDL_CE_pct <- M_LDL_FC <- M_LDL_FC_pct <-
    M_LDL_L <- M_LDL_PL <- M_LDL_PL_pct <- M_LDL_TG <- M_LDL_TG_pct <- M_VLDL_C <-
    M_VLDL_C_pct <- M_VLDL_CE <- M_VLDL_CE_pct <- M_VLDL_FC <- M_VLDL_FC_pct <-
    M_VLDL_L <- M_VLDL_PL <- M_VLDL_PL_pct <- M_VLDL_TG <- M_VLDL_TG_pct <- MUFA <-
    MUFA_pct <- Omega_3 <- Omega_3_pct <- Omega_6 <- Omega_6_by_Omega_3 <- Omega_6_pct <-
    Phosphoglyc <- PUFA <- PUFA_by_MUFA <- PUFA_pct <- S_HDL_C <- S_HDL_C_pct <-
    S_HDL_CE <- S_HDL_CE_pct <- S_HDL_FC <- S_HDL_FC_pct <- S_HDL_L <- S_HDL_PL <-
    S_HDL_PL_pct <- S_HDL_TG <- S_HDL_TG_pct <- S_LDL_C <- S_LDL_C_pct <- S_LDL_CE <-
    S_LDL_CE_pct <- S_LDL_FC <- S_LDL_FC_pct <- S_LDL_L <- S_LDL_PL <- S_LDL_PL_pct <-
    S_LDL_TG <- S_LDL_TG_pct <- S_VLDL_C <- S_VLDL_C_pct <- S_VLDL_CE <- S_VLDL_CE_pct <-
    S_VLDL_FC <- S_VLDL_FC_pct <- S_VLDL_L <- S_VLDL_PL <- S_VLDL_PL_pct <- S_VLDL_TG <-
    S_VLDL_TG_pct <- SFA <- SFA_pct <- TG_by_PG <- Total_FA <- Total_TG <- XL_HDL_C <-
    XL_HDL_C_pct <- XL_HDL_CE <- XL_HDL_CE_pct <- XL_HDL_FC <- XL_HDL_FC_pct <-
    XL_HDL_L <- XL_HDL_PL <- XL_HDL_PL_pct <- XL_HDL_TG <- XL_HDL_TG_pct <- XL_VLDL_C <-
    XL_VLDL_C_pct <- XL_VLDL_CE <- XL_VLDL_CE_pct <- XL_VLDL_FC <- XL_VLDL_FC_pct <-
    XL_VLDL_L <- XL_VLDL_PL <- XL_VLDL_PL_pct <- XL_VLDL_TG <- XL_VLDL_TG_pct <-
    XS_VLDL_C <- XS_VLDL_C_pct <- XS_VLDL_CE <- XS_VLDL_CE_pct <- XS_VLDL_FC <-
    XS_VLDL_FC_pct <- XS_VLDL_L <- XS_VLDL_PL <- XS_VLDL_PL_pct <- XS_VLDL_TG <-
    XS_VLDL_TG_pct <- XXL_VLDL_C <- XXL_VLDL_C_pct <- XXL_VLDL_CE <- XXL_VLDL_CE_pct <-
    XXL_VLDL_FC <- XXL_VLDL_FC_pct <- XXL_VLDL_L <- XXL_VLDL_PL <- XXL_VLDL_PL_pct <-
    XXL_VLDL_TG <- XXL_VLDL_TG_pct <- NULL



  # For each of the 14 lipoprotein subclasses, compute percentage of total
  # lipids composed of cholesteryl esters (CE), free cholesterol (FC), total
  # cholester (C), phospholipids (PL), and triglycerides (TG).
  if (exists("XXL_VLDL_CE", x) && exists("XXL_VLDL_L", x)) {
    x[, XXL_VLDL_CE_pct := collate_flags(XXL_VLDL_CE, XXL_VLDL_L)]
  } else if (!(exists("XXL_VLDL_CE", x) && exists("XXL_VLDL_L", x)) && exists("XXL_VLDL_CE_pct", x)) {
    x[, XXL_VLDL_CE_pct := NULL]
  }

  if (exists("XXL_VLDL_FC", x) && exists("XXL_VLDL_L", x)) {
    x[, XXL_VLDL_FC_pct := collate_flags(XXL_VLDL_FC, XXL_VLDL_L)]
  } else if (!(exists("XXL_VLDL_FC", x) && exists("XXL_VLDL_L", x)) && exists("XXL_VLDL_FC_pct", x)) {
    x[, XXL_VLDL_FC_pct := NULL]
  }

  if (exists("XXL_VLDL_C", x) && exists("XXL_VLDL_L", x)) {
    x[, XXL_VLDL_C_pct := collate_flags(XXL_VLDL_C, XXL_VLDL_L)]
  } else if (!(exists("XXL_VLDL_C", x) && exists("XXL_VLDL_L", x)) && exists("XXL_VLDL_C_pct", x)) {
    x[, XXL_VLDL_C_pct := NULL]
  }

  if (exists("XXL_VLDL_PL", x) && exists("XXL_VLDL_L", x)) {
    x[, XXL_VLDL_PL_pct := collate_flags(XXL_VLDL_PL, XXL_VLDL_L)]
  } else if (!(exists("XXL_VLDL_PL", x) && exists("XXL_VLDL_L", x)) && exists("XXL_VLDL_PL_pct", x)) {
    x[, XXL_VLDL_PL_pct := NULL]
  }

  if (exists("XXL_VLDL_TG", x) && exists("XXL_VLDL_L", x)) {
    x[, XXL_VLDL_TG_pct := collate_flags(XXL_VLDL_TG, XXL_VLDL_L)]
  } else if (!(exists("XXL_VLDL_TG", x) && exists("XXL_VLDL_L", x)) && exists("XXL_VLDL_TG_pct", x)) {
    x[, XXL_VLDL_TG_pct := NULL]
  }


  if (exists("XL_VLDL_CE", x) && exists("XL_VLDL_L", x)) {
    x[, XL_VLDL_CE_pct := collate_flags(XL_VLDL_CE, XL_VLDL_L)]
  } else if (!(exists("XL_VLDL_CE", x) && exists("XL_VLDL_L", x)) && exists("XL_VLDL_CE_pct", x)) {
    x[, XL_VLDL_CE_pct := NULL]
  }

  if (exists("XL_VLDL_FC", x) && exists("XL_VLDL_L", x)) {
    x[, XL_VLDL_FC_pct := collate_flags(XL_VLDL_FC, XL_VLDL_L)]
  } else if (!(exists("XL_VLDL_FC", x) && exists("XL_VLDL_L", x)) && exists("XL_VLDL_FC_pct", x)) {
    x[, XL_VLDL_FC_pct := NULL]
  }

  if (exists("XL_VLDL_C", x) && exists("XL_VLDL_L", x)) {
    x[, XL_VLDL_C_pct := collate_flags(XL_VLDL_C, XL_VLDL_L)]
  } else if (!(exists("XL_VLDL_C", x) && exists("XL_VLDL_L", x)) && exists("XL_VLDL_C_pct", x)) {
    x[, XL_VLDL_C_pct := NULL]
  }

  if (exists("XL_VLDL_PL", x) && exists("XL_VLDL_L", x)) {
    x[, XL_VLDL_PL_pct := collate_flags(XL_VLDL_PL, XL_VLDL_L)]
  } else if (!(exists("XL_VLDL_PL", x) && exists("XL_VLDL_L", x)) && exists("XL_VLDL_PL_pct", x)) {
    x[, XL_VLDL_PL_pct := NULL]
  }

  if (exists("XL_VLDL_TG", x) && exists("XL_VLDL_L", x)) {
    x[, XL_VLDL_TG_pct := collate_flags(XL_VLDL_TG, XL_VLDL_L)]
  } else if (!(exists("XL_VLDL_TG", x) && exists("XL_VLDL_L", x)) && exists("XL_VLDL_TG_pct", x)) {
    x[, XL_VLDL_TG_pct := NULL]
  }


  if (exists("L_VLDL_CE", x) && exists("L_VLDL_L", x)) {
    x[, L_VLDL_CE_pct := collate_flags(L_VLDL_CE, L_VLDL_L)]
  } else if (!(exists("L_VLDL_CE", x) && exists("L_VLDL_L", x)) && exists("L_VLDL_CE_pct", x)) {
    x[, L_VLDL_CE_pct := NULL]
  }

  if (exists("L_VLDL_FC", x) && exists("L_VLDL_L", x)) {
    x[, L_VLDL_FC_pct := collate_flags(L_VLDL_FC, L_VLDL_L)]
  } else if (!(exists("L_VLDL_FC", x) && exists("L_VLDL_L", x)) && exists("L_VLDL_FC_pct", x)) {
    x[, L_VLDL_FC_pct := NULL]
  }

  if (exists("L_VLDL_C", x) && exists("L_VLDL_L", x)) {
    x[, L_VLDL_C_pct := collate_flags(L_VLDL_C, L_VLDL_L)]
  } else if (!(exists("L_VLDL_C", x) && exists("L_VLDL_L", x)) && exists("L_VLDL_C_pct", x)) {
    x[, L_VLDL_C_pct := NULL]
  }

  if (exists("L_VLDL_PL", x) && exists("L_VLDL_L", x)) {
    x[, L_VLDL_PL_pct := collate_flags(L_VLDL_PL, L_VLDL_L)]
  } else if (!(exists("L_VLDL_PL", x) && exists("L_VLDL_L", x)) && exists("L_VLDL_PL_pct", x)) {
    x[, L_VLDL_PL_pct := NULL]
  }

  if (exists("L_VLDL_TG", x) && exists("L_VLDL_L", x)) {
    x[, L_VLDL_TG_pct := collate_flags(L_VLDL_TG, L_VLDL_L)]
  } else if (!(exists("L_VLDL_TG", x) && exists("L_VLDL_L", x)) && exists("L_VLDL_TG_pct", x)) {
    x[, L_VLDL_TG_pct := NULL]
  }


  if (exists("M_VLDL_CE", x) && exists("M_VLDL_L", x)) {
    x[, M_VLDL_CE_pct := collate_flags(M_VLDL_CE, M_VLDL_L)]
  } else if (!(exists("M_VLDL_CE", x) && exists("M_VLDL_L", x)) && exists("M_VLDL_CE_pct", x)) {
    x[, M_VLDL_CE_pct := NULL]
  }

  if (exists("M_VLDL_FC", x) && exists("M_VLDL_L", x)) {
    x[, M_VLDL_FC_pct := collate_flags(M_VLDL_FC, M_VLDL_L)]
  } else if (!(exists("M_VLDL_FC", x) && exists("M_VLDL_L", x)) && exists("M_VLDL_FC_pct", x)) {
    x[, M_VLDL_FC_pct := NULL]
  }

  if (exists("M_VLDL_C", x) && exists("M_VLDL_L", x)) {
    x[, M_VLDL_C_pct := collate_flags(M_VLDL_C, M_VLDL_L)]
  } else if (!(exists("M_VLDL_C", x) && exists("M_VLDL_L", x)) && exists("M_VLDL_C_pct", x)) {
    x[, M_VLDL_C_pct := NULL]
  }

  if (exists("M_VLDL_PL", x) && exists("M_VLDL_L", x)) {
    x[, M_VLDL_PL_pct := collate_flags(M_VLDL_PL, M_VLDL_L)]
  } else if (!(exists("M_VLDL_PL", x) && exists("M_VLDL_L", x)) && exists("M_VLDL_PL_pct", x)) {
    x[, M_VLDL_PL_pct := NULL]
  }

  if (exists("M_VLDL_TG", x) && exists("M_VLDL_L", x)) {
    x[, M_VLDL_TG_pct := collate_flags(M_VLDL_TG, M_VLDL_L)]
  } else if (!(exists("M_VLDL_TG", x) && exists("M_VLDL_L", x)) && exists("M_VLDL_TG_pct", x)) {
    x[, M_VLDL_TG_pct := NULL]
  }


  if (exists("S_VLDL_CE", x) && exists("S_VLDL_L", x)) {
    x[, S_VLDL_CE_pct := collate_flags(S_VLDL_CE, S_VLDL_L)]
  } else if (!(exists("S_VLDL_CE", x) && exists("S_VLDL_L", x)) && exists("S_VLDL_CE_pct", x)) {
    x[, S_VLDL_CE_pct := NULL]
  }

  if (exists("S_VLDL_FC", x) && exists("S_VLDL_L", x)) {
    x[, S_VLDL_FC_pct := collate_flags(S_VLDL_FC, S_VLDL_L)]
  } else if (!(exists("S_VLDL_FC", x) && exists("S_VLDL_L", x)) && exists("S_VLDL_FC_pct", x)) {
    x[, S_VLDL_FC_pct := NULL]
  }

  if (exists("S_VLDL_C", x) && exists("S_VLDL_L", x)) {
    x[, S_VLDL_C_pct := collate_flags(S_VLDL_C, S_VLDL_L)]
  } else if (!(exists("S_VLDL_C", x) && exists("S_VLDL_L", x)) && exists("S_VLDL_C_pct", x)) {
    x[, S_VLDL_C_pct := NULL]
  }

  if (exists("S_VLDL_PL", x) && exists("S_VLDL_L", x)) {
    x[, S_VLDL_PL_pct := collate_flags(S_VLDL_PL, S_VLDL_L)]
  } else if (!(exists("S_VLDL_PL", x) && exists("S_VLDL_L", x)) && exists("S_VLDL_PL_pct", x)) {
    x[, S_VLDL_PL_pct := NULL]
  }

  if (exists("S_VLDL_TG", x) && exists("S_VLDL_L", x)) {
    x[, S_VLDL_TG_pct := collate_flags(S_VLDL_TG, S_VLDL_L)]
  } else if (!(exists("S_VLDL_TG", x) && exists("S_VLDL_L", x)) && exists("S_VLDL_TG_pct", x)) {
    x[, S_VLDL_TG_pct := NULL]
  }


  if (exists("XS_VLDL_CE", x) && exists("XS_VLDL_L", x)) {
    x[, XS_VLDL_CE_pct := collate_flags(XS_VLDL_CE, XS_VLDL_L)]
  } else if (!(exists("XS_VLDL_CE", x) && exists("XS_VLDL_L", x)) && exists("XS_VLDL_CE_pct", x)) {
    x[, XS_VLDL_CE_pct := NULL]
  }

  if (exists("XS_VLDL_FC", x) && exists("XS_VLDL_L", x)) {
    x[, XS_VLDL_FC_pct := collate_flags(XS_VLDL_FC, XS_VLDL_L)]
  } else if (!(exists("XS_VLDL_FC", x) && exists("XS_VLDL_L", x)) && exists("XS_VLDL_FC_pct", x)) {
    x[, XS_VLDL_FC_pct := NULL]
  }

  if (exists("XS_VLDL_C", x) && exists("XS_VLDL_L", x)) {
    x[, XS_VLDL_C_pct := collate_flags(XS_VLDL_C, XS_VLDL_L)]
  } else if (!(exists("XS_VLDL_C", x) && exists("XS_VLDL_L", x)) && exists("XS_VLDL_C_pct", x)) {
    x[, XS_VLDL_C_pct := NULL]
  }

  if (exists("XS_VLDL_PL", x) && exists("XS_VLDL_L", x)) {
    x[, XS_VLDL_PL_pct := collate_flags(XS_VLDL_PL, XS_VLDL_L)]
  } else if (!(exists("XS_VLDL_PL", x) && exists("XS_VLDL_L", x)) && exists("XS_VLDL_PL_pct", x)) {
    x[, XS_VLDL_PL_pct := NULL]
  }

  if (exists("XS_VLDL_TG", x) && exists("XS_VLDL_L", x)) {
    x[, XS_VLDL_TG_pct := collate_flags(XS_VLDL_TG, XS_VLDL_L)]
  } else if (!(exists("XS_VLDL_TG", x) && exists("XS_VLDL_L", x)) && exists("XS_VLDL_TG_pct", x)) {
    x[, XS_VLDL_TG_pct := NULL]
  }


  if (exists("L_LDL_CE", x) && exists("L_LDL_L", x)) {
    x[, L_LDL_CE_pct := collate_flags(L_LDL_CE, L_LDL_L)]
  } else if (!(exists("L_LDL_CE", x) && exists("L_LDL_L", x)) && exists("L_LDL_CE_pct", x)) {
    x[, L_LDL_CE_pct := NULL]
  }

  if (exists("L_LDL_FC", x) && exists("L_LDL_L", x)) {
    x[, L_LDL_FC_pct := collate_flags(L_LDL_FC, L_LDL_L)]
  } else if (!(exists("L_LDL_FC", x) && exists("L_LDL_L", x)) && exists("L_LDL_FC_pct", x)) {
    x[, L_LDL_FC_pct := NULL]
  }

  if (exists("L_LDL_C", x) && exists("L_LDL_L", x)) {
    x[, L_LDL_C_pct := collate_flags(L_LDL_C, L_LDL_L)]
  } else if (!(exists("L_LDL_C", x) && exists("L_LDL_L", x)) && exists("L_LDL_C_pct", x)) {
    x[, L_LDL_C_pct := NULL]
  }

  if (exists("L_LDL_PL", x) && exists("L_LDL_L", x)) {
    x[, L_LDL_PL_pct := collate_flags(L_LDL_PL, L_LDL_L)]
  } else if (!(exists("L_LDL_PL", x) && exists("L_LDL_L", x)) && exists("L_LDL_PL_pct", x)) {
    x[, L_LDL_PL_pct := NULL]
  }

  if (exists("L_LDL_TG", x) && exists("L_LDL_L", x)) {
    x[, L_LDL_TG_pct := collate_flags(L_LDL_TG, L_LDL_L)]
  } else if (!(exists("L_LDL_TG", x) && exists("L_LDL_L", x)) && exists("L_LDL_TG_pct", x)) {
    x[, L_LDL_TG_pct := NULL]
  }


  if (exists("M_LDL_CE", x) && exists("M_LDL_L", x)) {
    x[, M_LDL_CE_pct := collate_flags(M_LDL_CE, M_LDL_L)]
  } else if (!(exists("M_LDL_CE", x) && exists("M_LDL_L", x)) && exists("M_LDL_CE_pct", x)) {
    x[, M_LDL_CE_pct := NULL]
  }

  if (exists("M_LDL_FC", x) && exists("M_LDL_L", x)) {
    x[, M_LDL_FC_pct := collate_flags(M_LDL_FC, M_LDL_L)]
  } else if (!(exists("M_LDL_FC", x) && exists("M_LDL_L", x)) && exists("M_LDL_FC_pct", x)) {
    x[, M_LDL_FC_pct := NULL]
  }

  if (exists("M_LDL_C", x) && exists("M_LDL_L", x)) {
    x[, M_LDL_C_pct := collate_flags(M_LDL_C, M_LDL_L)]
  } else if (!(exists("M_LDL_C", x) && exists("M_LDL_L", x)) && exists("M_LDL_C_pct", x)) {
    x[, M_LDL_C_pct := NULL]
  }

  if (exists("M_LDL_PL", x) && exists("M_LDL_L", x)) {
    x[, M_LDL_PL_pct := collate_flags(M_LDL_PL, M_LDL_L)]
  } else if (!(exists("M_LDL_PL", x) && exists("M_LDL_L", x)) && exists("M_LDL_PL_pct", x)) {
    x[, M_LDL_PL_pct := NULL]
  }

  if (exists("M_LDL_TG", x) && exists("M_LDL_L", x)) {
    x[, M_LDL_TG_pct := collate_flags(M_LDL_TG, M_LDL_L)]
  } else if (!(exists("M_LDL_TG", x) && exists("M_LDL_L", x)) && exists("M_LDL_TG_pct", x)) {
    x[, M_LDL_TG_pct := NULL]
  }


  if (exists("S_LDL_CE", x) && exists("S_LDL_L", x)) {
    x[, S_LDL_CE_pct := collate_flags(S_LDL_CE, S_LDL_L)]
  } else if (!(exists("S_LDL_CE", x) && exists("S_LDL_L", x)) && exists("S_LDL_CE_pct", x)) {
    x[, S_LDL_CE_pct := NULL]
  }

  if (exists("S_LDL_FC", x) && exists("S_LDL_L", x)) {
    x[, S_LDL_FC_pct := collate_flags(S_LDL_FC, S_LDL_L)]
  } else if (!(exists("S_LDL_FC", x) && exists("S_LDL_L", x)) && exists("S_LDL_FC_pct", x)) {
    x[, S_LDL_FC_pct := NULL]
  }

  if (exists("S_LDL_C", x) && exists("S_LDL_L", x)) {
    x[, S_LDL_C_pct := collate_flags(S_LDL_C, S_LDL_L)]
  } else if (!(exists("S_LDL_C", x) && exists("S_LDL_L", x)) && exists("S_LDL_C_pct", x)) {
    x[, S_LDL_C_pct := NULL]
  }

  if (exists("S_LDL_PL", x) && exists("S_LDL_L", x)) {
    x[, S_LDL_PL_pct := collate_flags(S_LDL_PL, S_LDL_L)]
  } else if (!(exists("S_LDL_PL", x) && exists("S_LDL_L", x)) && exists("S_LDL_PL_pct", x)) {
    x[, S_LDL_PL_pct := NULL]
  }

  if (exists("S_LDL_TG", x) && exists("S_LDL_L", x)) {
    x[, S_LDL_TG_pct := collate_flags(S_LDL_TG, S_LDL_L)]
  } else if (!(exists("S_LDL_TG", x) && exists("S_LDL_L", x)) && exists("S_LDL_TG_pct", x)) {
    x[, S_LDL_TG_pct := NULL]
  }


  if (exists("IDL_CE", x) && exists("IDL_L", x)) {
    x[, IDL_CE_pct := collate_flags(IDL_CE, IDL_L)]
  } else if (!(exists("IDL_CE", x) && exists("IDL_L", x)) && exists("IDL_CE_pct", x)) {
    x[, IDL_CE_pct := NULL]
  }

  if (exists("IDL_FC", x) && exists("IDL_L", x)) {
    x[, IDL_FC_pct := collate_flags(IDL_FC, IDL_L)]
  } else if (!(exists("IDL_FC", x) && exists("IDL_L", x)) && exists("IDL_FC_pct", x)) {
    x[, IDL_FC_pct := NULL]
  }

  if (exists("IDL_C", x) && exists("IDL_L", x)) {
    x[, IDL_C_pct := collate_flags(IDL_C, IDL_L)]
  } else if (!(exists("IDL_C", x) && exists("IDL_L", x)) && exists("IDL_C_pct", x)) {
    x[, IDL_C_pct := NULL]
  }

  if (exists("IDL_PL", x) && exists("IDL_L", x)) {
    x[, IDL_PL_pct := collate_flags(IDL_PL, IDL_L)]
  } else if (!(exists("IDL_PL", x) && exists("IDL_L", x)) && exists("IDL_PL_pct", x)) {
    x[, IDL_PL_pct := NULL]
  }

  if (exists("IDL_TG", x) && exists("IDL_L", x)) {
    x[, IDL_TG_pct := collate_flags(IDL_TG, IDL_L)]
  } else if (!(exists("IDL_TG", x) && exists("IDL_L", x)) && exists("IDL_TG_pct", x)) {
    x[, IDL_TG_pct := NULL]
  }


  if (exists("XL_HDL_CE", x) && exists("XL_HDL_L", x)) {
    x[, XL_HDL_CE_pct := collate_flags(XL_HDL_CE, XL_HDL_L)]
  } else if (!(exists("XL_HDL_CE", x) && exists("XL_HDL_L", x)) && exists("XL_HDL_CE_pct", x)) {
    x[, XL_HDL_CE_pct := NULL]
  }

  if (exists("XL_HDL_FC", x) && exists("XL_HDL_L", x)) {
    x[, XL_HDL_FC_pct := collate_flags(XL_HDL_FC, XL_HDL_L)]
  } else if (!(exists("XL_HDL_FC", x) && exists("XL_HDL_L", x)) && exists("XL_HDL_FC_pct", x)) {
    x[, XL_HDL_FC_pct := NULL]
  }

  if (exists("XL_HDL_C", x) && exists("XL_HDL_L", x)) {
    x[, XL_HDL_C_pct := collate_flags(XL_HDL_C, XL_HDL_L)]
  } else if (!(exists("XL_HDL_C", x) && exists("XL_HDL_L", x)) && exists("XL_HDL_C_pct", x)) {
    x[, XL_HDL_C_pct := NULL]
  }

  if (exists("XL_HDL_PL", x) && exists("XL_HDL_L", x)) {
    x[, XL_HDL_PL_pct := collate_flags(XL_HDL_PL, XL_HDL_L)]
  } else if (!(exists("XL_HDL_PL", x) && exists("XL_HDL_L", x)) && exists("XL_HDL_PL_pct", x)) {
    x[, XL_HDL_PL_pct := NULL]
  }

  if (exists("XL_HDL_TG", x) && exists("XL_HDL_L", x)) {
    x[, XL_HDL_TG_pct := collate_flags(XL_HDL_TG, XL_HDL_L)]
  } else if (!(exists("XL_HDL_TG", x) && exists("XL_HDL_L", x)) && exists("XL_HDL_TG_pct", x)) {
    x[, XL_HDL_TG_pct := NULL]
  }


  if (exists("L_HDL_CE", x) && exists("L_HDL_L", x)) {
    x[, L_HDL_CE_pct := collate_flags(L_HDL_CE, L_HDL_L)]
  } else if (!(exists("L_HDL_CE", x) && exists("L_HDL_L", x)) && exists("L_HDL_CE_pct", x)) {
    x[, L_HDL_CE_pct := NULL]
  }

  if (exists("L_HDL_FC", x) && exists("L_HDL_L", x)) {
    x[, L_HDL_FC_pct := collate_flags(L_HDL_FC, L_HDL_L)]
  } else if (!(exists("L_HDL_FC", x) && exists("L_HDL_L", x)) && exists("L_HDL_FC_pct", x)) {
    x[, L_HDL_FC_pct := NULL]
  }

  if (exists("L_HDL_C", x) && exists("L_HDL_L", x)) {
    x[, L_HDL_C_pct := collate_flags(L_HDL_C, L_HDL_L)]
  } else if (!(exists("L_HDL_C", x) && exists("L_HDL_L", x)) && exists("L_HDL_C_pct", x)) {
    x[, L_HDL_C_pct := NULL]
  }

  if (exists("L_HDL_PL", x) && exists("L_HDL_L", x)) {
    x[, L_HDL_PL_pct := collate_flags(L_HDL_PL, L_HDL_L)]
  } else if (!(exists("L_HDL_PL", x) && exists("L_HDL_L", x)) && exists("L_HDL_PL_pct", x)) {
    x[, L_HDL_PL_pct := NULL]
  }

  if (exists("L_HDL_TG", x) && exists("L_HDL_L", x)) {
    x[, L_HDL_TG_pct := collate_flags(L_HDL_TG, L_HDL_L)]
  } else if (!(exists("L_HDL_TG", x) && exists("L_HDL_L", x)) && exists("L_HDL_TG_pct", x)) {
    x[, L_HDL_TG_pct := NULL]
  }


  if (exists("M_HDL_CE", x) && exists("M_HDL_L", x)) {
    x[, M_HDL_CE_pct := collate_flags(M_HDL_CE, M_HDL_L)]
  } else if (!(exists("M_HDL_CE", x) && exists("M_HDL_L", x)) && exists("M_HDL_CE_pct", x)) {
    x[, M_HDL_CE_pct := NULL]
  }

  if (exists("M_HDL_FC", x) && exists("M_HDL_L", x)) {
    x[, M_HDL_FC_pct := collate_flags(M_HDL_FC, M_HDL_L)]
  } else if (!(exists("M_HDL_FC", x) && exists("M_HDL_L", x)) && exists("M_HDL_FC_pct", x)) {
    x[, M_HDL_FC_pct := NULL]
  }

  if (exists("M_HDL_C", x) && exists("M_HDL_L", x)) {
    x[, M_HDL_C_pct := collate_flags(M_HDL_C, M_HDL_L)]
  } else if (!(exists("M_HDL_C", x) && exists("M_HDL_L", x)) && exists("M_HDL_C_pct", x)) {
    x[, M_HDL_C_pct := NULL]
  }

  if (exists("M_HDL_PL", x) && exists("M_HDL_L", x)) {
    x[, M_HDL_PL_pct := collate_flags(M_HDL_PL, M_HDL_L)]
  } else if (!(exists("M_HDL_PL", x) && exists("M_HDL_L", x)) && exists("M_HDL_PL_pct", x)) {
    x[, M_HDL_PL_pct := NULL]
  }

  if (exists("M_HDL_TG", x) && exists("M_HDL_L", x)) {
    x[, M_HDL_TG_pct := collate_flags(M_HDL_TG, M_HDL_L)]
  } else if (!(exists("M_HDL_TG", x) && exists("M_HDL_L", x)) && exists("M_HDL_TG_pct", x)) {
    x[, M_HDL_TG_pct := NULL]
  }


  if (exists("S_HDL_CE", x) && exists("S_HDL_L", x)) {
    x[, S_HDL_CE_pct := collate_flags(S_HDL_CE, S_HDL_L)]
  } else if (!(exists("S_HDL_CE", x) && exists("S_HDL_L", x)) && exists("S_HDL_CE_pct", x)) {
    x[, S_HDL_CE_pct := NULL]
  }

  if (exists("S_HDL_FC", x) && exists("S_HDL_L", x)) {
    x[, S_HDL_FC_pct := collate_flags(S_HDL_FC, S_HDL_L)]
  } else if (!(exists("S_HDL_FC", x) && exists("S_HDL_L", x)) && exists("S_HDL_FC_pct", x)) {
    x[, S_HDL_FC_pct := NULL]
  }

  if (exists("S_HDL_C", x) && exists("S_HDL_L", x)) {
    x[, S_HDL_C_pct := collate_flags(S_HDL_C, S_HDL_L)]
  } else if (!(exists("S_HDL_C", x) && exists("S_HDL_L", x)) && exists("S_HDL_C_pct", x)) {
    x[, S_HDL_C_pct := NULL]
  }

  if (exists("S_HDL_PL", x) && exists("S_HDL_L", x)) {
    x[, S_HDL_PL_pct := collate_flags(S_HDL_PL, S_HDL_L)]
  } else if (!(exists("S_HDL_PL", x) && exists("S_HDL_L", x)) && exists("S_HDL_PL_pct", x)) {
    x[, S_HDL_PL_pct := NULL]
  }

  if (exists("S_HDL_TG", x) && exists("S_HDL_L", x)) {
    x[, S_HDL_TG_pct := collate_flags(S_HDL_TG, S_HDL_L)]
  } else if (!(exists("S_HDL_TG", x) && exists("S_HDL_L", x)) && exists("S_HDL_TG_pct", x)) {
    x[, S_HDL_TG_pct := NULL]
  }


  # Compute fatty acid percentages
  if (exists("Omega_3", x) && exists("Total_FA", x)) {
    x[, Omega_3_pct := collate_flags(Omega_3, Total_FA)]
  } else if (!(exists("Omega_3", x) && exists("Total_FA", x)) && exists("Omega_3_pct", x)) {
    x[, Omega_3_pct := NULL]
  }

  if (exists("Omega_6", x) && exists("Total_FA", x)) {
    x[, Omega_6_pct := collate_flags(Omega_6, Total_FA)]
  } else if (!(exists("Omega_6", x) && exists("Total_FA", x)) && exists("Omega_6_pct", x)) {
    x[, Omega_6_pct := NULL]
  }

  if (exists("LA", x) && exists("Total_FA", x)) {
    x[, LA_pct := collate_flags(LA, Total_FA)]
  } else if (!(exists("LA", x) && exists("Total_FA", x)) && exists("LA_pct", x)) {
    x[, LA_pct := NULL]
  }

  if (exists("MUFA", x) && exists("Total_FA", x)) {
    x[, MUFA_pct := collate_flags(MUFA, Total_FA)]
  } else if (!(exists("MUFA", x) && exists("Total_FA", x)) && exists("MUFA_pct", x)) {
    x[, MUFA_pct := NULL]
  }

  if (exists("PUFA", x) && exists("Total_FA", x)) {
    x[, PUFA_pct := collate_flags(PUFA, Total_FA)]
  } else if (!(exists("PUFA", x) && exists("Total_FA", x)) && exists("PUFA_pct", x)) {
    x[, PUFA_pct := NULL]
  }

  if (exists("SFA", x) && exists("Total_FA", x)) {
    x[, SFA_pct := collate_flags(SFA, Total_FA)]
  } else if (!(exists("SFA", x) && exists("Total_FA", x)) && exists("SFA_pct", x)) {
    x[, SFA_pct := NULL]
  }

  if (exists("DHA", x) && exists("Total_FA", x)) {
    x[, DHA_pct := collate_flags(DHA, Total_FA)]
  } else if (!(exists("DHA", x) && exists("Total_FA", x)) && exists("DHA_pct", x)) {
    x[, DHA_pct := NULL]
  }


  # Miscellaneous ratios
  if (exists("Omega_6", x) && exists("Omega_3", x)) {
    x[, Omega_6_by_Omega_3 := collate_flags(Omega_6, Omega_3)]
  } else if (!(exists("Omega_6", x) && exists("Omega_3", x)) && exists("Omega_6_by_Omega_3", x)) {
    x[, Omega_6_by_Omega_3 := NULL]
  }

  if (exists("ApoB", x) && exists("ApoA1", x)) {
    x[, ApoB_by_ApoA1 := collate_flags(ApoB, ApoA1)]
  } else if (!(exists("ApoB", x) && exists("ApoA1", x)) && exists("ApoB_by_ApoA1", x)) {
    x[, ApoB_by_ApoA1 := NULL]
  }

  if (exists("PUFA", x) && exists("MUFA", x)) {
    x[, PUFA_by_MUFA := collate_flags(PUFA, MUFA)]
  } else if (!(exists("PUFA", x) && exists("MUFA", x)) && exists("PUFA_by_MUFA", x)) {
    x[, PUFA_by_MUFA := NULL]
  }

  if (exists("Total_TG", x) && exists("Phosphoglyc", x)) {
    x[, TG_by_PG := collate_flags(Total_TG, Phosphoglyc)]
  } else if (!(exists("Total_TG", x) && exists("Phosphoglyc", x)) && exists("TG_by_PG", x)) {
    x[, TG_by_PG := NULL]
  }

  return(x)
}

nightingale_composite_biomarker_flags <- function(x) {
  # Silence CRAN NOTES about undefined global variables (columns in the scope of x)
  XL_VLDL_C <- HDL_C <- HDL_CE <- HDL_FC <- HDL_L <- HDL_P <- HDL_PL <- HDL_TG <-
    IDL_C <- IDL_CE <- IDL_FC <- IDL_L <- IDL_P <- IDL_PL <- IDL_TG <- Ile <-
    L_HDL_C <- L_HDL_CE <- L_HDL_FC <- L_HDL_L <- L_HDL_P <- L_HDL_PL <- L_HDL_TG <-
    L_LDL_C <- L_LDL_CE <- L_LDL_FC <- L_LDL_L <- L_LDL_P <- L_LDL_PL <- L_LDL_TG <-
    L_VLDL_C <- L_VLDL_CE <- L_VLDL_FC <- L_VLDL_L <- L_VLDL_P <- L_VLDL_PL <-
    L_VLDL_TG <- LDL_C <- LDL_CE <- LDL_FC <- LDL_L <- LDL_P <- LDL_PL <- LDL_TG <-
    Leu <- M_HDL_C <- M_HDL_CE <- M_HDL_FC <- M_HDL_L <- M_HDL_P <- M_HDL_PL <-
    M_HDL_TG <- M_LDL_C <- M_LDL_CE <- M_LDL_FC <- M_LDL_L <- M_LDL_P <- M_LDL_PL <-
    M_LDL_TG <- M_VLDL_C <- M_VLDL_CE <- M_VLDL_FC <- M_VLDL_L <- M_VLDL_P <-
    M_VLDL_PL <- M_VLDL_TG <- MUFA <- non_HDL_C <- Omega_3 <- Omega_6 <- PUFA <-
    Remnant_C <- S_HDL_C <- S_HDL_CE <- S_HDL_FC <- S_HDL_L <- S_HDL_P <- S_HDL_PL <-
    S_HDL_TG <- S_LDL_C <- S_LDL_CE <- S_LDL_FC <- S_LDL_L <- S_LDL_P <- S_LDL_PL <-
    S_LDL_TG <- S_VLDL_C <- S_VLDL_CE <- S_VLDL_FC <- S_VLDL_L <- S_VLDL_P <-
    S_VLDL_PL <- S_VLDL_TG <- SFA <- Total_BCAA <- Total_C <- Total_CE <- Total_FA <-
    Total_FC <- Total_L <- Total_P <- Total_PL <- Total_TG <- Val <- VLDL_C <-
    VLDL_CE <- VLDL_FC <- VLDL_L <- VLDL_P <- VLDL_PL <- VLDL_TG <- XL_HDL_C <-
    XL_HDL_CE <- XL_HDL_FC <- XL_HDL_L <- XL_HDL_P <- XL_HDL_PL <- XL_HDL_TG <-
    XL_VLDL_CE <- XL_VLDL_FC <- XL_VLDL_L <- XL_VLDL_P <- XL_VLDL_PL <- XL_VLDL_TG <-
    XS_VLDL_C <- XS_VLDL_CE <- XS_VLDL_FC <- XS_VLDL_L <- XS_VLDL_P <- XS_VLDL_PL <-
    XS_VLDL_TG <- XXL_VLDL_C <- XXL_VLDL_CE <- XXL_VLDL_FC <- XXL_VLDL_L <-
    XXL_VLDL_P <- XXL_VLDL_PL <- XXL_VLDL_TG <- NULL


  # First compute total cholesterol in 14 lipoprotein subclasses: these are the
  # sums of free cholesterol and cholesteryl esters.
  if (exists("XXL_VLDL_CE", x) && exists("XXL_VLDL_FC", x)) {
    x[, XXL_VLDL_C := collate_flags(XXL_VLDL_CE, XXL_VLDL_FC)]
  } else if (!(exists("XXL_VLDL_CE", x) && exists("XXL_VLDL_FC", x)) && exists("XXL_VLDL_C", x)) {
    x[, XXL_VLDL_C := NULL]
  }

  if (exists("XL_VLDL_CE", x) && exists("XL_VLDL_FC", x)) {
    x[, XL_VLDL_C := collate_flags(XL_VLDL_CE, XL_VLDL_FC)]
  } else if (!(exists("XL_VLDL_CE", x) && exists("XL_VLDL_FC", x)) && exists("XL_VLDL_C", x)) {
    x[, XL_VLDL_C := NULL]
  }

  if (exists("L_VLDL_CE", x) && exists("L_VLDL_FC", x)) {
    x[, L_VLDL_C := collate_flags(L_VLDL_CE, L_VLDL_FC)]
  } else if (!(exists("L_VLDL_CE", x) && exists("L_VLDL_FC", x)) && exists("L_VLDL_C", x)) {
    x[, L_VLDL_C := NULL]
  }

  if (exists("M_VLDL_CE", x) && exists("M_VLDL_FC", x)) {
    x[, M_VLDL_C := collate_flags(M_VLDL_CE, M_VLDL_FC)]
  } else if (!(exists("M_VLDL_CE", x) && exists("M_VLDL_FC", x)) && exists("M_VLDL_C", x)) {
    x[, M_VLDL_C := NULL]
  }

  if (exists("S_VLDL_CE", x) && exists("S_VLDL_FC", x)) {
    x[, S_VLDL_C := collate_flags(S_VLDL_CE, S_VLDL_FC)]
  } else if (!(exists("S_VLDL_CE", x) && exists("S_VLDL_FC", x)) && exists("S_VLDL_C", x)) {
    x[, S_VLDL_C := NULL]
  }

  if (exists("XS_VLDL_CE", x) && exists("XS_VLDL_FC", x)) {
    x[, XS_VLDL_C := collate_flags(XS_VLDL_CE, XS_VLDL_FC)]
  } else if (!(exists("XS_VLDL_CE", x) && exists("XS_VLDL_FC", x)) && exists("XS_VLDL_C", x)) {
    x[, XS_VLDL_C := NULL]
  }

  if (exists("IDL_CE", x) && exists("IDL_FC", x)) {
    x[, IDL_C := collate_flags(IDL_CE, IDL_FC)]
  } else if (!(exists("IDL_CE", x) && exists("IDL_FC", x)) && exists("IDL_C", x)) {
    x[, IDL_C := NULL]
  }

  if (exists("L_LDL_CE", x) && exists("L_LDL_FC", x)) {
    x[, L_LDL_C := collate_flags(L_LDL_CE, L_LDL_FC)]
  } else if (!(exists("L_LDL_CE", x) && exists("L_LDL_FC", x)) && exists("L_LDL_C", x)) {
    x[, L_LDL_C := NULL]
  }

  if (exists("M_LDL_CE", x) && exists("M_LDL_FC", x)) {
    x[, M_LDL_C := collate_flags(M_LDL_CE, M_LDL_FC)]
  } else if (!(exists("M_LDL_CE", x) && exists("M_LDL_FC", x)) && exists("M_LDL_C", x)) {
    x[, M_LDL_C := NULL]
  }

  if (exists("S_LDL_CE", x) && exists("S_LDL_FC", x)) {
    x[, S_LDL_C := collate_flags(S_LDL_CE, S_LDL_FC)]
  } else if (!(exists("S_LDL_CE", x) && exists("S_LDL_FC", x)) && exists("S_LDL_C", x)) {
    x[, S_LDL_C := NULL]
  }

  if (exists("XL_HDL_CE", x) && exists("XL_HDL_FC", x)) {
    x[, XL_HDL_C := collate_flags(XL_HDL_CE, XL_HDL_FC)]
  } else if (!(exists("XL_HDL_CE", x) && exists("XL_HDL_FC", x)) && exists("XL_HDL_C", x)) {
    x[, XL_HDL_C := NULL]
  }

  if (exists("L_HDL_CE", x) && exists("L_HDL_FC", x)) {
    x[, L_HDL_C := collate_flags(L_HDL_CE, L_HDL_FC)]
  } else if (!(exists("L_HDL_CE", x) && exists("L_HDL_FC", x)) && exists("L_HDL_C", x)) {
    x[, L_HDL_C := NULL]
  }

  if (exists("M_HDL_CE", x) && exists("M_HDL_FC", x)) {
    x[, M_HDL_C := collate_flags(M_HDL_CE, M_HDL_FC)]
  } else if (!(exists("M_HDL_CE", x) && exists("M_HDL_FC", x)) && exists("M_HDL_C", x)) {
    x[, M_HDL_C := NULL]
  }

  if (exists("S_HDL_CE", x) && exists("S_HDL_FC", x)) {
    x[, S_HDL_C := collate_flags(S_HDL_CE, S_HDL_FC)]
  } else if (!(exists("S_HDL_CE", x) && exists("S_HDL_FC", x)) && exists("S_HDL_C", x)) {
    x[, S_HDL_C := NULL]
  }


  # Now compute total lipids in lipoprotein subclasses
  # Total Lipids = cholesterol, phospholipids, triglycerides
  if (exists("XXL_VLDL_C", x) && exists("XXL_VLDL_PL", x) && exists("XXL_VLDL_TG", x)) {
    x[, XXL_VLDL_L := collate_flags(XXL_VLDL_C, XXL_VLDL_PL, XXL_VLDL_TG)]
  } else if (!(exists("XXL_VLDL_C", x) && exists("XXL_VLDL_PL", x) && exists("XXL_VLDL_TG", x)) && exists("XXL_VLDL_L", x)) {
    x[, XXL_VLDL_L := NULL]
  }

  if (exists("XL_VLDL_C", x) && exists("XL_VLDL_PL", x) && exists("XL_VLDL_TG", x)) {
    x[, XL_VLDL_L := collate_flags(XL_VLDL_C, XL_VLDL_PL, XL_VLDL_TG)]
  } else if (!(exists("XL_VLDL_C", x) && exists("XL_VLDL_PL", x) && exists("XL_VLDL_TG", x)) && exists("XL_VLDL_L", x)) {
    x[, XL_VLDL_L := NULL]
  }

  if (exists("L_VLDL_C", x) && exists("L_VLDL_PL", x) && exists("L_VLDL_TG", x)) {
    x[, L_VLDL_L := collate_flags(L_VLDL_C, L_VLDL_PL, L_VLDL_TG)]
  } else if (!(exists("L_VLDL_C", x) && exists("L_VLDL_PL", x) && exists("L_VLDL_TG", x)) && exists("L_VLDL_L", x)) {
    x[, L_VLDL_L := NULL]
  }

  if (exists("M_VLDL_C", x) && exists("M_VLDL_PL", x) && exists("M_VLDL_TG", x)) {
    x[, M_VLDL_L := collate_flags(M_VLDL_C, M_VLDL_PL, M_VLDL_TG)]
  } else if (!(exists("M_VLDL_C", x) && exists("M_VLDL_PL", x) && exists("M_VLDL_TG", x)) && exists("M_VLDL_L", x)) {
    x[, M_VLDL_L := NULL]
  }

  if (exists("S_VLDL_C", x) && exists("S_VLDL_PL", x) && exists("S_VLDL_TG", x)) {
    x[, S_VLDL_L := collate_flags(S_VLDL_C, S_VLDL_PL, S_VLDL_TG)]
  } else if (!(exists("S_VLDL_C", x) && exists("S_VLDL_PL", x) && exists("S_VLDL_TG", x)) && exists("S_VLDL_L", x)) {
    x[, S_VLDL_L := NULL]
  }

  if (exists("XS_VLDL_C", x) && exists("XS_VLDL_PL", x) && exists("XS_VLDL_TG", x)) {
    x[, XS_VLDL_L := collate_flags(XS_VLDL_C, XS_VLDL_PL, XS_VLDL_TG)]
  } else if (!(exists("XS_VLDL_C", x) && exists("XS_VLDL_PL", x) && exists("XS_VLDL_TG", x)) && exists("XS_VLDL_L", x)) {
    x[, XS_VLDL_L := NULL]
  }

  if (exists("IDL_C", x) && exists("IDL_PL", x) && exists("IDL_TG", x)) {
    x[, IDL_L := collate_flags(IDL_C, IDL_PL, IDL_TG)]
  } else if (!(exists("IDL_C", x) && exists("IDL_PL", x) && exists("IDL_TG", x)) && exists("IDL_L", x)) {
    x[, IDL_L := NULL]
  }

  if (exists("L_LDL_C", x) && exists("L_LDL_PL", x) && exists("L_LDL_TG", x)) {
    x[, L_LDL_L := collate_flags(L_LDL_C, L_LDL_PL, L_LDL_TG)]
  } else if (!(exists("L_LDL_C", x) && exists("L_LDL_PL", x) && exists("L_LDL_TG", x)) && exists("L_LDL_L", x)) {
    x[, L_LDL_L := NULL]
  }

  if (exists("M_LDL_C", x) && exists("M_LDL_PL", x) && exists("M_LDL_TG", x)) {
    x[, M_LDL_L := collate_flags(M_LDL_C, M_LDL_PL, M_LDL_TG)]
  } else if (!(exists("M_LDL_C", x) && exists("M_LDL_PL", x) && exists("M_LDL_TG", x)) && exists("M_LDL_L", x)) {
    x[, M_LDL_L := NULL]
  }

  if (exists("S_LDL_C", x) && exists("S_LDL_PL", x) && exists("S_LDL_TG", x)) {
    x[, S_LDL_L := collate_flags(S_LDL_C, S_LDL_PL, S_LDL_TG)]
  } else if (!(exists("S_LDL_C", x) && exists("S_LDL_PL", x) && exists("S_LDL_TG", x)) && exists("S_LDL_L", x)) {
    x[, S_LDL_L := NULL]
  }

  if (exists("XL_HDL_C", x) && exists("XL_HDL_PL", x) && exists("XL_HDL_TG", x)) {
    x[, XL_HDL_L := collate_flags(XL_HDL_C, XL_HDL_PL, XL_HDL_TG)]
  } else if (!(exists("XL_HDL_C", x) && exists("XL_HDL_PL", x) && exists("XL_HDL_TG", x)) && exists("XL_HDL_L", x)) {
    x[, XL_HDL_L := NULL]
  }

  if (exists("L_HDL_C", x) && exists("L_HDL_PL", x) && exists("L_HDL_TG", x)) {
    x[, L_HDL_L := collate_flags(L_HDL_C, L_HDL_PL, L_HDL_TG)]
  } else if (!(exists("L_HDL_C", x) && exists("L_HDL_PL", x) && exists("L_HDL_TG", x)) && exists("L_HDL_L", x)) {
    x[, L_HDL_L := NULL]
  }

  if (exists("M_HDL_C", x) && exists("M_HDL_PL", x) && exists("M_HDL_TG", x)) {
    x[, M_HDL_L := collate_flags(M_HDL_C, M_HDL_PL, M_HDL_TG)]
  } else if (!(exists("M_HDL_C", x) && exists("M_HDL_PL", x) && exists("M_HDL_TG", x)) && exists("M_HDL_L", x)) {
    x[, M_HDL_L := NULL]
  }

  if (exists("S_HDL_C", x) && exists("S_HDL_PL", x) && exists("S_HDL_TG", x)) {
    x[, S_HDL_L := collate_flags(S_HDL_C, S_HDL_PL, S_HDL_TG)]
  } else if (!(exists("S_HDL_C", x) && exists("S_HDL_PL", x) && exists("S_HDL_TG", x)) && exists("S_HDL_L", x)) {
    x[, S_HDL_L := NULL]
  }


  # Now compute totals for lipoprotein classes.
  # Eg. free cholesterol in LDL (LDL_FC) = sum(free choleseterol in LDL of different sizes).
  if (exists("XXL_VLDL_CE", x) && exists("XL_VLDL_CE", x) && exists("L_VLDL_CE", x) && exists("M_VLDL_CE", x) && exists("S_VLDL_CE", x) && exists("XS_VLDL_CE", x)) {
    x[, VLDL_CE := collate_flags(XXL_VLDL_CE, XL_VLDL_CE, L_VLDL_CE, M_VLDL_CE, S_VLDL_CE, XS_VLDL_CE)]
  } else if (!(exists("XXL_VLDL_CE", x) && exists("XL_VLDL_CE", x) && exists("L_VLDL_CE", x) && exists("M_VLDL_CE", x) && exists("S_VLDL_CE", x) && exists("XS_VLDL_CE", x)) && exists("VLDL_CE", x)) {
    x[, VLDL_CE := NULL]
  }

  if (exists("XXL_VLDL_FC", x) && exists("XL_VLDL_FC", x) && exists("L_VLDL_FC", x) && exists("M_VLDL_FC", x) && exists("S_VLDL_FC", x) && exists("XS_VLDL_FC", x)) {
    x[, VLDL_FC := collate_flags(XXL_VLDL_FC, XL_VLDL_FC, L_VLDL_FC, M_VLDL_FC, S_VLDL_FC, XS_VLDL_FC)]
  } else if (!(exists("XXL_VLDL_FC", x) && exists("XL_VLDL_FC", x) && exists("L_VLDL_FC", x) && exists("M_VLDL_FC", x) && exists("S_VLDL_FC", x) && exists("XS_VLDL_FC", x)) && exists("VLDL_FC", x)) {
    x[, VLDL_FC := NULL]
  }

  if (exists("XXL_VLDL_C", x) && exists("XL_VLDL_C", x) && exists("L_VLDL_C", x) && exists("M_VLDL_C", x) && exists("S_VLDL_C", x) && exists("XS_VLDL_C", x)) {
    x[, VLDL_C := collate_flags(XXL_VLDL_C, XL_VLDL_C, L_VLDL_C, M_VLDL_C, S_VLDL_C, XS_VLDL_C)]
  } else if (!(exists("XXL_VLDL_C", x) && exists("XL_VLDL_C", x) && exists("L_VLDL_C", x) && exists("M_VLDL_C", x) && exists("S_VLDL_C", x) && exists("XS_VLDL_C", x)) && exists("VLDL_C", x)) {
    x[, VLDL_C := NULL]
  }

  if (exists("XXL_VLDL_PL", x) && exists("XL_VLDL_PL", x) && exists("L_VLDL_PL", x) && exists("M_VLDL_PL", x) && exists("S_VLDL_PL", x) && exists("XS_VLDL_PL", x)) {
    x[, VLDL_PL := collate_flags(XXL_VLDL_PL, XL_VLDL_PL, L_VLDL_PL, M_VLDL_PL, S_VLDL_PL, XS_VLDL_PL)]
  } else if (!(exists("XXL_VLDL_PL", x) && exists("XL_VLDL_PL", x) && exists("L_VLDL_PL", x) && exists("M_VLDL_PL", x) && exists("S_VLDL_PL", x) && exists("XS_VLDL_PL", x)) && exists("VLDL_PL", x)) {
    x[, VLDL_PL := NULL]
  }

  if (exists("XXL_VLDL_TG", x) && exists("XL_VLDL_TG", x) && exists("L_VLDL_TG", x) && exists("M_VLDL_TG", x) && exists("S_VLDL_TG", x) && exists("XS_VLDL_TG", x)) {
    x[, VLDL_TG := collate_flags(XXL_VLDL_TG, XL_VLDL_TG, L_VLDL_TG, M_VLDL_TG, S_VLDL_TG, XS_VLDL_TG)]
  } else if (!(exists("XXL_VLDL_TG", x) && exists("XL_VLDL_TG", x) && exists("L_VLDL_TG", x) && exists("M_VLDL_TG", x) && exists("S_VLDL_TG", x) && exists("XS_VLDL_TG", x)) && exists("VLDL_TG", x)) {
    x[, VLDL_TG := NULL]
  }

  if (exists("XXL_VLDL_L", x) && exists("XL_VLDL_L", x) && exists("L_VLDL_L", x) && exists("M_VLDL_L", x) && exists("S_VLDL_L", x) && exists("XS_VLDL_L", x)) {
    x[, VLDL_L := collate_flags(XXL_VLDL_L, XL_VLDL_L, L_VLDL_L, M_VLDL_L, S_VLDL_L, XS_VLDL_L)]
  } else if (!(exists("XXL_VLDL_L", x) && exists("XL_VLDL_L", x) && exists("L_VLDL_L", x) && exists("M_VLDL_L", x) && exists("S_VLDL_L", x) && exists("XS_VLDL_L", x)) && exists("VLDL_L", x)) {
    x[, VLDL_L := NULL]
  }

  if (exists("XXL_VLDL_P", x) && exists("XL_VLDL_P", x) && exists("L_VLDL_P", x) && exists("M_VLDL_P", x) && exists("S_VLDL_P", x) && exists("XS_VLDL_P", x)) {
    x[, VLDL_P := collate_flags(XXL_VLDL_P, XL_VLDL_P, L_VLDL_P, M_VLDL_P, S_VLDL_P, XS_VLDL_P)]
  } else if (!(exists("XXL_VLDL_P", x) && exists("XL_VLDL_P", x) && exists("L_VLDL_P", x) && exists("M_VLDL_P", x) && exists("S_VLDL_P", x) && exists("XS_VLDL_P", x)) && exists("VLDL_P", x)) {
    x[, VLDL_P := NULL]
  }


  if (exists("L_LDL_CE", x) && exists("M_LDL_CE", x) && exists("S_LDL_CE", x)) {
    x[, LDL_CE := collate_flags(L_LDL_CE, M_LDL_CE, S_LDL_CE)]
  } else if (!(exists("L_LDL_CE", x) && exists("M_LDL_CE", x) && exists("S_LDL_CE", x)) && exists("LDL_CE", x)) {
    x[, LDL_CE := NULL]
  }

  if (exists("L_LDL_FC", x) && exists("M_LDL_FC", x) && exists("S_LDL_FC", x)) {
    x[, LDL_FC := collate_flags(L_LDL_FC, M_LDL_FC, S_LDL_FC)]
  } else if (!(exists("L_LDL_FC", x) && exists("M_LDL_FC", x) && exists("S_LDL_FC", x)) && exists("LDL_FC", x)) {
    x[, LDL_FC := NULL]
  }

  if (exists("L_LDL_C", x) && exists("M_LDL_C", x) && exists("S_LDL_C", x)) {
    x[, LDL_C := collate_flags(L_LDL_C, M_LDL_C, S_LDL_C)]
  } else if (!(exists("L_LDL_C", x) && exists("M_LDL_C", x) && exists("S_LDL_C", x)) && exists("LDL_C", x)) {
    x[, LDL_C := NULL]
  }

  if (exists("L_LDL_PL", x) && exists("M_LDL_PL", x) && exists("S_LDL_PL", x)) {
    x[, LDL_PL := collate_flags(L_LDL_PL, M_LDL_PL, S_LDL_PL)]
  } else if (!(exists("L_LDL_PL", x) && exists("M_LDL_PL", x) && exists("S_LDL_PL", x)) && exists("LDL_PL", x)) {
    x[, LDL_PL := NULL]
  }

  if (exists("L_LDL_TG", x) && exists("M_LDL_TG", x) && exists("S_LDL_TG", x)) {
    x[, LDL_TG := collate_flags(L_LDL_TG, M_LDL_TG, S_LDL_TG)]
  } else if (!(exists("L_LDL_TG", x) && exists("M_LDL_TG", x) && exists("S_LDL_TG", x)) && exists("LDL_TG", x)) {
    x[, LDL_TG := NULL]
  }

  if (exists("L_LDL_L", x) && exists("M_LDL_L", x) && exists("S_LDL_L", x)) {
    x[, LDL_L := collate_flags(L_LDL_L, M_LDL_L, S_LDL_L)]
  } else if (!(exists("L_LDL_L", x) && exists("M_LDL_L", x) && exists("S_LDL_L", x)) && exists("LDL_L", x)) {
    x[, LDL_L := NULL]
  }

  if (exists("L_LDL_P", x) && exists("M_LDL_P", x) && exists("S_LDL_P", x)) {
    x[, LDL_P := collate_flags(L_LDL_P, M_LDL_P, S_LDL_P)]
  } else if (!(exists("L_LDL_P", x) && exists("M_LDL_P", x) && exists("S_LDL_P", x)) && exists("LDL_P", x)) {
    x[, LDL_P := NULL]
  }


  if (exists("XL_HDL_CE", x) && exists("L_HDL_CE", x) && exists("M_HDL_CE", x) && exists("S_HDL_CE", x)) {
    x[, HDL_CE := collate_flags(XL_HDL_CE, L_HDL_CE, M_HDL_CE, S_HDL_CE)]
  } else if (!(exists("XL_HDL_CE", x) && exists("L_HDL_CE", x) && exists("M_HDL_CE", x) && exists("S_HDL_CE", x)) && exists("HDL_CE", x)) {
    x[, HDL_CE := NULL]
  }

  if (exists("XL_HDL_FC", x) && exists("L_HDL_FC", x) && exists("M_HDL_FC", x) && exists("S_HDL_FC", x)) {
    x[, HDL_FC := collate_flags(XL_HDL_FC, L_HDL_FC, M_HDL_FC, S_HDL_FC)]
  } else if (!(exists("XL_HDL_FC", x) && exists("L_HDL_FC", x) && exists("M_HDL_FC", x) && exists("S_HDL_FC", x)) && exists("HDL_FC", x)) {
    x[, HDL_FC := NULL]
  }

  if (exists("XL_HDL_C", x) && exists("L_HDL_C", x) && exists("M_HDL_C", x) && exists("S_HDL_C", x)) {
    x[, HDL_C := collate_flags(XL_HDL_C, L_HDL_C, M_HDL_C, S_HDL_C)]
  } else if (!(exists("XL_HDL_C", x) && exists("L_HDL_C", x) && exists("M_HDL_C", x) && exists("S_HDL_C", x)) && exists("HDL_C", x)) {
    x[, HDL_C := NULL]
  }

  if (exists("XL_HDL_PL", x) && exists("L_HDL_PL", x) && exists("M_HDL_PL", x) && exists("S_HDL_PL", x)) {
    x[, HDL_PL := collate_flags(XL_HDL_PL, L_HDL_PL, M_HDL_PL, S_HDL_PL)]
  } else if (!(exists("XL_HDL_PL", x) && exists("L_HDL_PL", x) && exists("M_HDL_PL", x) && exists("S_HDL_PL", x)) && exists("HDL_PL", x)) {
    x[, HDL_PL := NULL]
  }

  if (exists("XL_HDL_TG", x) && exists("L_HDL_TG", x) && exists("M_HDL_TG", x) && exists("S_HDL_TG", x)) {
    x[, HDL_TG := collate_flags(XL_HDL_TG, L_HDL_TG, M_HDL_TG, S_HDL_TG)]
  } else if (!(exists("XL_HDL_TG", x) && exists("L_HDL_TG", x) && exists("M_HDL_TG", x) && exists("S_HDL_TG", x)) && exists("HDL_TG", x)) {
    x[, HDL_TG := NULL]
  }

  if (exists("XL_HDL_L", x) && exists("L_HDL_L", x) && exists("M_HDL_L", x) && exists("S_HDL_L", x)) {
    x[, HDL_L := collate_flags(XL_HDL_L, L_HDL_L, M_HDL_L, S_HDL_L)]
  } else if (!(exists("XL_HDL_L", x) && exists("L_HDL_L", x) && exists("M_HDL_L", x) && exists("S_HDL_L", x)) && exists("HDL_L", x)) {
    x[, HDL_L := NULL]
  }

  if (exists("XL_HDL_P", x) && exists("L_HDL_P", x) && exists("M_HDL_P", x) && exists("S_HDL_P", x)) {
    x[, HDL_P := collate_flags(XL_HDL_P, L_HDL_P, M_HDL_P, S_HDL_P)]
  } else if (!(exists("XL_HDL_P", x) && exists("L_HDL_P", x) && exists("M_HDL_P", x) && exists("S_HDL_P", x)) && exists("HDL_P", x)) {
    x[, HDL_P := NULL]
  }


  # Next compute serum totals for lipids
  if (exists("VLDL_CE", x) && exists("LDL_CE", x) && exists("IDL_CE", x) && exists("HDL_CE", x)) {
    x[, Total_CE := collate_flags(VLDL_CE, LDL_CE, IDL_CE, HDL_CE)]
  } else if (!(exists("VLDL_CE", x) && exists("LDL_CE", x) && exists("IDL_CE", x) && exists("HDL_CE", x)) && exists("Total_CE", x)) {
    x[, Total_CE := NULL]
  }

  if (exists("VLDL_FC", x) && exists("LDL_FC", x) && exists("IDL_FC", x) && exists("HDL_FC", x)) {
    x[, Total_FC := collate_flags(VLDL_FC, LDL_FC, IDL_FC, HDL_FC)]
  } else if (!(exists("VLDL_FC", x) && exists("LDL_FC", x) && exists("IDL_FC", x) && exists("HDL_FC", x)) && exists("Total_FC", x)) {
    x[, Total_FC := NULL]
  }

  if (exists("VLDL_C", x) && exists("LDL_C", x) && exists("IDL_C", x) && exists("HDL_C", x)) {
    x[, Total_C := collate_flags(VLDL_C, LDL_C, IDL_C, HDL_C)]
  } else if (!(exists("VLDL_C", x) && exists("LDL_C", x) && exists("IDL_C", x) && exists("HDL_C", x)) && exists("Total_C", x)) {
    x[, Total_C := NULL]
  }

  if (exists("VLDL_PL", x) && exists("LDL_PL", x) && exists("IDL_PL", x) && exists("HDL_PL", x)) {
    x[, Total_PL := collate_flags(VLDL_PL, LDL_PL, IDL_PL, HDL_PL)]
  } else if (!(exists("VLDL_PL", x) && exists("LDL_PL", x) && exists("IDL_PL", x) && exists("HDL_PL", x)) && exists("Total_PL", x)) {
    x[, Total_PL := NULL]
  }

  if (exists("VLDL_TG", x) && exists("LDL_TG", x) && exists("IDL_TG", x) && exists("HDL_TG", x)) {
    x[, Total_TG := collate_flags(VLDL_TG, LDL_TG, IDL_TG, HDL_TG)]
  } else if (!(exists("VLDL_TG", x) && exists("LDL_TG", x) && exists("IDL_TG", x) && exists("HDL_TG", x)) && exists("Total_TG", x)) {
    x[, Total_TG := NULL]
  }

  if (exists("VLDL_L", x) && exists("LDL_L", x) && exists("IDL_L", x) && exists("HDL_L", x)) {
    x[, Total_L := collate_flags(VLDL_L, LDL_L, IDL_L, HDL_L)]
  } else if (!(exists("VLDL_L", x) && exists("LDL_L", x) && exists("IDL_L", x) && exists("HDL_L", x)) && exists("Total_L", x)) {
    x[, Total_L := NULL]
  }

  if (exists("VLDL_P", x) && exists("LDL_P", x) && exists("IDL_P", x) && exists("HDL_P", x)) {
    x[, Total_P := collate_flags(VLDL_P, LDL_P, IDL_P, HDL_P)]
  } else if (!(exists("VLDL_P", x) && exists("LDL_P", x) && exists("IDL_P", x) && exists("HDL_P", x)) && exists("Total_P", x)) {
    x[, Total_P := NULL]
  }


  # Finally miscellaneous composite biomarkers
  if (exists("Omega_3", x) && exists("Omega_6", x)) {
    x[, PUFA := collate_flags(Omega_3, Omega_6)]
  } else if (!(exists("Omega_3", x) && exists("Omega_6", x)) && exists("PUFA", x)) {
    x[, PUFA := NULL]
  }

  if (exists("PUFA", x) && exists("MUFA", x) && exists("SFA", x)) {
    x[, Total_FA := collate_flags(PUFA, MUFA, SFA)]
  } else if (!(exists("PUFA", x) && exists("MUFA", x) && exists("SFA", x)) && exists("Total_FA", x)) {
    x[, Total_FA := NULL]
  }


  # Miscellaneous composite markers
  if (exists("Leu", x) && exists("Ile", x) && exists("Val", x)) {
    x[, Total_BCAA := collate_flags(Leu, Ile, Val)] # total branched chain amino acids
  } else if (!(exists("Leu", x) && exists("Ile", x) && exists("Val", x)) && exists("Total_BCAA", x)) {
    x[, Total_BCAA := NULL]
  }

  if (exists("Total_C", x) && exists("HDL_C", x)) {
    x[, non_HDL_C := collate_flags(Total_C, HDL_C)] # non HDL cholesterol
  } else if (!(exists("Total_C", x) && exists("HDL_C", x)) && exists("non_HDL_C", x)) {
    x[, non_HDL_C := NULL]
  }

  if (exists("Total_C", x) && exists("HDL_C", x) && exists("LDL_C", x)) {
    x[, Remnant_C := collate_flags(Total_C, HDL_C, LDL_C)] # remnant cholesterol
  } else if (!(exists("Total_C", x) && exists("HDL_C", x) && exists("LDL_C", x)) && exists("Remnant_C", x)) {
    x[, Remnant_C := NULL]
  }

  # Finished
  return(x)
}

extended_ratios_flags <- function(x) {
  # Silence CRAN NOTES about undefined global variables (columns in the scope of x)
  HDL_C <- HDL_C_pct <- HDL_CE <- HDL_CE_pct <- HDL_CE_pct_C <- HDL_FC <- HDL_FC_by_CE <-
    HDL_FC_pct <- HDL_FC_pct_C <- HDL_L <- HDL_PL <- HDL_PL_pct <- HDL_TG <-
    HDL_TG_pct <- IDL_C <- IDL_CE <- IDL_CE_pct_C <- IDL_FC <- IDL_FC_by_CE <-
    IDL_FC_pct_C <- L_HDL_C <- L_HDL_CE <- L_HDL_CE_pct_C <- L_HDL_FC <- L_HDL_FC_by_CE <-
    L_HDL_FC_pct_C <- L_LDL_C <- L_LDL_CE <- L_LDL_CE_pct_C <- L_LDL_FC <-
    L_LDL_FC_by_CE <- L_LDL_FC_pct_C <- L_VLDL_C <- L_VLDL_CE <- L_VLDL_CE_pct_C <-
    L_VLDL_FC <- L_VLDL_FC_by_CE <- L_VLDL_FC_pct_C <- LDL_C <- LDL_C_pct <- LDL_CE <-
    LDL_CE_pct <- LDL_CE_pct_C <- LDL_FC <- LDL_FC_by_CE <- LDL_FC_pct <- LDL_FC_pct_C <-
    LDL_L <- LDL_PL <- LDL_PL_pct <- LDL_TG <- LDL_TG_pct <- M_HDL_C <- M_HDL_CE <-
    M_HDL_CE_pct_C <- M_HDL_FC <- M_HDL_FC_by_CE <- M_HDL_FC_pct_C <- M_LDL_C <-
    M_LDL_CE <- M_LDL_CE_pct_C <- M_LDL_FC <- M_LDL_FC_by_CE <- M_LDL_FC_pct_C <-
    M_VLDL_C <- M_VLDL_CE <- M_VLDL_CE_pct_C <- M_VLDL_FC <- M_VLDL_FC_by_CE <-
    M_VLDL_FC_pct_C <- Omega_3 <- Omega_3_pct_PUFA <- Omega_6 <- Omega_6_pct_PUFA <-
    PUFA <- S_HDL_C <- S_HDL_CE <- S_HDL_CE_pct_C <- S_HDL_FC <- S_HDL_FC_by_CE <-
    S_HDL_FC_pct_C <- S_LDL_C <- S_LDL_CE <- S_LDL_CE_pct_C <- S_LDL_FC <-
    S_LDL_FC_by_CE <- S_LDL_FC_pct_C <- S_VLDL_C <- S_VLDL_CE <- S_VLDL_CE_pct_C <-
    S_VLDL_FC <- S_VLDL_FC_by_CE <- S_VLDL_FC_pct_C <- Total_C <- Total_C_pct <-
    Total_CE <- Total_CE_pct <- Total_CE_pct_C <- Total_FC <- Total_FC_by_CE <-
    Total_FC_pct <- Total_FC_pct_C <- Total_L <- Total_PL <- Total_PL_pct <-
    Total_TG <- Total_TG_pct <- VLDL_C <- VLDL_C_pct <- VLDL_CE <- VLDL_CE_pct <-
    VLDL_CE_pct_C <- VLDL_FC <- VLDL_FC_by_CE <- VLDL_FC_pct <- VLDL_FC_pct_C <-
    VLDL_L <- VLDL_PL <- VLDL_PL_pct <- VLDL_TG <- VLDL_TG_pct <- XL_HDL_C <-
    XL_HDL_CE <- XL_HDL_CE_pct_C <- XL_HDL_FC <- XL_HDL_FC_by_CE <- XL_HDL_FC_pct_C <-
    XL_VLDL_C <- XL_VLDL_CE <- XL_VLDL_CE_pct_C <- XL_VLDL_FC <- XL_VLDL_FC_by_CE <-
    XL_VLDL_FC_pct_C <- XS_VLDL_C <- XS_VLDL_CE <- XS_VLDL_CE_pct_C <- XS_VLDL_FC <-
    XS_VLDL_FC_by_CE <- XS_VLDL_FC_pct_C <- XXL_VLDL_C <- XXL_VLDL_CE <-
    XXL_VLDL_CE_pct_C <- XXL_VLDL_FC <- XXL_VLDL_FC_by_CE <-
    XXL_VLDL_FC_pct_C <- NULL

  # Compute percentage of cholesterol from free vs. esterified cholesterol
  if (exists("XXL_VLDL_FC", x) && exists("XXL_VLDL_C", x)) {
    x[, XXL_VLDL_FC_pct_C := collate_flags(XXL_VLDL_FC, XXL_VLDL_C)]
  } else if (!(exists("XXL_VLDL_FC", x) && exists("XXL_VLDL_C", x)) && exists("XXL_VLDL_FC_pct_C", x)) {
    x[, XXL_VLDL_FC_pct_C := NULL]
  }

  if (exists("XL_VLDL_FC", x) && exists("XL_VLDL_C", x)) {
    x[, XL_VLDL_FC_pct_C := collate_flags(XL_VLDL_FC, XL_VLDL_C)]
  } else if (!(exists("XL_VLDL_FC", x) && exists("XL_VLDL_C", x)) && exists("XL_VLDL_FC_pct_C", x)) {
    x[, XL_VLDL_FC_pct_C := NULL]
  }

  if (exists("L_VLDL_FC", x) && exists("L_VLDL_C", x)) {
    x[, L_VLDL_FC_pct_C := collate_flags(L_VLDL_FC, L_VLDL_C)]
  } else if (!(exists("L_VLDL_FC", x) && exists("L_VLDL_C", x)) && exists("L_VLDL_FC_pct_C", x)) {
    x[, L_VLDL_FC_pct_C := NULL]
  }

  if (exists("M_VLDL_FC", x) && exists("M_VLDL_C", x)) {
    x[, M_VLDL_FC_pct_C := collate_flags(M_VLDL_FC, M_VLDL_C)]
  } else if (!(exists("M_VLDL_FC", x) && exists("M_VLDL_C", x)) && exists("M_VLDL_FC_pct_C", x)) {
    x[, M_VLDL_FC_pct_C := NULL]
  }

  if (exists("S_VLDL_FC", x) && exists("S_VLDL_C", x)) {
    x[, S_VLDL_FC_pct_C := collate_flags(S_VLDL_FC, S_VLDL_C)]
  } else if (!(exists("S_VLDL_FC", x) && exists("S_VLDL_C", x)) && exists("S_VLDL_FC_pct_C", x)) {
    x[, S_VLDL_FC_pct_C := NULL]
  }

  if (exists("XS_VLDL_FC", x) && exists("XS_VLDL_C", x)) {
    x[, XS_VLDL_FC_pct_C := collate_flags(XS_VLDL_FC, XS_VLDL_C)]
  } else if (!(exists("XS_VLDL_FC", x) && exists("XS_VLDL_C", x)) && exists("XS_VLDL_FC_pct_C", x)) {
    x[, XS_VLDL_FC_pct_C := NULL]
  }

  if (exists("L_LDL_FC", x) && exists("L_LDL_C", x)) {
    x[, L_LDL_FC_pct_C := collate_flags(L_LDL_FC, L_LDL_C)]
  } else if (!(exists("L_LDL_FC", x) && exists("L_LDL_C", x)) && exists("L_LDL_FC_pct_C", x)) {
    x[, L_LDL_FC_pct_C := NULL]
  }

  if (exists("M_LDL_FC", x) && exists("M_LDL_C", x)) {
    x[, M_LDL_FC_pct_C := collate_flags(M_LDL_FC, M_LDL_C)]
  } else if (!(exists("M_LDL_FC", x) && exists("M_LDL_C", x)) && exists("M_LDL_FC_pct_C", x)) {
    x[, M_LDL_FC_pct_C := NULL]
  }

  if (exists("S_LDL_FC", x) && exists("S_LDL_C", x)) {
    x[, S_LDL_FC_pct_C := collate_flags(S_LDL_FC, S_LDL_C)]
  } else if (!(exists("S_LDL_FC", x) && exists("S_LDL_C", x)) && exists("S_LDL_FC_pct_C", x)) {
    x[, S_LDL_FC_pct_C := NULL]
  }

  if (exists("IDL_FC", x) && exists("IDL_C", x)) {
    x[, IDL_FC_pct_C := collate_flags(IDL_FC, IDL_C)]
  } else if (!(exists("IDL_FC", x) && exists("IDL_C", x)) && exists("IDL_FC_pct_C", x)) {
    x[, IDL_FC_pct_C := NULL]
  }

  if (exists("XL_HDL_FC", x) && exists("XL_HDL_C", x)) {
    x[, XL_HDL_FC_pct_C := collate_flags(XL_HDL_FC, XL_HDL_C)]
  } else if (!(exists("XL_HDL_FC", x) && exists("XL_HDL_C", x)) && exists("XL_HDL_FC_pct_C", x)) {
    x[, XL_HDL_FC_pct_C := NULL]
  }

  if (exists("L_HDL_FC", x) && exists("L_HDL_C", x)) {
    x[, L_HDL_FC_pct_C := collate_flags(L_HDL_FC, L_HDL_C)]
  } else if (!(exists("L_HDL_FC", x) && exists("L_HDL_C", x)) && exists("L_HDL_FC_pct_C", x)) {
    x[, L_HDL_FC_pct_C := NULL]
  }

  if (exists("M_HDL_FC", x) && exists("M_HDL_C", x)) {
    x[, M_HDL_FC_pct_C := collate_flags(M_HDL_FC, M_HDL_C)]
  } else if (!(exists("M_HDL_FC", x) && exists("M_HDL_C", x)) && exists("M_HDL_FC_pct_C", x)) {
    x[, M_HDL_FC_pct_C := NULL]
  }

  if (exists("S_HDL_FC", x) && exists("S_HDL_C", x)) {
    x[, S_HDL_FC_pct_C := collate_flags(S_HDL_FC, S_HDL_C)]
  } else if (!(exists("S_HDL_FC", x) && exists("S_HDL_C", x)) && exists("S_HDL_FC_pct_C", x)) {
    x[, S_HDL_FC_pct_C := NULL]
  }


  if (exists("XXL_VLDL_CE", x) && exists("XXL_VLDL_C", x)) {
    x[, XXL_VLDL_CE_pct_C := collate_flags(XXL_VLDL_CE, XXL_VLDL_C)]
  } else if (!(exists("XXL_VLDL_CE", x) && exists("XXL_VLDL_C", x)) && exists("XXL_VLDL_CE_pct_C", x)) {
    x[, XXL_VLDL_CE_pct_C := NULL]
  }

  if (exists("XL_VLDL_CE", x) && exists("XL_VLDL_C", x)) {
    x[, XL_VLDL_CE_pct_C := collate_flags(XL_VLDL_CE, XL_VLDL_C)]
  } else if (!(exists("XL_VLDL_CE", x) && exists("XL_VLDL_C", x)) && exists("XL_VLDL_CE_pct_C", x)) {
    x[, XL_VLDL_CE_pct_C := NULL]
  }

  if (exists("L_VLDL_CE", x) && exists("L_VLDL_C", x)) {
    x[, L_VLDL_CE_pct_C := collate_flags(L_VLDL_CE, L_VLDL_C)]
  } else if (!(exists("L_VLDL_CE", x) && exists("L_VLDL_C", x)) && exists("L_VLDL_CE_pct_C", x)) {
    x[, L_VLDL_CE_pct_C := NULL]
  }

  if (exists("M_VLDL_CE", x) && exists("M_VLDL_C", x)) {
    x[, M_VLDL_CE_pct_C := collate_flags(M_VLDL_CE, M_VLDL_C)]
  } else if (!(exists("M_VLDL_CE", x) && exists("M_VLDL_C", x)) && exists("M_VLDL_CE_pct_C", x)) {
    x[, M_VLDL_CE_pct_C := NULL]
  }

  if (exists("S_VLDL_CE", x) && exists("S_VLDL_C", x)) {
    x[, S_VLDL_CE_pct_C := collate_flags(S_VLDL_CE, S_VLDL_C)]
  } else if (!(exists("S_VLDL_CE", x) && exists("S_VLDL_C", x)) && exists("S_VLDL_CE_pct_C", x)) {
    x[, S_VLDL_CE_pct_C := NULL]
  }

  if (exists("XS_VLDL_CE", x) && exists("XS_VLDL_C", x)) {
    x[, XS_VLDL_CE_pct_C := collate_flags(XS_VLDL_CE, XS_VLDL_C)]
  } else if (!(exists("XS_VLDL_CE", x) && exists("XS_VLDL_C", x)) && exists("XS_VLDL_CE_pct_C", x)) {
    x[, XS_VLDL_CE_pct_C := NULL]
  }

  if (exists("L_LDL_CE", x) && exists("L_LDL_C", x)) {
    x[, L_LDL_CE_pct_C := collate_flags(L_LDL_CE, L_LDL_C)]
  } else if (!(exists("L_LDL_CE", x) && exists("L_LDL_C", x)) && exists("L_LDL_CE_pct_C", x)) {
    x[, L_LDL_CE_pct_C := NULL]
  }

  if (exists("M_LDL_CE", x) && exists("M_LDL_C", x)) {
    x[, M_LDL_CE_pct_C := collate_flags(M_LDL_CE, M_LDL_C)]
  } else if (!(exists("M_LDL_CE", x) && exists("M_LDL_C", x)) && exists("M_LDL_CE_pct_C", x)) {
    x[, M_LDL_CE_pct_C := NULL]
  }

  if (exists("S_LDL_CE", x) && exists("S_LDL_C", x)) {
    x[, S_LDL_CE_pct_C := collate_flags(S_LDL_CE, S_LDL_C)]
  } else if (!(exists("S_LDL_CE", x) && exists("S_LDL_C", x)) && exists("S_LDL_CE_pct_C", x)) {
    x[, S_LDL_CE_pct_C := NULL]
  }

  if (exists("IDL_CE", x) && exists("IDL_C", x)) {
    x[, IDL_CE_pct_C := collate_flags(IDL_CE, IDL_C)]
  } else if (!(exists("IDL_CE", x) && exists("IDL_C", x)) && exists("IDL_CE_pct_C", x)) {
    x[, IDL_CE_pct_C := NULL]
  }

  if (exists("XL_HDL_CE", x) && exists("XL_HDL_C", x)) {
    x[, XL_HDL_CE_pct_C := collate_flags(XL_HDL_CE, XL_HDL_C)]
  } else if (!(exists("XL_HDL_CE", x) && exists("XL_HDL_C", x)) && exists("XL_HDL_CE_pct_C", x)) {
    x[, XL_HDL_CE_pct_C := NULL]
  }

  if (exists("L_HDL_CE", x) && exists("L_HDL_C", x)) {
    x[, L_HDL_CE_pct_C := collate_flags(L_HDL_CE, L_HDL_C)]
  } else if (!(exists("L_HDL_CE", x) && exists("L_HDL_C", x)) && exists("L_HDL_CE_pct_C", x)) {
    x[, L_HDL_CE_pct_C := NULL]
  }

  if (exists("M_HDL_CE", x) && exists("M_HDL_C", x)) {
    x[, M_HDL_CE_pct_C := collate_flags(M_HDL_CE, M_HDL_C)]
  } else if (!(exists("M_HDL_CE", x) && exists("M_HDL_C", x)) && exists("M_HDL_CE_pct_C", x)) {
    x[, M_HDL_CE_pct_C := NULL]
  }

  if (exists("S_HDL_CE", x) && exists("S_HDL_C", x)) {
    x[, S_HDL_CE_pct_C := collate_flags(S_HDL_CE, S_HDL_C)]
  } else if (!(exists("S_HDL_CE", x) && exists("S_HDL_C", x)) && exists("S_HDL_CE_pct_C", x)) {
    x[, S_HDL_CE_pct_C := NULL]
  }


  # Compute ratio of free cholesterol to esterified cholesterol
  if (exists("XXL_VLDL_FC", x) && exists("XXL_VLDL_CE", x)) {
    x[, XXL_VLDL_FC_by_CE := collate_flags(XXL_VLDL_FC, XXL_VLDL_CE)]
  } else if (!(exists("XXL_VLDL_FC", x) && exists("XXL_VLDL_CE", x)) && exists("XXL_VLDL_FC_by_CE", x)) {
    x[, XXL_VLDL_FC_by_CE := NULL]
  }

  if (exists("XL_VLDL_FC", x) && exists("XL_VLDL_CE", x)) {
    x[, XL_VLDL_FC_by_CE := collate_flags(XL_VLDL_FC, XL_VLDL_CE)]
  } else if (!(exists("XL_VLDL_FC", x) && exists("XL_VLDL_CE", x)) && exists("XL_VLDL_FC_by_CE", x)) {
    x[, XL_VLDL_FC_by_CE := NULL]
  }

  if (exists("L_VLDL_FC", x) && exists("L_VLDL_CE", x)) {
    x[, L_VLDL_FC_by_CE := collate_flags(L_VLDL_FC, L_VLDL_CE)]
  } else if (!(exists("L_VLDL_FC", x) && exists("L_VLDL_CE", x)) && exists("L_VLDL_FC_by_CE", x)) {
    x[, L_VLDL_FC_by_CE := NULL]
  }

  if (exists("M_VLDL_FC", x) && exists("M_VLDL_CE", x)) {
    x[, M_VLDL_FC_by_CE := collate_flags(M_VLDL_FC, M_VLDL_CE)]
  } else if (!(exists("M_VLDL_FC", x) && exists("M_VLDL_CE", x)) && exists("M_VLDL_FC_by_CE", x)) {
    x[, M_VLDL_FC_by_CE := NULL]
  }

  if (exists("S_VLDL_FC", x) && exists("S_VLDL_CE", x)) {
    x[, S_VLDL_FC_by_CE := collate_flags(S_VLDL_FC, S_VLDL_CE)]
  } else if (!(exists("S_VLDL_FC", x) && exists("S_VLDL_CE", x)) && exists("S_VLDL_FC_by_CE", x)) {
    x[, S_VLDL_FC_by_CE := NULL]
  }

  if (exists("XS_VLDL_FC", x) && exists("XS_VLDL_CE", x)) {
    x[, XS_VLDL_FC_by_CE := collate_flags(XS_VLDL_FC, XS_VLDL_CE)]
  } else if (!(exists("XS_VLDL_FC", x) && exists("XS_VLDL_CE", x)) && exists("XS_VLDL_FC_by_CE", x)) {
    x[, XS_VLDL_FC_by_CE := NULL]
  }

  if (exists("L_LDL_FC", x) && exists("L_LDL_CE", x)) {
    x[, L_LDL_FC_by_CE := collate_flags(L_LDL_FC, L_LDL_CE)]
  } else if (!(exists("L_LDL_FC", x) && exists("L_LDL_CE", x)) && exists("L_LDL_FC_by_CE", x)) {
    x[, L_LDL_FC_by_CE := NULL]
  }

  if (exists("M_LDL_FC", x) && exists("M_LDL_CE", x)) {
    x[, M_LDL_FC_by_CE := collate_flags(M_LDL_FC, M_LDL_CE)]
  } else if (!(exists("M_LDL_FC", x) && exists("M_LDL_CE", x)) && exists("M_LDL_FC_by_CE", x)) {
    x[, M_LDL_FC_by_CE := NULL]
  }

  if (exists("S_LDL_FC", x) && exists("S_LDL_CE", x)) {
    x[, S_LDL_FC_by_CE := collate_flags(S_LDL_FC, S_LDL_CE)]
  } else if (!(exists("S_LDL_FC", x) && exists("S_LDL_CE", x)) && exists("S_LDL_FC_by_CE", x)) {
    x[, S_LDL_FC_by_CE := NULL]
  }

  if (exists("IDL_FC", x) && exists("IDL_CE", x)) {
    x[, IDL_FC_by_CE := collate_flags(IDL_FC, IDL_CE)]
  } else if (!(exists("IDL_FC", x) && exists("IDL_CE", x)) && exists("IDL_FC_by_CE", x)) {
    x[, IDL_FC_by_CE := NULL]
  }

  if (exists("XL_HDL_FC", x) && exists("XL_HDL_CE", x)) {
    x[, XL_HDL_FC_by_CE := collate_flags(XL_HDL_FC, XL_HDL_CE)]
  } else if (!(exists("XL_HDL_FC", x) && exists("XL_HDL_CE", x)) && exists("XL_HDL_FC_by_CE", x)) {
    x[, XL_HDL_FC_by_CE := NULL]
  }

  if (exists("L_HDL_FC", x) && exists("L_HDL_CE", x)) {
    x[, L_HDL_FC_by_CE := collate_flags(L_HDL_FC, L_HDL_CE)]
  } else if (!(exists("L_HDL_FC", x) && exists("L_HDL_CE", x)) && exists("L_HDL_FC_by_CE", x)) {
    x[, L_HDL_FC_by_CE := NULL]
  }

  if (exists("M_HDL_FC", x) && exists("M_HDL_CE", x)) {
    x[, M_HDL_FC_by_CE := collate_flags(M_HDL_FC, M_HDL_CE)]
  } else if (!(exists("M_HDL_FC", x) && exists("M_HDL_CE", x)) && exists("M_HDL_FC_by_CE", x)) {
    x[, M_HDL_FC_by_CE := NULL]
  }

  if (exists("S_HDL_FC", x) && exists("S_HDL_CE", x)) {
    x[, S_HDL_FC_by_CE := collate_flags(S_HDL_FC, S_HDL_CE)]
  } else if (!(exists("S_HDL_FC", x) && exists("S_HDL_CE", x)) && exists("S_HDL_FC_by_CE", x)) {
    x[, S_HDL_FC_by_CE := NULL]
  }


  # Compute lipid percentages in lipoprotein classes
  if (exists("VLDL_CE", x) && exists("VLDL_L", x)) {
    x[, VLDL_CE_pct := collate_flags(VLDL_CE, VLDL_L)]
  } else if (!(exists("VLDL_CE", x) && exists("VLDL_L", x)) && exists("VLDL_CE_pct", x)) {
    x[, VLDL_CE_pct := NULL]
  }

  if (exists("VLDL_FC", x) && exists("VLDL_L", x)) {
    x[, VLDL_FC_pct := collate_flags(VLDL_FC, VLDL_L)]
  } else if (!(exists("VLDL_FC", x) && exists("VLDL_L", x)) && exists("VLDL_FC_pct", x)) {
    x[, VLDL_FC_pct := NULL]
  }

  if (exists("VLDL_C", x) && exists("VLDL_L", x)) {
    x[, VLDL_C_pct := collate_flags(VLDL_C, VLDL_L)]
  } else if (!(exists("VLDL_C", x) && exists("VLDL_L", x)) && exists("VLDL_C_pct", x)) {
    x[, VLDL_C_pct := NULL]
  }

  if (exists("VLDL_PL", x) && exists("VLDL_L", x)) {
    x[, VLDL_PL_pct := collate_flags(VLDL_PL, VLDL_L)]
  } else if (!(exists("VLDL_PL", x) && exists("VLDL_L", x)) && exists("VLDL_PL_pct", x)) {
    x[, VLDL_PL_pct := NULL]
  }

  if (exists("VLDL_TG", x) && exists("VLDL_L", x)) {
    x[, VLDL_TG_pct := collate_flags(VLDL_TG, VLDL_L)]
  } else if (!(exists("VLDL_TG", x) && exists("VLDL_L", x)) && exists("VLDL_TG_pct", x)) {
    x[, VLDL_TG_pct := NULL]
  }


  if (exists("LDL_CE", x) && exists("LDL_L", x)) {
    x[, LDL_CE_pct := collate_flags(LDL_CE, LDL_L)]
  } else if (!(exists("LDL_CE", x) && exists("LDL_L", x)) && exists("LDL_CE_pct", x)) {
    x[, LDL_CE_pct := NULL]
  }

  if (exists("LDL_FC", x) && exists("LDL_L", x)) {
    x[, LDL_FC_pct := collate_flags(LDL_FC, LDL_L)]
  } else if (!(exists("LDL_FC", x) && exists("LDL_L", x)) && exists("LDL_FC_pct", x)) {
    x[, LDL_FC_pct := NULL]
  }

  if (exists("LDL_C", x) && exists("LDL_L", x)) {
    x[, LDL_C_pct := collate_flags(LDL_C, LDL_L)]
  } else if (!(exists("LDL_C", x) && exists("LDL_L", x)) && exists("LDL_C_pct", x)) {
    x[, LDL_C_pct := NULL]
  }

  if (exists("LDL_PL", x) && exists("LDL_L", x)) {
    x[, LDL_PL_pct := collate_flags(LDL_PL, LDL_L)]
  } else if (!(exists("LDL_PL", x) && exists("LDL_L", x)) && exists("LDL_PL_pct", x)) {
    x[, LDL_PL_pct := NULL]
  }

  if (exists("LDL_TG", x) && exists("LDL_L", x)) {
    x[, LDL_TG_pct := collate_flags(LDL_TG, LDL_L)]
  } else if (!(exists("LDL_TG", x) && exists("LDL_L", x)) && exists("LDL_TG_pct", x)) {
    x[, LDL_TG_pct := NULL]
  }


  if (exists("HDL_CE", x) && exists("HDL_L", x)) {
    x[, HDL_CE_pct := collate_flags(HDL_CE, HDL_L)]
  } else if (!(exists("HDL_CE", x) && exists("HDL_L", x)) && exists("HDL_CE_pct", x)) {
    x[, HDL_CE_pct := NULL]
  }

  if (exists("HDL_FC", x) && exists("HDL_L", x)) {
    x[, HDL_FC_pct := collate_flags(HDL_FC, HDL_L)]
  } else if (!(exists("HDL_FC", x) && exists("HDL_L", x)) && exists("HDL_FC_pct", x)) {
    x[, HDL_FC_pct := NULL]
  }

  if (exists("HDL_C", x) && exists("HDL_L", x)) {
    x[, HDL_C_pct := collate_flags(HDL_C, HDL_L)]
  } else if (!(exists("HDL_C", x) && exists("HDL_L", x)) && exists("HDL_C_pct", x)) {
    x[, HDL_C_pct := NULL]
  }

  if (exists("HDL_PL", x) && exists("HDL_L", x)) {
    x[, HDL_PL_pct := collate_flags(HDL_PL, HDL_L)]
  } else if (!(exists("HDL_PL", x) && exists("HDL_L", x)) && exists("HDL_PL_pct", x)) {
    x[, HDL_PL_pct := NULL]
  }

  if (exists("HDL_TG", x) && exists("HDL_L", x)) {
    x[, HDL_TG_pct := collate_flags(HDL_TG, HDL_L)]
  } else if (!(exists("HDL_TG", x) && exists("HDL_L", x)) && exists("HDL_TG_pct", x)) {
    x[, HDL_TG_pct := NULL]
  }


  # Compute cholesterol percentages
  if (exists("VLDL_FC", x) && exists("VLDL_C", x)) {
    x[, VLDL_FC_pct_C := collate_flags(VLDL_FC, VLDL_C)]
  } else if (!(exists("VLDL_FC", x) && exists("VLDL_C", x)) && exists("VLDL_FC_pct_C", x)) {
    x[, VLDL_FC_pct_C := NULL]
  }

  if (exists("LDL_FC", x) && exists("LDL_C", x)) {
    x[, LDL_FC_pct_C := collate_flags(LDL_FC, LDL_C)]
  } else if (!(exists("LDL_FC", x) && exists("LDL_C", x)) && exists("LDL_FC_pct_C", x)) {
    x[, LDL_FC_pct_C := NULL]
  }

  if (exists("HDL_FC", x) && exists("HDL_C", x)) {
    x[, HDL_FC_pct_C := collate_flags(HDL_FC, HDL_C)]
  } else if (!(exists("HDL_FC", x) && exists("HDL_C", x)) && exists("HDL_FC_pct_C", x)) {
    x[, HDL_FC_pct_C := NULL]
  }


  if (exists("VLDL_CE", x) && exists("VLDL_C", x)) {
    x[, VLDL_CE_pct_C := collate_flags(VLDL_CE, VLDL_C)]
  } else if (!(exists("VLDL_CE", x) && exists("VLDL_C", x)) && exists("VLDL_CE_pct_C", x)) {
    x[, VLDL_CE_pct_C := NULL]
  }

  if (exists("LDL_CE", x) && exists("LDL_C", x)) {
    x[, LDL_CE_pct_C := collate_flags(LDL_CE, LDL_C)]
  } else if (!(exists("LDL_CE", x) && exists("LDL_C", x)) && exists("LDL_CE_pct_C", x)) {
    x[, LDL_CE_pct_C := NULL]
  }

  if (exists("HDL_CE", x) && exists("HDL_C", x)) {
    x[, HDL_CE_pct_C := collate_flags(HDL_CE, HDL_C)]
  } else if (!(exists("HDL_CE", x) && exists("HDL_C", x)) && exists("HDL_CE_pct_C", x)) {
    x[, HDL_CE_pct_C := NULL]
  }


  # Ratios of free cholesterol to esterified cholesterol:
  if (exists("VLDL_FC", x) && exists("VLDL_CE", x)) {
    x[, VLDL_FC_by_CE := collate_flags(VLDL_FC, VLDL_CE)]
  } else if (!(exists("VLDL_FC", x) && exists("VLDL_CE", x)) && exists("VLDL_FC_by_CE", x)) {
    x[, VLDL_FC_by_CE := NULL]
  }

  if (exists("LDL_FC", x) && exists("LDL_CE", x)) {
    x[, LDL_FC_by_CE := collate_flags(LDL_FC, LDL_CE)]
  } else if (!(exists("LDL_FC", x) && exists("LDL_CE", x)) && exists("LDL_FC_by_CE", x)) {
    x[, LDL_FC_by_CE := NULL]
  }

  if (exists("HDL_FC", x) && exists("HDL_CE", x)) {
    x[, HDL_FC_by_CE := collate_flags(HDL_FC, HDL_CE)]
  } else if (!(exists("HDL_FC", x) && exists("HDL_CE", x)) && exists("HDL_FC_by_CE", x)) {
    x[, HDL_FC_by_CE := NULL]
  }


  # Lipid and cholesterol fractions in total serum
  if (exists("Total_CE", x) && exists("Total_L", x)) {
    x[, Total_CE_pct := collate_flags(Total_CE, Total_L)]
  } else if (!(exists("Total_CE", x) && exists("Total_L", x)) && exists("Total_CE_pct", x)) {
    x[, Total_CE_pct := NULL]
  }

  if (exists("Total_FC", x) && exists("Total_L", x)) {
    x[, Total_FC_pct := collate_flags(Total_FC, Total_L)]
  } else if (!(exists("Total_FC", x) && exists("Total_L", x)) && exists("Total_FC_pct", x)) {
    x[, Total_FC_pct := NULL]
  }

  if (exists("Total_C", x) && exists("Total_L", x)) {
    x[, Total_C_pct := collate_flags(Total_C, Total_L)]
  } else if (!(exists("Total_C", x) && exists("Total_L", x)) && exists("Total_C_pct", x)) {
    x[, Total_C_pct := NULL]
  }

  if (exists("Total_PL", x) && exists("Total_L", x)) {
    x[, Total_PL_pct := collate_flags(Total_PL, Total_L)]
  } else if (!(exists("Total_PL", x) && exists("Total_L", x)) && exists("Total_PL_pct", x)) {
    x[, Total_PL_pct := NULL]
  }

  if (exists("Total_TG", x) && exists("Total_L", x)) {
    x[, Total_TG_pct := collate_flags(Total_TG, Total_L)]
  } else if (!(exists("Total_TG", x) && exists("Total_L", x)) && exists("Total_TG_pct", x)) {
    x[, Total_TG_pct := NULL]
  }


  if (exists("Total_FC", x) && exists("Total_C", x)) {
    x[, Total_FC_pct_C := collate_flags(Total_FC, Total_C)]
  } else if (!(exists("Total_FC", x) && exists("Total_C", x)) && exists("Total_FC_pct_C", x)) {
    x[, Total_FC_pct_C := NULL]
  }

  if (exists("Total_CE", x) && exists("Total_C", x)) {
    x[, Total_CE_pct_C := collate_flags(Total_CE, Total_C)]
  } else if (!(exists("Total_CE", x) && exists("Total_C", x)) && exists("Total_CE_pct_C", x)) {
    x[, Total_CE_pct_C := NULL]
  }

  if (exists("Total_FC", x) && exists("Total_CE", x)) {
    x[, Total_FC_by_CE := collate_flags(Total_FC, Total_CE)]
  } else if (!(exists("Total_FC", x) && exists("Total_CE", x)) && exists("Total_FC_by_CE", x)) {
    x[, Total_FC_by_CE := NULL]
  }


  # Omega 3 and Omega 6 percent of PUFA
  if (exists("Omega_3", x) && exists("PUFA", x)) {
    x[, Omega_3_pct_PUFA := collate_flags(Omega_3, PUFA)]
  } else if (!(exists("Omega_3", x) && exists("PUFA", x)) && exists("Omega_3_pct_PUFA", x)) {
    x[, Omega_3_pct_PUFA := NULL]
  }

  if (exists("Omega_6", x) && exists("PUFA", x)) {
    x[, Omega_6_pct_PUFA := collate_flags(Omega_6, PUFA)]
  } else if (!(exists("Omega_6", x) && exists("PUFA", x)) && exists("Omega_6_pct_PUFA", x)) {
    x[, Omega_6_pct_PUFA := NULL]
  }

  # Finished
  return(x)
}

# Helper function for processing user input data
#
# Columns in the UK Biobank data extracted on the Research Analysis Platform
# following the naming scheme p<field_id>_i<instance> where <field_id>
# corresponds to the biomarker, e.g. 23474 for 3-Hydroxybutyrate (given the variable
# name bOHbutyrate). <instance> corresponds to the timepoint of biomarker quantification,
# in the NMR data, either 0 for baseline assessment, or 1 for first repeat. The
# column names may also optionally have an addition _a<array_index> component
# reserved for multiple measures at the same timepoint, or in the case of the NMR
# biomarker data, the presence of multiple QC flags for a given measurement.
#
# This function extracts from raw UK Biobank data the fields corresponding to
# either (1) biomarker concentrations, (2) biomarker QC Flags, or (3) sample QC
# Flags (as per the calling function, indicated by the 'type' argument). Columns
# corresponding to each <instance> are aggregated instead to multiple rows per
# participant (i.e. so there is only 1 column per biomarker / QC Flag). In
# instances where there are multiple QC Flags for a biomarker measurement (at
# a single time point for a single person), these are collated into a single
# comma separated entry. Fields are renamed with the short biomarker variable
# name typically provided by Nightingale Health, listed in the Biomarker column
# in the nmr_info data sheet included with this package
#
# This function also handles column naming schemes for datasets predating the
# UKB Research Analysis Platform, e.g. those extracted by ukbconv which follow
# the naming scheme <field_id>-<instance>.<array_index>, by ukbconv_r which
# follow the naming scheme f<field_id>.<instance>.<array_index>, and the ukbtools
# R package.
#
process_data <- function(x, type) {
  # Silence CRAN NOTES about undefined global variables (columns in data.tables)
  Biomarker <- UKB.Field.ID <- QC.Flag.Field.ID <- visit_index <- repeat_index <-
    Name <- Shipment.Plate <- Well.Position.Within.Plate <- tmp_cname <- NULL

  # Determine format of data
  data_format <- detect_format(x, type)

  # Make sure its a data.table
  x <- getDT(x)

  # If already in correct format, no need to do anything extra, return.
  if (data_format == "processed") {
    return(x)
  }

  # Data extracted by ukbconv with the txt option and loaded in with fread
  if (names(x)[1] == "eid" && all(is.na(x[[1]]))) {
    warning("Fixing malformed column headers for loaded dataset generated by ukbconv txt")
    setnames(x, names(x), c("V1", names(x)[-ncol(x)]))
  }

  # Give field ids consistent names based on data format
  if (data_format == "raw") {
    setnames(x, gsub("-|\\.", "_", names(x)))
  } else if (data_format == "ukbtools") {
    setnames(x, gsub(".*_f", "", names(x)))
  } else if (data_format == "ukbconv_r") {
    setnames(x, gsub("\\.", "_", gsub("f\\.", "", names(x))))
  } else if (data_format == "ukb_rap") {
    missing_array_index <- names(x)[which(names(x) %like% "i[0-9]$")]
    setnames(x, missing_array_index, paste0(missing_array_index, "_a0"))
    setnames(x, gsub("_a", "_", names(x)))
    setnames(x, gsub("_i", "_", names(x)))
    setnames(x, gsub("^p", "", names(x)))
  }

  # Extract field IDs present
  field_ids <- data.table(UKB.Field.ID = names(x))
  field_ids[, UKB.Field.ID := suppressWarnings(as.integer(gsub("_.*", "", UKB.Field.ID)))]
  field_ids <- unique(field_ids[!is.na(UKB.Field.ID)])

  # Filter to those we want to extract
  if (type == "biomarkers") {
    field_ids <- field_ids[UKB.Field.ID %in% na.omit(ukbnmr::nmr_info$UKB.Field.ID)]
  } else if (type == "biomarker_qc_flags") {
    # Not all biomarkers have QC flags, but we need to know where the column is
    # missing due to this, or due to the user not having access.
    biomarker_fields <- na.omit(ukbnmr::nmr_info$UKB.Field.ID)
    biomarkers_present <- field_ids[UKB.Field.ID %in% biomarker_fields]
    biomarkers_present[ukbnmr::nmr_info, on = list(UKB.Field.ID), QC.Flag.Field.ID := QC.Flag.Field.ID]
    qc_flag_fields <- na.omit(ukbnmr::nmr_info$QC.Flag.Field.ID)
    qc_flags_present <- field_ids[UKB.Field.ID %in% qc_flag_fields]
    empty_flags_with_biomarkers <- biomarkers_present[!(QC.Flag.Field.ID %in% qc_flags_present$UKB.Field.ID)]
    field_ids <- rbind(qc_flags_present, empty_flags_with_biomarkers[,list(UKB.Field.ID=na.omit(QC.Flag.Field.ID))])
  } else if (type == "sample_qc_flags") {
    field_ids <- field_ids[UKB.Field.ID %in% na.omit(ukbnmr::sample_qc_info$UKB.Field.ID)]
  } else {
    stop("internal error: 'type' must be one of \"biomarkers\", \"biomarker_qc_flags\", or \"sample_qc_flags\"")
  }

  # Map to biomarker variable names
  if (type == "biomarkers") {
    field_ids[ukbnmr::nmr_info, on = list(UKB.Field.ID), Biomarker := Biomarker]
  } else if (type == "biomarker_qc_flags") {
    field_ids[ukbnmr::nmr_info, on = list(UKB.Field.ID=QC.Flag.Field.ID), Biomarker := Biomarker]
  } else if (type == "sample_qc_flags") {
    field_ids[ukbnmr::sample_qc_info, on = list(UKB.Field.ID), Biomarker := Name] # Not biomarkers but harmonises with rest of code
  }

  # Split out instance (visit) and array index (repeat measure) fields so they
  # are rows instead of columns
  visit_repeats <- setdiff(unique(gsub("^[0-9]+_", "", names(x))), "eid")
  x <- rbindlist(fill=TRUE, use.names=TRUE, lapply(visit_repeats, function(vr) {
    # Find columns matching this visit repeat pair (e.g. ending in _0_0)
    this_cols <- names(x)[grepl(pattern=paste0(vr, "$"), names(x))]

    # Filter to these columns
    this_x <- x[, .SD, .SDcols=c("eid", this_cols)]

    # Drop repeat visit pair label from column name
    setnames(this_x, this_cols, gsub(paste0("_", vr, "$"), "", this_cols))

    # Get fields present in the visit repeat pair
    this_fields <- intersect(names(this_x), field_ids$UKB.Field.ID)

    # Return empty for this visit repeat pair if no fields of interest present
    if (length(this_fields) == 0) {
      return(NULL)
    }

    # Add columns for visit and repeat index
    this_x[, visit_index := as.integer(gsub("_.*", "", vr))]
    this_x[, repeat_index := as.integer(gsub(".*_", "", vr))]

    # Filter to field IDs of interest
    this_x <- this_x[, .SD, .SDcols=c("eid", "visit_index", "repeat_index", this_fields)]

    # For biomarkers with no QC flags, add in missing columns
    miss_fields <- setdiff(field_ids$UKB.Field.ID, names(this_x))
    if (length(miss_fields) > 0) {
      for (fid in miss_fields) {
        this_x[, as.character(fid) := NA_integer_]
      }
      this_fields <- intersect(names(this_x), field_ids$UKB.Field.ID)
    }

    # Rename Field IDs to biomarker variable names
    this_names <- field_ids[UKB.Field.ID %in% this_fields]
    setnames(this_x, as.character(this_names$UKB.Field.ID), this_names$Biomarker)

    # Shipment.Plate may be integer64, convert to char
    if ("Shipment.Plate" %in% names(this_x) & inherits(this_x$Shipment.Plate, "integer64")) {
      this_x[, Shipment.Plate := as.character(Shipment.Plate)]
    } else if ("Shipment.Plate" %in% names(this_x)) {
      # If already a character string, may have "" instead of NA_character_ for
      # samples without NMR data
      this_x[Shipment.Plate == "", Shipment.Plate := NA_character_]
    }

    # Some columns may have "" instead of NA_character_ for samples without NMR data, fix
    for (this_col in names(this_x)[which(sapply(this_x, is.character))]) {
      setnames(this_x, this_col, "tmp_cname")
      this_x[tmp_cname == "", tmp_cname := NA_character_]
      setnames(this_x, "tmp_cname", this_col)
    }

    # Well.Position.Within.Plate may have lowercase row letters in the July 2023
    # UKB data release, fix:
    if ("Well.Position.Within.Plate" %in% names(this_x)) {
      this_x[, Well.Position.Within.Plate := toupper(Well.Position.Within.Plate)]
    }

    # Drop instance and array index combinations with all missing data
    # eid, visit_index, and array_index always non-missing
    this_x <- this_x[apply(this_x, 1, function(row) { sum(!is.na(row)) > 3L })]

    # Return to rbindlist - which will row-bind all data.tables in the list
    # (one data.table per visit repeat pair)
    return(this_x)
  }))


  # For biomarker QC flags, map integers to flag values:
  # https://biobank.ndph.ox.ac.uk/showcase/coding.cgi?id=2310
  if (type == "biomarker_qc_flags") {
    x <- get_biomarker_qc_flag_values(x)
  } else if (type == "sample_qc_flags") {
    x <- get_sample_qc_flag_values(x)
  }

  # Remove repeat_index column no longer needed
  x[, repeat_index := NULL]

  # Set column keys for fast joining by user
  setkeyv(x, c("eid", "visit_index"))

  # Finished processing
  return(x)
}

# Map integer representation of biomarker QC flags to string
# https://biobank.ndph.ox.ac.uk/showcase/coding.cgi?id=2310
get_biomarker_qc_flag_values <- function(x) {
  # Silence CRAN NOTES about undefined global variables (columns in data.tables)
  value <- visit_index <- repeat_index <- integer_rep <- flag <- variable <-
    eid <- NULL

  x <- suppressWarnings(melt(x, id.vars=c("eid", "visit_index", "repeat_index")))

  # Sometimes the integer representations have already been mapped to their
  # character representations, e.g. when loading data using the Rscript made by
  # ukbconv r, in which case we can skip the mapping step.
  if (is.integer(x$value)) { # check if values are integer (or character representations of integer values)
    if (!inherits(x$value, "integer")) {
      x[, value := as.integer(value)]
    }

    x[biomarker_flag_map, on = list(value=integer_rep), flag := flag]
    x[, value := NULL]
    setnames(x, "flag", "value")
  }

  # Where there are multiple flags per visit_index, concatenate
  x <- x[, list(repeat_index=0,
    value=ifelse(all(is.na(value)), NA_character_, paste(sort(na.omit(value)), collapse="; "))),
    by=list(eid, visit_index, variable)]

  # Cast back to wide format
  x <- dcast(x, eid + visit_index + repeat_index ~ variable, value.var="value")

  return(x)
}

# Map integer representation of biomarker QC flags to string
# https://biobank.ndph.ox.ac.uk/showcase/coding.cgi?id=2310
get_sample_qc_flag_values <- function(x) {
  # Silence CRAN NOTES about undefined global variables (columns in data.tables)
  Spectrometer <- Shipment.Plate <- High.Lactate <- High.Pyruvate <-
    Low.Glucose <- Low.Protein <- Measurement.Quality.Flagged <- integer_rep <-
    flag <- Resolved.Plate.Swaps <- Processing.Batch <- NULL

  if (exists("High.Lactate", where=x))
    x[, High.Lactate := ifelse(is.na(High.Lactate), NA_character_, "Yes")]
  if (exists("High.Pyruvate", where=x))
    x[, High.Pyruvate := ifelse(is.na(High.Pyruvate), NA_character_, "Yes")]
  if (exists("Low.Glucose", where=x))
    x[, Low.Glucose := ifelse(is.na(Low.Glucose), NA_character_, "Yes")]
  if (exists("Low.Protein", where=x))
    x[, Low.Protein := ifelse(is.na(Low.Protein), NA_character_, "Yes")]
  if (exists("Resolved.Plate.Swaps", where=x))
    x[, Resolved.Plate.Swaps := ifelse(is.na(Resolved.Plate.Swaps), NA_character_, "Yes")]

  if ("Spectrometer" %in% names(x)) {
    if (is.integer(x$Spectrometer)) {
      x[, Spectrometer := ifelse(is.na(Spectrometer), NA_character_, paste("Spectrometer", Spectrometer))]
    } else if (is.factor(x$Spectrometer)) {
      x[, Spectrometer := as.character(Spectrometer)]
    }
  }

  if ("Measurement.Quality.Flagged" %in% names(x)) {
    if (is.integer(x$Measurement.Quality.Flagged)) {
      x[, Measurement.Quality.Flagged := as.character(Measurement.Quality.Flagged)]
      x[measure_quality_map, on = list(Measurement.Quality.Flagged=integer_rep), Measurement.Quality.Flagged := flag]
    } else if (is.factor(x$Measurement.Quality.Flagged)) {
      x[, Measurement.Quality.Flagged := as.character(Measurement.Quality.Flagged)]
    }
  }

  if ("Shipment.Plate" %in% names(x)) {
    if (bit64::is.integer64(x$Shipment.Plate)) {
      x[, Shipment.Plate := as.character(Shipment.Plate)]
    }
    if (!is.character(x$Shipment.Plate) | !grepl("^Plate", stats::na.omit(x$Shipment.Plate)[1])) {
      x[, Shipment.Plate := ifelse(is.na(Shipment.Plate), NA_character_, paste("Plate", Shipment.Plate))]
    }
    # Sometimes the leading 0 gets dropped (e.g. if loaded as integer64), add these back in to match UK Biobank release
    x[!(Shipment.Plate %like% "^Plate 0"), Shipment.Plate := gsub("^Plate ", "Plate 0", Shipment.Plate)]
  }

  if ("Processing.Batch" %in% names(x)) {
    if (!is.character(x$Processing.Batch) | !grepl("^Plate", stats::na.omit(x$Processing.Batch)[1])) {
      x[, Processing.Batch := ifelse(is.na(Processing.Batch), NA_character_, paste("Batch", Processing.Batch))]
    }
  }

  return(x)
}

#' Compute extended set of biomarker ratios
#'
#' Computes 76 additional ratios not provided by the Nightingale platform. These
#' include lipid fractions in HDL, LDL, VLDL, and total serum, as well as
#' cholesterol fractions, and omega to polyunsaturated fatty acid ratios. See
#' \code{\link{nmr_info}} for details.
#'
#' @details
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' All biomarkers in the input data are also returned alongside the ratios computed
#' by this function.
#'
#' @param x \code{data.frame} containing NMR metabolomics data from UK Biobank.
#'   May either be raw field data output by
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.
#'
#' @return a \code{data.frame} with the additional computed biomarker ratios.
#'
#' @seealso \code{\link{nmr_info}} for list of computed biomarker ratios,
#'   \code{\link{compute_extended_ratio_qc_flags}()} for obtaining an
#'   aggregate of the biomarker QC flags from the biomarkers underlying each
#'   computed ratio, and \code{\link{extract_biomarkers}()} for details on how
#'   raw data from
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   is processed.
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' nmr <- compute_extended_ratios(ukb_data)
#'
#' @export
compute_extended_ratios <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  # Process data to correct format
  x <- process_data(x, type="biomarkers") # copy of x created if already in right format

  # Are the nightingale ratios already computed? If not, also compute these
  Type <- Nightingale <- Biomarker <- NULL # Silence CRAN NOTES for data.table columns
  nightingale_ratios <- ukbnmr::nmr_info[
    (Nightingale) & !(Type %in% c("Non-derived", "Composite")),
    Biomarker]
  has_nightingale_ratios <- (length(intersect(names(x), nightingale_ratios)) > 0)

  # compute ratios
  if (!has_nightingale_ratios)
    x <- nightingale_ratio_compute(x)

  x <- extended_ratios_compute(x)

  # Return
  returnDT(x)
}

#' Recompute composite biomarkers and ratios from the 107 non-derived biomarkers
#'
#' When adjusting biomarkers for unwanted biological covariates, it is desirable
#' to recompute composite biomarkers and ratios to ensure consistency in the
#' adjusted dataset. This function will compute all composite biomarkers and
#' ratios from their parts (see \code{\link{nmr_info}} for biomarker details).
#'
#' @details
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' All biomarkers in the input data are also returned alongside those computed
#' by this function.
#'
#' @param x \code{data.frame} containing NMR metabolomics data from UK Biobank.
#'   May either be raw field data output by
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.
#'
#' @return a \code{data.frame} with all composite biomarkers and ratios
#'   (re)computed from the 107 non-derived biomarkers (see \code{\link{nmr_info}}
#'   for details).
#'
#' @seealso \code{\link{nmr_info}} for list of computed biomarker ratios,
#'   \code{\link{recompute_derived_biomarker_qc_flags}()} for obtaining an
#'   aggregate of the biomarker QC flags from the biomarkers underlying each
#'   computed biomarker, and \code{\link{extract_biomarkers}()} for details on
#'   how the raw field data extracted by the Table Exporter tool is processed.
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' bio_qc <- recompute_derived_biomarkers(ukb_data)
#'
#' @export
recompute_derived_biomarkers <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  # Process data to correct format
  x <- process_data(x, type="biomarkers") # copy of x created if already in right format

  # Composite biomarkers *must* be recomputed before downstream ratios
  x <- nightingale_composite_biomarker_compute(x)
  x <- nightingale_ratio_compute(x)
  x <- extended_ratios_compute(x)

  # Return
  returnDT(x)
}

#' Aggregate QC Flags when computing the extended set of biomarker ratios
#'
#' For the 76 biomarker ratios computed by \code{\link{compute_extended_ratios}()},
#' aggregates the biomarker QC flags from the biomarkers composing each ratio
#' (see \code{\link{nmr_info}}), which can be useful for determining the reason
#' underlying missing values in the biomarker ratios.
#'
#' @details
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' Biomarker QC Flags in the input data are also returned alongside those
#' aggregated by this function for the computed biomarker ratios.
#'
#' @param x \code{data.frame} containing NMR metabolomics data from UK Biobank.
#'   May either be raw field data output by
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.
#'
#' @return a \code{data.frame} with QC flags aggregated for all computed
#'         biomarker ratios.
#'
#' @seealso \code{\link{nmr_info}} for list of computed biomarker ratios and
#'   \code{\link{extract_biomarkers}()} for details on how raw data from
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   is processed.
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' biomarker_qc_flags <- compute_extended_ratio_qc_flags(ukb_data)
#'
#' @export
compute_extended_ratio_qc_flags <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  # Process data to correct format
  x <- process_data(x, type="biomarker_qc_flags") # copy of x created if already in right format

  # Are the nightingale ratios already computed? If not, also compute these
  Type <- Nightingale <- Biomarker <- NULL # Silence CRAN NOTES for data.table columns
  nightingale_ratios <- ukbnmr::nmr_info[
    (Nightingale) & !(Type %in% c("Non-derived", "Composite")),
    Biomarker]
  has_nightingale_ratios <- (length(intersect(names(x), nightingale_ratios)) > 0)

  # compute ratios
  if (!has_nightingale_ratios)
    x <- nightingale_ratio_flags(x)

  x <- extended_ratios_flags(x)

  # Return
  returnDT(x)
}

#' Aggregate QC Flags when recomputing all composite and derived biomarkers
#'
#' For the 61 composite biomarkers, 81 Nightingale biomarker ratios, and 76
#' extended biomarker ratios computed by \code{\link{recompute_derived_biomarkers}()},
#' aggregates the biomarker QC flags from the underlying biomarkers
#' (see \code{\link{nmr_info}}).
#'
#' @details
#' If your UK Biobank project only has access to a subset of biomarkers, then
#' this function will only return the subset of ratios that can be computed from
#' the biomarker data provided.
#'
#' Biomarker QC Flags in the input data are also returned alongside those
#' aggregated by this function for the computed biomarker ratios.
#'
#' @param x \code{data.frame} containing NMR metabolomics data from UK Biobank.
#'   May either be raw field data output by
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.
#'
#' @return a \code{data.frame} with QC flags aggregated for all computed
#'         biomarkers and ratios.
#'
#' @seealso \code{\link{nmr_info}} for list of computed biomarker ratios and
#'   \code{\link{extract_biomarkers}()} for details on how raw data from
#'   \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
#'   is processed.
#'
#' @examples
#' ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
#' biomarker_qc_flags <- recompute_derived_biomarker_qc_flags(ukb_data)
#'
#' @export
recompute_derived_biomarker_qc_flags <- function(x) {
  # Check if running on package test_data, and if so, force data.table to be
  # single threaded so that we can avoid a NOTE on CRAN submission due to their
  # misconfigured debian server
  if (isTRUE(all.equal(x, ukbnmr::test_data))) {
    registered_threads <- getDTthreads()
    setDTthreads(1)
    on.exit({ setDTthreads(registered_threads) }) # re-register so no unintended side effects for users
  }

  # Process data to correct format
  x <- process_data(x, type="biomarker_qc_flags") # copy of x created if already in right format

  # Composite biomarkers *must* be recomputed before downstream ratios
  x <- nightingale_composite_biomarker_flags(x)
  x <- nightingale_ratio_flags(x)
  x <- extended_ratios_flags(x)

  # Return
  returnDT(x)
}

#' Nightingale biomarker information
#'
#' Contains details on the Nightingale biomarkers available in UK Biobank and
#' computed by this package.
#'
#' @format A data table with 325 rows and 8 columns:
#' \describe{
#'   \item{Biomarker}{Short column name assigned to the biomarker}
#'   \item{Description}{Biomarker description, matching the description field provided by UK Biobank and Nightingale Health}
#'   \item{Units}{Units of measurement for the biomarker ("mmol/L", "g/L", "nm", "degree", "ratio", or "\%")}
#'   \item{Type}{Biomarker type ("Non-derived", "Composite", "Ratio", or "Percentage")}
#'   \item{Group}{Biomarker group as provided by Nightingale Health}
#'   \item{Sub.Group}{Biomarker sub-group as provided by Nightingale Health}
#'   \item{Nightingale}{Logical. Indicates biomarker is quantified by the Nightingale Health platform}
#'   \item{UKB.Field.ID}{Field ID in UK Biobank, see \url{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}}
#'   \item{QC.Flag.Field.ID}{Field ID in UK Biobank for the biomarker QC Flags, see \url{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}}
#'   \item{Full.Formula}{For composite biomarkers and ratios, details formula through which the biomarker can be derived from the 107 non-derived biomarkers}
#'   \item{Simplified.Formula}{Simplified form of the full formula most clearly expressing how each composite biomarker and ratio can be rederived from other biomarkers}
#' }
"nmr_info"

#' Nightingale biomarker sample processing information
#'
#' Contains details on the sample processing and quality control information for
#' the NMR biomarker data in UK Biobank.
#'
#' @format A data table with 18 rows and 3 columns:
#' \describe{
#'   \item{Name}{Column name assigned to the sample information field}
#'   \item{Description}{Brief description of the field contents. Further details
#'   on the Nightingale sample QC columns can be found in the \href{https://biobank.ndph.ox.ac.uk/showcase/refer.cgi?id=3004}{accompanying resource on the UK Biobank showcase}.}
#'   \item{UKB.Field.ID}{Field ID in UK Biobank, see \url{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}.
#'   Rows missing \code{UKB.Field.ID} entries correspond to additional sample
#'   processing information derived from these fields and returned by \code{\link{remove_technical_variation}}.}
#' }
"sample_qc_info"

#' Data for testing package functions
#'
#' Dataset mimicking structure of decoded UK Biobank dataset of NMR metabolomics
#' biomarker concentrations and associated processing variables for testing
#' package functions.
#'
#' @format A data table with 50 rows and 735 columns with column names "eid"
#' followed by extracted UK Biobank field data of the format "p23649_i0",
#' "p23649_i1", \dots, "p23655_i1".
#'
#' @source Data in each column has been randomly drawn from the distribution
#' present in the UK Biobank dataset. Importantly, random sampling was performed
#' for each column separately, thus no rows represent real observations or
#' participants in UK Biobank.
"test_data"

---
title: "Guide to using the ukbnmr package"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Guide to using the ukbnmr package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

This package provides utilities for working with the [UK Biobank NMR metabolomics data](https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220).

There are three groups of functions in this package: 

1. [Data extraction](#data-extraction-functions)
2. [Removal of technical variation](#removal-of-technical-variation), 
3. [Recomputing derived biomarkers and computing additional biomarker ratios after adjustment for biological covariates](#methods-for-computing-derived-biomarkers-and-ratios-after-adjusting-for-biological-variation).

All functions are designed to be applied directly to the UK Biobank phenotype data on the [UK Biobank Research Analysis Platform](https://ukbiobank.dnanexus.com/landing) after the NMR metabolomics fields have been extracted using the [Table Exporter](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#create-a-tsv-or-csv-file-using-table-exporter) tool. 

This package also works with datasets predating the Research Analysis Platform, which have been extractedusing the [ukbconv](https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide) tool or processed with the [ukbtools](https://cran.r-project.org/package=ukbtools) R package.

This package also provides a `data.frame` of biomarker information, loaded as `nmr_info`, and a `data.frame` of sample processing information, loaded as `sample_qc_info`. See `help("nmr_info")` and `help("sample_qc_info")` for details on column contents. 

## Citation

If using this package to remove additional technical variation or compute additional biomarker ratios, please cite:

Ritchie S. C. *et al.*, Quality control and removal of technical variation of NMR metabolic biomarker data in ~120,000 UK Biobank participants, <ins>Sci Data</ins> *10* 64 (2023). doi: [10.1038/s41597-023-01949-y](https://www.nature.com/articles/s41597-023-01949-y).

Note that several updates have been made to the package and algorithm based on subsequent releases of NMR metabolomic biomarker data that have expanded to cover all ~500,000 UK Biobank participants with blood samples. These updates are described in more detail in the [Algorithms for removing technical varation](#algorithms-for-removing-technical-variation) section below. The impact of technical variation and its removal in the full UK Biobank data are also shown in the [Technical variation in the full UK Biobank NMR data](#technical-variation-in-the-full-uk-biobank-nmr-data) section below.

Citation is appreciated, but not expected, if simply using the data extraction functions for convenience to extract the NMR biomarker data and associated information as-is into analysis-ready data.frames. 

## Data Extraction Functions

Three data extraction functions are supplied by this package for extracting the UK Biobank NMR data and associated processing information and quality control tags into an analysis-ready format from the CSV or TSV files of field data saved by the [Table Exporter](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#create-a-tsv-or-csv-file-using-table-exporter) tool on the [UK Biobank Research Analysis Platform](https://ukbiobank.dnanexus.com/landing).

Exported field data saved by the [Table Exporter](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#create-a-tsv-or-csv-file-using-table-exporter) has column names following a [naming scheme](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#database-columns) with the format "p<field_id>_i<visit_index>_a<repeat_index>".   

The `extract_biomarkers()` function extracts from this raw field data a `data.frame` that contains one column per [NMR biomarker](https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220) which are labelled with short descriptive ( and analysis-friendly) column names for each biomarker. Each row of the extracted `data.frame` corresponds to a single observation for a UK Biobank participant at either baseline assessment (2006-2010) or the first repeat assessment (2012-2013): rows are uniquely identifiable by their combination of `"eid"` and `"visit_index"` columns. The `"eid"` column contains the project-specific identifier for each participant and the `"visit_index"` column contains either a 0 or 1 depending on whether the biomarker was quantified from blood samples taken at baseline assessemt (visit_index == 0) or at the first repeat assessment (visit_index == 1). Mappings between biomarker column names and UK Biobank field identifiers, along with detailed descriptions of each biomarker, are provided in the `nmr_info` `data.frame` that is bundled with this package.

The `extract_biomarker_qc_flags()` function similarly returns a `data.frame` with one column for each biomarker, with observations containing the [quality control flags](https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221) for the measurement of the respective biomarker for the UK Biobank participant and timepoint indicated in the `"eid"` and `"visit_index"` columns. Observations with no quality control flags contain `NA`. In instances where there were multiple quality control flags, the individual flags are separated by `"; "`.

The `extract_sample_qc_flags()` function similarly returns a `data.frame` with one column for each of the [NMR sample processing flags and quality control flags](https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222) for each sample for the respective UK Biobank participant (`"eid"`) and timepoint (`"visit_index"`). Mappings between sample processing column names and UK Biobank field identifiers, along with detailed descriptions of each sample processing flag, are provided in the `sample_qc_info` `data.frame` that is bundled with this package.

An example workflow for extracting these data and saving them for later use:

```{r eval=FALSE}
library(ukbnmr)
library(data.table) # for fast reading and writing of csv files using fread() and fwrite()

# Load exported field data saved by the Table Exporter tool on the RAP
exported <- fread("path/to/exported_ukbiobank_phenotype_data.csv")

nmr <- extract_biomarkers(exported)
biomarker_qc_flags <- extract_biomarker_qc_flags(exported)
sample_qc_flags <- extract_sample_qc_flags(exported)

fwrite(nmr, file="path/to/nmr_biomarker_data.csv")
fwrite(biomarker_qc_flags, file="path/to/nmr_biomarker_qc_flags.csv")
fwrite(sample_qc_flags, file="path/to/nmr_sample_qc_flags.csv")
```

Remember to use the `dx upload` tool provided by the UK Biobank Research Analysis Platform to [save these files to your persistant project storage](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/running-analysis-jobs/rstudio#uploading-local-files-to-the-project) for later use.

You can try this out using the test dataset bundled with the `ukbnmr` package:

```{r}
library(ukbnmr)

exported <- ukbnmr::test_data # see help("test_data") for more details

nmr <- extract_biomarkers(exported)
biomarker_qc_flags <- extract_biomarker_qc_flags(exported)
sample_qc_flags <- extract_sample_qc_flags(exported)
```

## Removal of technical variation

The `remove_technical_variation()` function removes additional technical variation present in the UK Biobank NMR data (see [section below](#algorithms-for-removing-technical-variation) for details), returning a `list` containing the corrected NMR biomarker data, biomarker QC flags, and sample processing information in analysis-ready `data.frame`s.

Note that the no prefiltering of samples or columns should be performed prior to running this function: the algorithms used for removing technical variation expect all the data to be present.

This function takes 40 minutes to run, and requires at least 32 GB of RAM, so you will want to save the output, rather than incorporate this function into your analysis scripts.
      
An example workflow for using this function and saving the output for loading into future R sessions or other programs:

```{r eval=FALSE}
library(ukbnmr)
library(data.table) # for fast reading and writing of csv files using fread() and fwrite()

# Load exported field data saved by the Table Exporter tool on the RAP
exported <- fread("path/to/exported_ukbiobank_phenotype_data.csv")

processed <- remove_technical_variation(exported) 

fwrite(processed$biomarkers, file="path/to/nmr_biomarker_data.csv")
fwrite(processed$biomarker_qc_flags, file="path/to/nmr_biomarker_qc_flags.csv")
fwrite(processed$sample_processing, file="path/to/nmr_sample_qc_flags.csv")
fwrite(processed$log_offset, file="path/to/nmr_biomarker_log_offset.csv")
fwrite(processed$outlier_plate_detection, file="path/to/outlier_plate_info.csv")
```

Remember to use the `dx upload` tool provided by the UK Biobank Research Analysis Platform to [save these files to your persistant project storage](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/running-analysis-jobs/rstudio#uploading-local-files-to-the-project) for later use.

You can try this out using the test dataset bundled with the `ukbnmr` package:

```{r}
library(ukbnmr)

exported <- ukbnmr::test_data # see help("test_data") for more details

processed <- remove_technical_variation(exported)
```

## Algorithms for removing technical variation

Three versions of the QC algorithm have been developed:

1. Version 1 was designed based on the first phase of data released to the public covering ~120,000 UK Biobank participants. 
2. Version 2 made several improvements to the algorithm based on the subsequent second public release of data covering an additional ~150,000 participants.
3. Version 3 **(the default)** makes some further minor tweaks primarily so that the algorithm is compatible with the full public data release covering all ~500,000 participants.

### Algorithm version 1

Version 1 of the algorithm is as described in [Ritchie *et al.* 2023](https://www.nature.com/articles/s41597-023-01949-y), which was developed based on the technical variation observed in the NMR metabolomics data in the first ~120,000 participants that were measured. In brief, this multi-step procedure applies the following steps in sequence:

  1. First biomarker data is filtered to the 107 biomarkers that cannot be derived from any combination of other biomarkers.
  2. Absolute concentrations are log transformed, with a small offset applied to biomarkers with concentrations of 0.
  3. Each biomarker is adjusted for the time between sample preparation and sample measurement (hours) on a log scale.
  4. Each biomarker is adjusted for systematic differences between rows (A-H) on the 96-well shipment plates.
  5. Each biomarker is adjusted for remaining systematic differences between columns (1-12) on the 96-well shipment plates.
  6. Each biomarker is adjusted for drift over time within each of the six spectrometers. To do so, samples are grouped into 10
     bins, within each spectrometer, by the date the majority of samples on their respective 96-well plates were measured.
  7. Regression residuals after the sequential adjustments are transformed back to absolute concentrations.
  8. Samples belonging to shipment plates that are outliers of non-biological origin are identified and set to missing.
  9. The 61 composite biomarkers and 81 biomarker ratios are recomputed from their adjusted parts.
  10. An additional 76 biomarker ratios of potential biological significance are computed.

### Algorithm version 2

Version 2 of the algorithm modifies this algorithm:

  - Steps 4 and 5 above are performed within each processing batch
  - Step 6 above is modified to:

      1. group samples into bins of ~2,000 samples 
      2. hard codes a bin split on spectrometer 5 between plates 0490000006726 and 0490000006714
      
The first modification was made as applying version 1 of the algorithm to the combined data from the first and second tranche of measurements revealed introduced stratification by well position when examining the correctedconcentrations in each data release separately.

The second modification was made to ensure consistent bin sizes across data releases when correcting for drift over time. Otherwise, spectrometers used in multiple data releases would have different bin sizes when adjusting different releases. A bin split is also hard coded on spectrometer 5 between plates 0490000006726 and 0490000006714 which correspond to a large change in concentrations akin to a spectrometer recalibration event most strongly observed for alanine concentrations.

### Algorithm version 3

Version 3 of the algorithm makes two further minor changes:

 1. Imputation of missing sample preparation times has been improved. Previously, any samples missing time of measurement (N=3 in the phase 2 public release) had their time of measurement set to 00:00. In version 3, the time of measurement is set to the median time of measurement for that spectrometer on that day, which is between 12:00-13:00, instead of 00:00.
 
 2. Underlying code for adjusting drift over time has been modified to accommodate the phase 3 public release, which includes one spectrometer with ~2,500 samples. Version 2 of the algorithm would split this into two bins, whereas version 3 keeps this as a single bin to better match the bin sizes of the rest of the spectrometers.

## Technical variation in the full UK Biobank NMR data

The phase 3 release (January 2025) of the UK Biobank data in covers all ~500,000 UK Biobank participants, including the ~122,000 measured as part of the phase 1 release (June 2021) and the ~170,000 measured as part of the phase 2 release (July 2023).

The following figures below summarise the impact of the possible sources of variation on this updated dataset, and the impact of applying version 3 of our algorithm for removing technical variation, similar to what was shown in Figure 2 and Figure 7 of [Ritchie *et al.* 2023](https://www.nature.com/articles/s41597-023-01949-y) for the phase 1 release data: 

![Boxplots showing variance explained (r<sup>2</sup> in linear regression) across the 249 NMR metabolic biomarkers by each possible technical covariate in the January 2025 release of the UK Biobank NMR data (left) and after application of version 3 of the algorithm for removing technical variation. The vertical red dashed line indicates 1% variance explained.](img/r2_explained_v3.png){width=98%}

![Repeatability improves amongst 11,483 blind duplicates (samples measured twice or thrice (N=4) with randomised sample identifier so that Nightingale Health were blinded to the identity of the repeated samples at time of sample measurement).](img/repeatability_v3.png){width=98%}

![Power for genetic associations increases after removing technical variation. Here, scatterplots compare the P-values from GWAS of 8.5 million common SNPs (MAF > 1%, INFO > 0.4) for selected biomarkers before removal of technical variation (x-axes) and after removal of technical variation with version 3 of our algorithm (y-axes). The dashed line on the diagonal shows y=x, where P-values are identical for both the original and post-QC concentrations. GWAS were performed in the 373,342 unrelated ["White British" genetic ancestry subset](https://www.nature.com/articles/s41586-018-0579-z) adjusting for age, sex, genotyping chip, and 10 genetic PCs as described in the Methods section of [Ritchie *et al.* 2023](https://www.nature.com/articles/s41597-023-01949-y). Biomarkers shown here are those most greatly affected by specific sources of technical variation: Alanine (drift over time within spectrometer), Albumin (outlier plates of undeterminable non-biological origin), Glycine (stratification by well position), and Histidine (Sample degredation time).](img/gwas_improvement_v3.png){width=98%}

Extended diagnostic plots showing the impact of technical variation and its removal on all biomarkers are available to download on FigShare at [10.6084/m9.figshare.27730101](https://doi.org/10.6084/m9.figshare.27730101).

## Technical variation in the phase 2 release of UK Biobank NMR data

The July 2023 release of the UK Biobank NMR data covered ~275,000 UK Biobank participants, including ~122,000 measured as part of the phase 1 release (June 2021).

The following figures below summarise the impact of the possible sources of variation on this updated dataset, and the impact of applying version 2 of our algorithm for removing technical variation, similar to what was shown in Figure 2 of [Ritchie *et al.* 2023](https://www.nature.com/articles/s41597-023-01949-y) for the phase 1 release data: 

![Boxplots showing variance explained (r<sup>2</sup> in linear regression) across the 249 NMR metabolic biomarkers by each possible technical covariate in the July 2023 release of the UK Biobank NMR data (left) and after application of version 2 of the algorithm for removing technical variation. The vertical red dashed line indicates 1% variance explained.](img/r2_explained_v2.png){width=98%}

![Repeatability improves amongst 6,811 blind duplicates (samples measured twice or thrice (N=4) with randomised sample identifier so that Nightingale Health were blinded to the identity of the repeated samples at time of sample measurement).](img/repeatability_v2.png){width=98%}

Extended diagnostic plots showing the impact of technical variation and its removal on all biomarkers are available to download on FigShare at [10.6084/m9.figshare.23686407](https://doi.org/10.6084/m9.figshare.23686407).

### Justification for algorithm version 2 updates

Our exploration of this updated data release (advance access under UK Biobank application 30418) revealed several changes were needed to our existing algorithm for removing technical variation developed on the phase 1 data.

First, we observed that correcting for systematic differences in well position (steps 4 and 5 of the algorithm) over all 275,000 participants introduced systematic differences between the phase 1 and phase 2 data release samples:

<center>
![](img/well_position_adjustment_v1.png){width=80%}
</center>

Further exploration revealed that the systematic differences in well position differed between processing batches, and we therefore modified our algorithm to correct for well position within each processing batch separately:

<center>
![](img/well_position_adjustment_v2.png){width=80%}
</center>

Second, we found that the method for correcting for drift over time (step 6 of the algorithm) needed updating for two reasons:

  1. Partitioning spectrometers into 10 bins meant that the bin sizes would differ between data releases, and that as the sample size increased within each spectrometer, more unwanted variation was being retained.
  2. A large change in concentrations between sequential sets of plates on spectrometer 5 was not being appropriately account for.

These issues were fixed as described above by setting a fixed bin size when correcting for drift over time, and hard-coding as a bin boundary the large change in concentrations observed on spectrometer 5 between plates 0490000006726 and 0490000006714. The fixed bin size chosen was 2,000 samples, matching the median bin size when applying version 1 of the algorithm to the phase 1 data release. This bin size is approximate, and differs between spectrometers. The number of bins each spectrometer is divided into is set as the total number of samples divided by 2,000. When allocating samples to bins, samples on the same plate or date are always allocated to the same bin.

The following figure illustrates the result of applying version 1 of the algorithm to the alanine concentrations in the July 2023 release of the UK Biobank NMR data covers ~275,000 UK Biobank participants:

<center>
![](img/drift_over_time_correction_v1.png){width=80%}
</center>

And how this changes with version 2 of the algorithm:

<center>
![](img/drift_over_time_correction_v2.png){width=80%}
</center>

## Methods for computing derived biomarkers and ratios after adjusting for biological variation

Analysts may wish to further adjust data for biological covariates. We provide an additional function, `recompute_derived_biomarkers()` to recompute all composite biomarkers and ratios from 107 non-derived biomarkers, which is useful for ensuring data consistency when adjusting for unwanted biological variation. A companion function, `recompute_derived_biomarker_qc_flags()` will aggregate the QC flags for the biomarkers underlying each composite biomarker and ratio.

Note these functions assume the data has been returned to absolute units after adjusting for covariates. For example the ratio of two biomarkers A and B is computed as A/B, which may not be true if the two biomarkers are on different scales (e.g. regression residuals) after adjustment.

If using these functions, please cite:

Ritchie S. C. *et al.*, Quality control and removal of technical variation of NMR metabolic biomarker data in ~120,000 UK Biobank participants, <ins>Sci Data</ins> *10* 64 (2023). doi: [10.1038/s41597-023-01949-y](https://www.nature.com/articles/s41597-023-01949-y).

A worked example of recomputing derived biomarkers after adjusting for age, sex, and BMI follows:

```{r eval=FALSE}
library(ukbnmr)
library(data.table) # for fast reading and writing of csv files using fread() and fwrite()
library(MASS) # Robust linear regression

# Load exported field data saved by the Table Exporter tool on the RAP
exported <- fread("path/to/exported_ukbiobank_phenotype_data.csv")

# First we need to remove the effects of technical variation before removing the 
# biological covariates
processed <- remove_technical_variation(exported)
tech_qc <- processed$biomarkers

# Write out all the component results as above
fwrite(tech_qc, file="path/to/nmr_biomarker_data.csv")
fwrite(processed$biomarker_qc_flags, file="path/to/nmr_biomarker_qc_flags.csv")
fwrite(processed$sample_processing, file="path/to/nmr_sample_qc_flags.csv")
fwrite(processed$log_offset, file="path/to/nmr_biomarker_log_offset.csv")
fwrite(processed$outlier_plate_detection, file="path/to/outlier_plate_info.csv")

# Second, we can now create an additional dataset that has been adjusted for 
# age, sex, and BMI.

# First, we need to extract the relevant fields from UK biobank and format as 
# above, assuming the relevant fields for age (field #21003), sex (field #31), 
# and BMI (field #21001) have also been saved by the Table Exporter tool in 
# path/to/exported_ukbiobank_phenotype_data.csv 
baseline_covar <- exported[, .(eid, age=p21003_i0, sex=p31_i0, bmi=p21001_i0)]
repeat_covar <- exported[, .(eid, age=p21003_i1, sex=p31_i1, bmi=p21001_i1)]
covar <- rbind(idcol="visit_index", "0"=baseline_covar, "1"=repeat_covar)
covar[, visit_index := as.integer(visit_index)]

# Combine covariates with processed NMR biomarker data. Filter the processed 
# biomarkers only to the non-derived set, as we will plan to rederive the sums 
# and ratios after adjusting for age sex and bmi
non_derived <- ukbnmr::nmr_info[Type == "Non-derived", Biomarker]
non_derived <- intersect(non_derived, names(nmr)) # In case some biomarkers are not in the data
bio_qc <- tech_qc[, .SD, .SDcols=c("eid", "visit_index", non_derived)]
bio_qc <- covar[bio_qc, on = .(eid, visit_index), nomatch=0]

# Transform to long format so we can operate on all biomarkers in bulk
bio_qc <- melt(bio_qc, id.vars=c("eid", "visit_index", "age", "sex", "bmi"), 
  variable.name="biomarker")

# Log transform biomarkers prior to adjustment. To do so, we need to add a small 
# offset to any measurements equal to 0. This code is copied directly from the 
# internals of the remove_technical_variation() function
log_offset <- bioqc[!is.na(value), 
  .(Minimum=min(value), Minimum.Non.Zero=min(value[value != 0])), 
  by=biomarker]
log_offset[, Log.Offset := ifelse(Minimum == 0, Minimum.Non.Zero / 2, 0)]
bio_qc[log_offset, on = .(biomarker), log_value := log(value + Log.Offset)]

# Now adjust for age, sex, and (log) BMI using robust linear regression from the MASS package
bio_qc[, adjusted := rlm(biomarker ~ age + factor(sex) + log(bmi))$residuals, by=biomarker]

# The adjusted residuals are normally distributed with mean 0. We can rescale to 
# absolute units by subtracting the mean of the original log concentrations, then 
# undoing the log transformation. This code is copied directly from the internals 
# of the remove_technical_variation() function.
bio_qc[, adjusted := adjusted + as.vector(coef(rlm(value ~ 1)))[1], by=biomarker]
bio_qc[, adjusted := exp(adjusted)]
bio_qc[log_offset, on = .(biomarker), adjusted := adjusted - Log.Offset] # remove log offset

# Some values that were 0 are now < 0, apply small right shift for these biomarkers 
# (the shift should be very small, essentially numeric error. This is worth double 
# checking!). Again, this code is copied directly from the internals of the 
# remove_technical_variation() function.
shift <- bio_qc[, .(Right.Shift=-pmin(0, min(adjusted))), by=biomarker]
log_offset <- log_offset[shift, on = .(biomarker)]
bio_qc[log_offset, on = .(biomarker), adjusted := adjusted + Right.Shift]

# Cast back to wide format so that each biomarker has its own column
bioqc <- dcast(bio_qc, eid + visit_index ~ biomarker, value.var="adjusted")

# Now we recompute the composite biomarkers and derived ratios and save the result.
bio_qc <- recompute_derived_biomarkers(bio_qc)
fwrite(bio_qc, file="path/to/nmr_biomarkers_adjusted_for_covariates.csv")

# You may also want to aggregate and save the quality control flags for each sample from 
# the biomarkers underlying each derived biomarker or ratio, adding them as additional 
# columns to the input data (see help("recompute_derived_biomarker_qc_flags")).
biomarker_qc_flags <- recompute_derived_biomarker_qc_flags(nmr)
fwrite(biomarker_qc_flags, file="path/to/biomarker_qc_flags.csv")
```

Finally, remember to use the `dx upload` tool provided by the UK Biobank Research Analysis Platform to [save these files to your persistant project storage](https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/running-analysis-jobs/rstudio#uploading-local-files-to-the-project) for later use.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_qc_flags_user.R
\name{recompute_derived_biomarker_qc_flags}
\alias{recompute_derived_biomarker_qc_flags}
\title{Aggregate QC Flags when recomputing all composite and derived biomarkers}
\usage{
recompute_derived_biomarker_qc_flags(x)
}
\arguments{
\item{x}{\code{data.frame} containing NMR metabolomics data from UK Biobank.
May either be raw field data output by
\href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.}
}
\value{
a \code{data.frame} with QC flags aggregated for all computed
        biomarkers and ratios.
}
\description{
For the 61 composite biomarkers, 81 Nightingale biomarker ratios, and 76
extended biomarker ratios computed by \code{\link{recompute_derived_biomarkers}()},
aggregates the biomarker QC flags from the underlying biomarkers
(see \code{\link{nmr_info}}).
}
\details{
If your UK Biobank project only has access to a subset of biomarkers, then
this function will only return the subset of ratios that can be computed from
the biomarker data provided.

Biomarker QC Flags in the input data are also returned alongside those
aggregated by this function for the computed biomarker ratios.
}
\examples{
ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
biomarker_qc_flags <- recompute_derived_biomarker_qc_flags(ukb_data)

}
\seealso{
\code{\link{nmr_info}} for list of computed biomarker ratios and
  \code{\link{extract_biomarkers}()} for details on how raw data from
  \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
  is processed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{nmr_info}
\alias{nmr_info}
\title{Nightingale biomarker information}
\format{
A data table with 325 rows and 8 columns:
\describe{
  \item{Biomarker}{Short column name assigned to the biomarker}
  \item{Description}{Biomarker description, matching the description field provided by UK Biobank and Nightingale Health}
  \item{Units}{Units of measurement for the biomarker ("mmol/L", "g/L", "nm", "degree", "ratio", or "\%")}
  \item{Type}{Biomarker type ("Non-derived", "Composite", "Ratio", or "Percentage")}
  \item{Group}{Biomarker group as provided by Nightingale Health}
  \item{Sub.Group}{Biomarker sub-group as provided by Nightingale Health}
  \item{Nightingale}{Logical. Indicates biomarker is quantified by the Nightingale Health platform}
  \item{UKB.Field.ID}{Field ID in UK Biobank, see \url{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}}
  \item{QC.Flag.Field.ID}{Field ID in UK Biobank for the biomarker QC Flags, see \url{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}}
  \item{Full.Formula}{For composite biomarkers and ratios, details formula through which the biomarker can be derived from the 107 non-derived biomarkers}
  \item{Simplified.Formula}{Simplified form of the full formula most clearly expressing how each composite biomarker and ratio can be rederived from other biomarkers}
}
}
\usage{
nmr_info
}
\description{
Contains details on the Nightingale biomarkers available in UK Biobank and
computed by this package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_biomarkers_user.R
\name{recompute_derived_biomarkers}
\alias{recompute_derived_biomarkers}
\title{Recompute composite biomarkers and ratios from the 107 non-derived biomarkers}
\usage{
recompute_derived_biomarkers(x)
}
\arguments{
\item{x}{\code{data.frame} containing NMR metabolomics data from UK Biobank.
May either be raw field data output by
\href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.}
}
\value{
a \code{data.frame} with all composite biomarkers and ratios
  (re)computed from the 107 non-derived biomarkers (see \code{\link{nmr_info}}
  for details).
}
\description{
When adjusting biomarkers for unwanted biological covariates, it is desirable
to recompute composite biomarkers and ratios to ensure consistency in the
adjusted dataset. This function will compute all composite biomarkers and
ratios from their parts (see \code{\link{nmr_info}} for biomarker details).
}
\details{
If your UK Biobank project only has access to a subset of biomarkers, then
this function will only return the subset of ratios that can be computed from
the biomarker data provided.

All biomarkers in the input data are also returned alongside those computed
by this function.
}
\examples{
ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
bio_qc <- recompute_derived_biomarkers(ukb_data)

}
\seealso{
\code{\link{nmr_info}} for list of computed biomarker ratios,
  \code{\link{recompute_derived_biomarker_qc_flags}()} for obtaining an
  aggregate of the biomarker QC flags from the biomarkers underlying each
  computed biomarker, and \code{\link{extract_biomarkers}()} for details on
  how the raw field data extracted by the Table Exporter tool is processed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ukbnmr.R
\docType{package}
\name{ukbnmr}
\alias{ukbnmr-package}
\alias{ukbnmr}
\title{Tools for processing the UK Biobank NMR metabolomics biomarker data}
\description{
@description
This package provides utilities for working with the
\href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{UK Biobank MR metabolomics data}.
Details are provided below, and in the package vignette (type \code{vignette("ukbnmr")} to view).
}
\details{
There are three groups of functions in this package: (1) data extraction,
(2) removal of technical variation, and (3) recomputing derived biomarkers
and biomarker ratios.

All functions can be applied directly to raw data extracted from UK Biobank.

This package also provides a \code{data.frame} of biomarker information, loaded
as \code{\link{nmr_info}}, and \code{data.frame} of sample processing information,
loaded as \code{\link{sample_qc_info}}.
}
\section{Data Extraction Functions}{

The \code{\link{extract_biomarkers}()} function will take a phenotype dataset extracted on
the UK Biobank Research Analysis Platform by the Table Exporter tool, extract the
\href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{NMR biomarker fields}
and give them short comprehensible column names as described in \code{\link{nmr_info}}.
Measurements are also split into multiple rows where a participant has measurements
at both baseline and first repeat assessment.

The \code{\link{extract_biomarker_qc_flags}()} function will take a phenotype
dataset extracted on the UK Biobank Research Analysis Platform by the Table
Exporter tool, extract the \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{Nightingale quality control flags}
for each biomarker measurement, returning a single column per biomarker
(corresponding to respective columns output by \code{\link{extract_biomarkers}()}).

The \code{\link{extract_sample_qc_flags}()} function will take a phenotype
dataset extracted on the UK Biobank Research Analysis Platform by the Table
Exporter tool and extract the \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{sample quality control tags}
for the Nightingale NMR metabolomics data.

These functions will also work with older datasets predating the UK Biobank
Research Analysis Platform, e.g. those extracted by \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv},
and/or by the ukbtools R package.
}

\section{Removal of technical variation}{

The \code{\link{remove_technical_variation}()} function  will take a phenotype
dataset extracted on the UK Biobank Research Analysis Platform by the Table
Exporter tool, extract all the biomarkers and QC flags, remove the effects of
technical variation on biomarker concentrations, and return a list
containing the adjusted NMR biomarker data, biomarker QC flags, and sample quality control
and processing information.

This applies a multistep process as described in Ritchie \emph{et al.} 2023:

\enumerate{
  \item{First biomarker data is filtered to the 107 biomarkers that
  cannot be derived from any combination of other biomarkers.}
  \item{Absolute concentrations are log transformed, with a small offset
  applied to biomarkers with concentrations of 0.}
  \item{Each biomarker is adjusted for the time between sample preparation
  and sample measurement (hours) on a log scale.}
  \item{Each biomarker is adjusted for systematic differences between rows
  (A-H) on the 96-well shipment plates.}
  \item{Each biomarker is adjusted for remaining systematic differences
  between columns (1-12) on the 96-well shipment plates.}
  \item{Each biomarker is adjusted for drift over time within each of the six
  spectrometers. To do so, samples are grouped into 10 bins, within each
  spectrometer, by the date the majority of samples on their respective
  96-well plates were measured.}
  \item{Regression residuals after the sequential adjustments are
  transformed back to absolute concentrations.}
  \item{Samples belonging to shipment plates that are outliers of
  non-biological origin are identified and set to missing.}
  \item{The 61 composite biomarkers and 81 biomarker ratios are recomputed
  from their adjusted parts.}
  \item{An additional 76 biomarker ratios of potential biological
  significance are computed.}
}

Further details can be found in Ritchie S. C. \emph{et al.} Quality control
and removal of technical variation of NMR metabolic biomarker data in
~120,000 UK Biobank participants, \emph{Sci Data} \strong{10}, 64 (2023). doi:
\href{https://www.nature.com/articles/s41597-023-01949-y}{10.1038/s41597-023-01949-y}
}

\section{Methods for computing biomarker ratios}{

The \code{\link{compute_extended_ratios}()} function will compute an extended
set of biomarker ratios expanding on the biomarkers available directly from
the Nightingale platform. A companion function, \code{\link{compute_extended_ratio_qc_flags}()},
will aggregate the QC flags for the biomarkers underlying each ratio.

The \code{\link{recompute_derived_biomarkers}()} function will recompute all
composite biomarkers and ratios from 107 non-derived biomarkers, which is
useful for ensuring data consistency when adjusting for unwanted biological
variation. This includes the extended biomarker rations computed by the
\code{\link{compute_extended_ratios}()} function. A companion function,
\code{\link{recompute_derived_biomarker_qc_flags}()} will aggregate the QC
flags for the biomarkers underlying each composite biomarker and ratio.
}

\seealso{
Useful links:
\itemize{
  \item Report bugs at \url{https://github.com/sritchie73/ukbnmr/issues}
}

}
\author{
\strong{Maintainer}: Scott C Ritchie \email{sritchie73@gmail.com} (\href{https://orcid.org/0000-0002-8454-9548}{ORCID})

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/biomarker_qc.R
\name{remove_technical_variation}
\alias{remove_technical_variation}
\title{Remove technical variation from NMR biomarker data in UK Biobank.}
\usage{
remove_technical_variation(
  x,
  remove.outlier.plates = TRUE,
  skip.biomarker.qc.flags = FALSE,
  version = 3L
)
}
\arguments{
\item{x}{\code{data.frame} containing a tabular phenotype dataset extracted by the
\href{https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#create-a-tsv-or-csv-file-using-table-exporter}{Table Exporter} tool on the
\href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}
containing the project specific sample id (eid) and all fields (and instances)
relating to the NMR metabolomics data (i.e. fields listed in showcase categories
\href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=220}{220},
\href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{221}, and
\href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{222}.}

\item{remove.outlier.plates}{logical, when set to \code{FALSE} biomarker
concentrations on outlier shipment plates (see details) are not set to
missing but simply flagged in the \code{biomarker_qc_flags} \code{data.frame}
in the returned \code{list}.}

\item{skip.biomarker.qc.flags}{logical, when set to \code{TRUE} biomarker QC
flags are not processed or returned.}

\item{version}{version of the QC algorithm to apply. Defaults to 3, the latest
version of the algorithm (see details).}
}
\value{
a \code{list} containing three \code{data.frames}: \describe{
  \item{biomarkers}{A \code{data.frame} with column names "eid",
       and "visit_index", containing project-specific sample identifier and
       UK Biobank visit index (0 for baseline assessment, 1 for first repeat
       assessment), followed by columns for each biomarker containing their
       absolute concentrations (or ratios thereof) adjusted for technical
       variation. See \code{\link{nmr_info}} for information on each biomarker.}
  \item{biomarker_qc_flags}{A \code{data.frame} with the same format as
        \code{biomarkers} with entries corresponding to the
        \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=221}{quality
        control indicators} for each sample. "High plate outlier" and "Low
        plate outlier" indicate the value was set to missing due to systematic
        abnormalities in the biomarker's concentration on the sample's shipment
        plate compared to all other shipment plates (see Details). For
        composite and derived biomarkers, quality control flags are aggregates
        of any quality control flags for the underlying biomarkers from which
        the composite biomarker or ratio is derived.}
  \item{sample_processing}{A \code{data.frame} containing the
        \href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{processing
        information and quality control indicators} for each sample, including
        those derived for removal of unwanted technical variation by this
        function. See \code{\link{sample_qc_info}} for details.}
  \item{log_offset}{A \code{data.frame} containing diagnostic information on
        the offset applied so that biomarkers with concentrations of 0 could
        be log transformed, and any right shift applied to prevent negative
        concentrations after rescaling adjusted residuals back to absolute
        concentrations. Should contain only biomarkers with minimum
        concentrations of 0 (in the "Minimum" column). "Minimum.Non.Zero"
        gives the smallest non-zero concentration for the biomarker.
        "Log.Offset" the small offset added to all samples prior to
        log transformation: half the mininum non-zero concentration.
        "Right.Shift" gives the small offset added to prevent negative
        concentrations that arise after rescaling residuals to log
        concentrations: this should be at least one order of magnitude
        smaller than the smallest non-zero value (i.e. the offset added
        should amount to noise in numeric precision for all samples). See
        publication for more details.}
  \item{outlier_plate_detection}{A \code{data.frame} containing diagnostic
        information and details of outlier plate detection. For each of the
        107 non-derived biomarkers, the median concentration on each of the
        1,352 plates was calculated, then plates were flagged as outliers if
        their median value deviated more than expected from the mean of plate
        medians. "Mean.Plate.Medians" gives the mean of the plate medians for
        each biomarker. "Lower.Limit" and "Upper.Limit" give the values below
        and above which plates are flagged as outliers based on their plate
        median. See publication for more details.}
  \item{algorithm_version}{Version of the algorithm run, either 1, 2, or 3
        (default).}
}
}
\description{
Remove technical variation from NMR biomarker data in UK Biobank.
}
\details{
Three versions of the QC algorithm have been developed. Version 1 was designed
based on the first phase of data released to the public covering ~120,000
UK Biobank participants. Version 2 made several improvements to the algorithm
based on the subsequent second public release of data covering an additional
~150,000 participants. Version 3, the default, makes some further minor tweaks
primarily so that the algorithm is compatible with the full public data
release covering all ~500,000 participants.

Details on the impact of these algorithms on technical variation in the latest
UK Biobank data are provided in the package vignette and on the github readme.

## Version 1
Setting version to 1 applies the algorithm as exactly described in Ritchie S.
C. *et al.* _Sci Data_ 2023. In brief, this multi-step procedure applies the
following steps in sequence:

1. First biomarker data is filtered to the 107 biomarkers that cannot be
   derived from any combination of other biomarkers.
2. Absolute concentrations are log transformed, with a small offset
   applied to biomarkers with concentrations of 0.
3. Each biomarker is adjusted for the time between sample preparation
   and sample measurement (hours).
4. Each biomarker is adjusted for systematic differences between rows
   (A-H) on the 96-well shipment plates.
5. Each biomarker is adjusted for remaining systematic differences
   between columns (1-12) on the 96-well shipment plates.
6. Each biomarker is adjusted for drift over time within each of the six
   spectrometers. To do so, samples are grouped into 10 bins, within each
   spectrometer, by the date the majority of samples on their respective
   96-well plates were measured.
7. Regression residuals after the sequential adjustments are
   transformed back to absolute concentrations.
8. Samples belonging to shipment plates that are outliers of
   non-biological origin are identified and set to missing.
9. The 61 composite biomarkers and 81 biomarker ratios are recomputed
   from their adjusted parts.
10. An additional 76 biomarker ratios of potential biological
    significance are computed.

At each step, adjustment for technical covariates is performed using
\link[MASS:rlm]{robust linear regression}. Plate row, plate column, and
sample measurement date bin are treated as factors, using the group with the
largest sample size as reference in the regression.

## Version 2
Version 2 of the algorithm modifies the above procedure to (1)
adjust for well and column within each processing batch separately in steps 4
and 5, and (2) in step 6 instead of splitting samples into 10 bins per
spectrometer uses a fixed bin size of approximately 2,000 samples per bin,
ensuring samples measured on the same plate and plates measured on the same
day are grouped into the same bin.

The first modification was made as applying version 1 of the algorithm to the
combined data from the first and second tranche of measurements revealed
introduced stratification by well position when examining the corrected
concentrations in each data release separately.

The second modification was made to ensure consistent bin sizes across data
releases when correcting for drift over time. Otherwise, spectrometers used
in multiple data releases would have different bin sizes when adjusting
different releases. A bin split is also hard coded on spectrometer 5 between
plates 0490000006726 and 0490000006714 which correspond to a large change in
concentrations akin to a spectrometer recalibration event most strongly
observed for alanine concentrations.

## Version 3
Version 3 of the algorithm makes two further minor changes:

1. Imputation of missing sample measurement times has been improved. Previously,
   any samples missing time of measurement (N=3 in the phase 2 public release)
   had their time of measurement set to 00:00. In version 3, the time of measurement
   is set to the median time of measurement for that spectrometer on that day,
   which is between 12:00-13:00, instead of 00:00.
2. Samples missing sample preparation dates and times (N=182 in the V20 UK
   Biobank data release) use their sample centrifugation date and time as a
   proxy to allow adjustment for time between sample prep and sample
   measurement. Sample centrifugation always takes place a short time after
   sample preparation. As sample centrifugation date and time is not made
   available via UK Biobank, this data and time is hard coded; all samples
   missing sample preparation date and time had a sample centrifugation date
   and time of 2022-12-20 06:39:03 in the extended advance access data, so
   we use this date and time for any samples missing sample preparation date
   and time.
3. Underlying code for adjusting drift over time has been modified to accommodate
   the phase 3 public release, which includes one spectrometer with ~2,500 samples.
   Version 2 of the algorithm would split this into two bins, whereas version 3
   keeps this as a single bin to better match the bin sizes of the rest of the
   spectrometers.
}
\examples{
ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
processed <- remove_technical_variation(ukb_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_biomarkers_user.R
\name{compute_extended_ratios}
\alias{compute_extended_ratios}
\title{Compute extended set of biomarker ratios}
\usage{
compute_extended_ratios(x)
}
\arguments{
\item{x}{\code{data.frame} containing NMR metabolomics data from UK Biobank.
May either be raw field data output by
\href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.}
}
\value{
a \code{data.frame} with the additional computed biomarker ratios.
}
\description{
Computes 76 additional ratios not provided by the Nightingale platform. These
include lipid fractions in HDL, LDL, VLDL, and total serum, as well as
cholesterol fractions, and omega to polyunsaturated fatty acid ratios. See
\code{\link{nmr_info}} for details.
}
\details{
If your UK Biobank project only has access to a subset of biomarkers, then
this function will only return the subset of ratios that can be computed from
the biomarker data provided.

All biomarkers in the input data are also returned alongside the ratios computed
by this function.
}
\examples{
ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
nmr <- compute_extended_ratios(ukb_data)

}
\seealso{
\code{\link{nmr_info}} for list of computed biomarker ratios,
  \code{\link{compute_extended_ratio_qc_flags}()} for obtaining an
  aggregate of the biomarker QC flags from the biomarkers underlying each
  computed ratio, and \code{\link{extract_biomarkers}()} for details on how
  raw data from
  \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
  is processed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{test_data}
\alias{test_data}
\title{Data for testing package functions}
\format{
A data table with 50 rows and 735 columns with column names "eid"
followed by extracted UK Biobank field data of the format "p23649_i0",
"p23649_i1", \dots, "p23655_i1".
}
\source{
Data in each column has been randomly drawn from the distribution
present in the UK Biobank dataset. Importantly, random sampling was performed
for each column separately, thus no rows represent real observations or
participants in UK Biobank.
}
\usage{
test_data
}
\description{
Dataset mimicking structure of decoded UK Biobank dataset of NMR metabolomics
biomarker concentrations and associated processing variables for testing
package functions.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{sample_qc_info}
\alias{sample_qc_info}
\title{Nightingale biomarker sample processing information}
\format{
A data table with 18 rows and 3 columns:
\describe{
  \item{Name}{Column name assigned to the sample information field}
  \item{Description}{Brief description of the field contents. Further details
  on the Nightingale sample QC columns can be found in the \href{https://biobank.ndph.ox.ac.uk/showcase/refer.cgi?id=3004}{accompanying resource on the UK Biobank showcase}.}
  \item{UKB.Field.ID}{Field ID in UK Biobank, see \url{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}.
  Rows missing \code{UKB.Field.ID} entries correspond to additional sample
  processing information derived from these fields and returned by \code{\link{remove_technical_variation}}.}
}
}
\usage{
sample_qc_info
}
\description{
Contains details on the sample processing and quality control information for
the NMR biomarker data in UK Biobank.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extractor_functions.R
\name{extract_sample_qc_flags}
\alias{extract_sample_qc_flags}
\title{Extract NMR sample QC flags from a data.frame of UK Biobank fields}
\usage{
extract_sample_qc_flags(x)
}
\arguments{
\item{x}{\code{data.frame} with column names "eid" followed by extracted
fields e.g. "p23649_i0", "p23649_i1", \dots, "p23655_i1".}
}
\value{
a \code{data.frame} or \code{data.table} with column names "eid"
       and "visit_index", followed by columns for each sample
       QC tag, e.g. "Shipment.Plate", \dots, "Low.Protein".
}
\description{
Given an input \code{data.frame} loaded from a dasaset of
\href{https://biobank.ndph.ox.ac.uk/showcase/label.cgi?id=222}{NMR metabolomics processing fields}
extracted by the Table Exporter tool on the
\href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}, this function
extracts the
\href{https://biobank.ndph.ox.ac.uk/showcase/showcase/docs/nmrm_app4.pdf}{sample quality control flags for the NMR metabolomics biomarker data} giving them short variable names as listed in the
\code{\link{sample_qc_info}} information data sheet available in this package.
}
\details{
Data sets extracted on the \href{https://ukbiobank.dnanexus.com/landing}{UK Biobank Research Analysis Platform}
have one row per UK Biobank participant, whose project specific sample
identifier is given in the first column named "eid". Columns following this
follow a \href{https://dnanexus.gitbook.io/uk-biobank-rap/working-on-the-research-analysis-platform/accessing-data/accessing-phenotypic-data#database-columns}{naming scheme}
based on the unique identifier of each field, assessment visit, and (optionally if relevant)
repeated measurement of "p<field_id>_i<visit_index>_a<repeat_index>". For example,
the Shipment Plate for each sample collected at baseline assessment has the
column name "p23649_i0". For the UKB NMR data, measurements are available at
baseline assessment and the first repeat assessment (e.g. "p23649_i1"). For the
UKB NMR data, the <repeat_index> is reserved for cases where biomarker
measurements have more than one QC Flag (see \code{\link{extract_biomarker_qc_flags}()}).

The \code{data.frame} returned by this function gives each field a unique
recognizable name, with measurements from baseline and repeat assessment
given in separate rows. The "visit_index" column immediately after the "eid"
column indicates whether the biomarker measurement was quantified from the
blood samples taken at baseline assessment (visit_index == 0) or first repeat
assessment (visit_index == 1). Rows are uniquely identifiable
by the combination of entries in columns "eid" and "visit_index".

This function will also work with data predating the Research Analysis Platform,
including data sets extracted by the \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
tool and/or the \code{ukbtools} R package.

If your UK Biobank project only has access to a subset of biomarkers, then
this function will only return the subset of ratios that can be computed from
the biomarker data provided.

A \code{data.table} will be returned instead of a \code{data.frame} if the
the user has loaded the package into their R session.
}
\examples{
ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
sample_qc_flags <- extract_sample_qc_flags(ukb_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compute_qc_flags_user.R
\name{compute_extended_ratio_qc_flags}
\alias{compute_extended_ratio_qc_flags}
\title{Aggregate QC Flags when computing the extended set of biomarker ratios}
\usage{
compute_extended_ratio_qc_flags(x)
}
\arguments{
\item{x}{\code{data.frame} containing NMR metabolomics data from UK Biobank.
May either be raw field data output by
\href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
or data with column names corresponding to biomarkers listed in \code{\link{nmr_info}}.}
}
\value{
a \code{data.frame} with QC flags aggregated for all computed
        biomarker ratios.
}
\description{
For the 76 biomarker ratios computed by \code{\link{compute_extended_ratios}()},
aggregates the biomarker QC flags from the biomarkers composing each ratio
(see \code{\link{nmr_info}}), which can be useful for determining the reason
underlying missing values in the biomarker ratios.
}
\details{
If your UK Biobank project only has access to a subset of biomarkers, then
this function will only return the subset of ratios that can be computed from
the biomarker data provided.

Biomarker QC Flags in the input data are also returned alongside those
aggregated by this function for the computed biomarker ratios.
}
\examples{
ukb_data <- ukbnmr::test_data # Toy example dataset for testing package
biomarker_qc_flags <- compute_extended_ratio_qc_flags(ukb_data)

}
\seealso{
\code{\link{nmr_info}} for list of computed biomarker ratios and
  \code{\link{extract_biomarkers}()} for details on how raw data from
  \href{https://biobank.ctsu.ox.ac.uk/crystal/exinfo.cgi?src=accessing_data_guide}{ukbconv}
  is processed.
}
